//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// 	$Id: AdvSkills.h,v 1.28 2000/10/19 15:27:41 goetz Exp $
//      $Manager: aarn (Axel Arnold)$
//
// ////////////////////////////////////////////////////////////////////

#ifndef SKILLS_H
#define SKILLS_H

/** all basic skills of the player.
    All skills calculate and execute one action leading to a target given
    by the agent. They return a state from which the agent can
    decide wether another action is needed or not.
*/
class AdvSkills
{
public:
  /** @name variable constants.
      these are values that normally should not change permanently during
      a game, but that may need to be adjusted to the specific settings of
      a match or of the agent using them.
  */
  //@{
  /// assumed number precision. Default: 1e-5
  double precision;
  /** the range in which a kick is likely successful.
      Default: .87(PLAYER_SIZE+BALL_SIZE+KICKABLE_MARGIN) */
  double kickRange;
  /** velocities below this are considered zero.
      Default: 0.1 */
  double minSpeed;
  /** how far player is placed from ball by prepare before going for it.
      Default: 3*kickRange */
  double prepareOuterLimit;
  /** distance player should keep to ball when preparing a
      kick in before directly intercepting the ball.
      Default: 2*collRange */
  double prepareInnerLimit;
  /// time until that intercept times should be calculated
  int minInterceptTime;
  /// angle threshold for goToPosition. Lower values mean higher correction
  double goToThreshStretch;
  /** turn(Heads) below this are marked FINISHED.
      Default: 15 */
  double turnMinAngle;
  /// time to wait between two turnHeads if last seems not to be executed
  int turnLatency;
  /// improvement threshold to choose direct kick over indirect
  double directThreshold;
  /// position error to actually lay back
  double laybackErr;
  /** the inner circle we do not want to get in.
    Default: PLAYER_SIZE+BALL_SIZE + 0.3*KICKABLE_MARGIN */
  double innerSafeCircle;
  /// the distance where the ball is handled optimally
  double safeCircle;
  /// the velocity which the ball may get maximally so that it can be handled
  double stopableV;
  /** the tolerated body direction angle error in dribble.
    Default: 10 */
  double dribbleBodyErr;
  /** the tolerated ball direction angle error in dribble.
    Default: 30 */
  double dribbleHoldErr;
  /// the maximal player velocity for dribble
  double dribbleMaxV;
  /// maximal number of steps simulated by control ball.
  int    controlMaxSteps;
  //@}

  /***************************** the constructor ****************************/
  /** the constructor.
    @param comm used for sending the calculated commands.
    */
  AdvSkills(Comm *comm);
  
  /********************************** the skills ****************************/

  /// kicking commands
  //@{
  /// compatability - translate modes into time lengths
  enum KickMode { KICKAWAY = 1, FIXED = 2, SUPERKICK = 5 };

  /**
    \begin{itemize}
    \item[KRANGE]     ball out of kick range
    \item[KFINISHING] ball will have wanted velocity after this kick
    \item[KTOOSHORT]  ball will be slower than wanted
    \item[KPLACING]   ball is placed to allow a better kick
    \item[KPLACEFORSKICK] ball is layed back
    \item[KSTOPPING]  requested velocity is near 0, so stopping ball
    \end{itemize}
   */
  enum KickState { KRANGE, KFINISHING, KTOOSHORT, KPLACING,
		   KPLACEFORSKICK, KSTOPPING };
  ///
  struct KickCommand {
    /// calculated ball velocity
    Vector2d  v;
    ///
    KickState state;
  };

  /// returns if a kick command is likely to be successful
  bool kickSafe(const Snapshot *snap);

  struct ExtKickCommand: public KickCommand {
    /// how many steps it will likely cost to complete the kick
    int time;
    ///
    int max_time;
    ///
    Vector2d target;
    /// if >= 0, the target velocity, indicates position kick
    double target_v;
  };

  /** try to give the ball a target velocity. Collisions are avoided by kicking
      the ball around the player if necessary. It is possible to let kick try
      to place the ball behind the player for a really hard kick (superkick).
    
      It is possible, that you have to call kick more than once. So call
      {\em rekick} to continue the last kick until it says RANGE,
      so that as long as possible kicks are sent.

      @param current_snap  the snapshot the agent believes to be current
      @param target_v      the velocity the ball should get
      @param max_time      maximal time until kick must be finished
      */
  ExtKickCommand kick(const Snapshot *current_snap, const Vector2d &target_v,
		      int max_time = 2);

  /** kick to an absolute target position.
      @param current_snap  the snapshot the agent believes to be current
      @param target_pos    the absolute coordinates where the ball should
                           be kicked.
      @param target_v      the velocity the ball should have when reaching
                           the target position
      */
  ExtKickCommand kick(const Snapshot *current_snap, double target_v,
		      const Vector2d &target_pos, int max_time = 2);

  /** does the last kick command again.
      @param current_snap  the snapshot the agent believes to be current
      @param kcm           the KickCommand from kick to continue
      */
  ExtKickCommand rekick(const Snapshot *current_snap,
			const ExtKickCommand &kcm);

  /** try to place ball on a safe circle around the player. The balls velocity
      will be slow enough to be stopped within one turn, if possible.
      @param current_snap  the snapshot the agent believes to be current
      @param target        the wanted position of the ball relative to the
                           player in absolute coordinates.
      */
  KickCommand placeBall(const Snapshot *current_snap, const Vector2d &target);

  /** try to stop ball.

      @param current_snap  the snapshot the agent believes to be current
      */
  KickCommand stopBall(const Snapshot *current_snap);

  /// execute part for kick commands
  void exec(const Snapshot *current_snap, const KickCommand &);
  //@}

  /// intercepting skills
  //@{
  /**
    \begin{itemize}
    \item[TFINISHED]       at target position/ball
    \item[TONWAY]          player is on way, but not in target position
    \item[TGOINGBACKWARDS] player shows in opposite direction
    \end{itemize}
   */
  enum TurtleState { TFINISHED, TONWAY, TGOINGBACKWARDS };
  ///
  struct TurtleCommand {
    /// estimated time until finished
    int         time;
    /// calculated player velocity
    Vector2d    v;
    /// players new body dir
    Angle       bodyDir;
    /// the angle to turn to reach bodyDir (affected by inertia moment)
    double      turnMoment;
    ///
    TurtleState state;
    // command storage
    enum { DASH, TURN } command;
  };

  /** try to get the ball under control, that is, intercept it.
      @param current_snap  the snapshot the agent believes to be current
      @param stamina       the stamina to use at most this step
      @param meet          the position we will meet the ball
      */
  TurtleCommand controlBall(const Snapshot *current_snap, double stamina, double error,
			    WorldBall &meet);
  ///
  TurtleCommand controlBall(const Snapshot *current_snap, double stamina);

  /** go to a position. This is intended for long range intercepts and absolute
      positions, so error is assumed to be relative large ( > PLAYER_SIZE).
      TIME INFORMATION OF TURTLECOMMAND IS NOT SET!!!

      @param current_snap  the snapshot the agent believes to be current
      @param target        absolute target position on field
      @param error         radius around the target in which the player
                           should end up
      @param stamina       the stamina to use at most this step
      @param dashBackwardsIfNeeded  keep the difference between bodyDir and
                           prefAngle < 90 by eventually dashing backwards
      @param prefAngle     the angle to keep if minAngle is set
      */
  TurtleCommand goToPosition(const Snapshot *current_snap, const Vector2d &target,
			     double error = 5.0,
			     double stamina = STAMINA_INC_MAX,
			     bool   dashBackwardsIfNeeded = false,
			     Angle  prefAngle = 0);

  /** go to a position.
      @param current_snap  the snapshot the agent believes to be current
      @param target        absolute target position on field
      @param stamina       the stamina to use for the complete intercept additional to
                           the stamina gained during the intercept.
      @param error         radius around the target in which the player
                           should end up
   */
  TurtleCommand goToPositionTimed(const Snapshot *current_snap, const Vector2d &target,
				  double stamina, double error);
  ///
  void exec(const Snapshot *current_snap, const TurtleCommand &);
  //@}

  /// dribbling
  //@{
  /**
    \begin{itemize}
    \item[DOK]           simply dribbling
    \item[DGOINGFORBALL] ball lost, actually intercepting it
    \end{itemize}
   */
  enum DribbleState { DOK, DGOINGFORBALL };
  ///
  struct DribbleCommand {
    /// calculated player and ball velocity
    Vector2d     v_p, v_b;
    ///
    Angle        bodyDir;
    ///
    DribbleState state;
    /// command storage
    enum { DASH, TURN, KICK } command;
  };

  /** dribble into an absolute direction. The agent can decide in which
      direction to the player the ball should be kept (in absolute direction).

      @param current_snap  the snapshot the agent believes to be current
      @param dribbledir    the direction in which to dribble
      @param hold_dir      the direction we want to hold the ball from us 
      */
  DribbleCommand dribble(const Snapshot *current_snap, Angle dribbledir,
			 Angle hold_dir, double stamina = STAMINA_INC_MAX);
  ///
  void exec(const Snapshot *current_snap, const DribbleCommand &);
  //@}

  /// goalie's catch skill
  //@{
  /**
    \begin{itemize}
    \item[CBAN]       catch still banned
    \item[CRANGE]     ball to far away or outside penalty area
    \item[CNOTGOALIE] player is not goalie and may not catch
    \item[COK]        done
    \end{itemize}
    */
  enum CatchState { CBAN, CRANGE, CNOTGOALIE, COK };

  /// returns if a catch command is likely to be successful
  bool catchSafe(const Snapshot *current_snap);

  ///
  struct CatchCommand {
    ///
    Angle catchDir;
    ///
    CatchState state;
  };

  /** catch the ball, if allowed. This is checked via worldmodel.
      @param current_snap the snapshot the agent believes to be current
      */
  CatchCommand goalieCatch(const Snapshot *current_snap);
  ///
  void exec(const Snapshot *current_snap, const CatchCommand &);
  //@}

  /// turn
  //@{
  /**
    \begin{itemize}
    \item[ROK]       done
    \item[RANGLE]    angle is too large, truncated
    \item[RFINISHED] angle to turn is small
    \end{itemize}
   */
  enum TurnState{ ROK, RANGLE, RFINISHED };
  ///
  struct TurnCommand {
    /// the optimal body dir actually achievable
    Angle bodyDir;
    /// the angle to turn to reach bodyDir (affected by inertia moment)
    double turnMoment;
    ///
    TurnState state;
  };
  /** turn into absolute direction dir.
      @param current_snap the snapshot the agent believes to be current
      */
  TurnCommand turn(const Snapshot *current_snap, Angle dir);
  ///
  void exec(const Snapshot *current_snap, const TurnCommand &);
  //@}

  /// turn neck
  //@{
  /**
    \begin{itemize}
    \item[THOK]       done
    \item[THANGLE]    angle is too large, truncated
    \item[THBAN]      last turn_neck seems not to be committed, waiting one turn
    \item[THFINISHED] angle to turn small
    \end{itemize}
  */
  enum TurnHeadState{ THOK, THANGLE, THBAN, THFINISHED };
  ///
  struct TurnHeadCommand {
    /// the optimal face dir actually achievable
    Angle faceDir;
    double turnHeadMoment;
    ///
    TurnHeadState state;
  };
  /** turns head into specified absolute direction.
      The resulting head angle is rounded to integer multiples of
      angleStepping to avoid too many turnHeads.

      @param current_snap the snapshot the agent believes to be current
      @param dir          the absolute dir we want to see
      */
  TurnHeadCommand turnHead(const Snapshot *current_snap, Angle dir);
  /** turn to ball considering a TurnCommand
   */
  TurnHeadCommand turnHeadToBall(const Snapshot *current_snap, const TurnCommand &tcm);
  /** turn to ball considering a TurtleCommand
   */
  TurnHeadCommand turnHeadToBall(const Snapshot *current_snap, const TurtleCommand &tcm);
  /** turn to ball considering a DribbleCommand
      @param dribbledir should be the same as for the dribble command
   */
  TurnHeadCommand turnHeadForDribble(const Snapshot *current_snap, const DribbleCommand &dcm,
				     double dribbledir);
  ///
  void exec(const Snapshot *current_snap, const TurnHeadCommand &);
  //@}

  /** changes view only when needed.
      @param current_snap the snapshot the agent believes to be current
      @param width        the wanted view width
   */
  void changeView(const Snapshot *current_snap, view_width width);

  /** changes view only when needed.
      @param current_snap the snapshot the agent believes to be current
      @param quality      the wanted view quality
   */
  void changeView(const Snapshot *current_snap, view_quality quality);

  /** changes view only when needed.
      @param current_snap the snapshot the agent believes to be current
      @param width        the wanted view width
      @param quality      the wanted view quality
   */
  void changeView(const Snapshot *current_snap, view_width width, view_quality quality);

  /** @name low-level commands.
      This are all commands that are directly aviable from comm.C
   */
  //@{
  ///
  void noAction();

  ///
  void basicDash (double pow);
  ///
  void basicKick (double pow, double dir);
  ///
  void basicTurn (double pow);
  ///
  void basicCatch(double dir);
  ///
  void basicMove(double x, double y);

  ///
  void sendTurnNeck(double dir);
  ///
  void sendChangeView(view_width width, view_quality quality);
  ///
  void sendSenseBody();
  //@}

private:

  Comm *comm;

  // turn head latency stuff
  int turnH_time;
  Angle turnH_oldHead, turnH_reqHead;

  // calculate current max and min accelerations for dashing respecting
  // reachability
  double reachRadius(double stamina, int numDashes);  
  double goToThresh(double error, double tdist);

  // vel constraints
  void calcAcc(const Snapshot *sn, double stamina, double &minacc, double &maxacc);
  // internal placeBall with more options
  KickCommand iplaceBall(const Snapshot *current_snap, const Vector2d &target,
			 int cond_mask, const int *drop_order);
};

#endif
