//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//-*- c++ -*-/////////////////////////////////////////////////////////
//
//  $Id: comm.h,v 1.13 2000/10/19 15:27:44 goetz Exp $
//  $Manager: eri$
// 
// 
//  comm.h - Communication with soccer server
// 
//  original version: Christian Bauer <cebix>, Achim Liese <wiz>
//
//  rebuilt for MRB99, 
//  tuned and cleaned up for MRB2000 by  Erich Kutschinski <eri>
//
//////////////////////////////////////////////////////////////////////


#ifndef _COMM_H_
#define _COMM_H_


#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <signal.h>

#include "constants.h"
#include "sensor.h"
 
#include "parser.h"
#include "datastore.h"
#include "asyncinfo.h"
#include "audiomgr.h"
#include "Ensure.h"

///////////////////////////////////////////////////////////////////////////

// error types returned by this module

enum comm_error {
  COMM_OK = 0,
  COMM_ERROR,	           	// Unspecified error
  COMM_ERROR_SERVER_NOT_FOUND,  // Server name not found
  COMM_ERROR_SOCKET_INIT,	// Couldn't init socket
  COMM_ERROR_SERVER_INIT,	// Server didn't accept our init command
  COMM_ERROR_NO_DATA,	        // Server has no new data
  COMM_ERROR_PARSE,	        // Server sent us something we could not parse
  COMM_ERROR_SIGNAL_INIT,       // initing the signal handlers failed
};


///////////////////////////////////////////////////////////////////////////


class Comm {

 public:

  Comm (const char* team);    // name of own team

  ~Comm();

  // interface to Player and Coach 
  comm_error 
  init(client_type type,          // type of sserver client (see above)
       const char* hostname,      // name of sserver host
       int port,                  // port on sserver host to connect to
       int reconnectNumber = 0,   // number to reconnect with (<0:no reconnect)
       bool printCommands    = false,     // print all commands sent to cout ?
       playing_side coachSide=SIDE_LEFT); // side of coach, has to be specified
  
  void         handleSignals();
  
  bool         isServerAlive();
  bool         isGameOver();
  int          getServerTime();
  bool         isTimeRunning();
  bool         isActionNeeded();

  int          getMyNumber();
  bool         amIGoalie();
  const char*  getMyTeam();
  playing_side getMySide();

  play_mode    getPlayMode();
  int          getOurScore();
  int          getOtherScore();
  bool         inSecondHalf();
  
  void         say(AudioMsg &msg);
  bool         getNextMsg(AudioMsg &msg);

  // action interface
  void noAction    ();
  void sendKick    (double power, double dir_rad);
  void sendDash    (double power);
  void sendTurn    (double moment_rad);
  void sendCatch   (double dir_rad);
  void sendMove    (double x, double y);

  void sendTurnNeck   (double angle_rad);
  void sendChangeView (view_width width, view_quality quality);
  void sendSenseBody  ();
  void sendSay        (const char *message);

  void sendChangeMode (play_mode playMode);
  void sendMoveObject (const char *object, double x, double y);
  void sendMoveObject (const char *object, 
		       double x, double y, double dir_rad);
  void sendMoveObject (const char *object, 
		       double x, double y, double dir_rad, 
		       double xvel, double yvel);
  void sendEye        (bool eyeOn);
  void sendEar        (bool earOn);
  void sendLook       ();  
  void sendCheckBall  ();
  void sendRecover    ();



  // interface to other modules (worldmodel, replayer)
  bool getNextData(SensorData &data);
  void getLastAction(CycleAction &action);


private:
  
  Parser          *parser;
  SensorDataStore *store;
  AsyncCommInfo   *info;
  AudioMgr        *audiomgr;
  
  bool hadMessage;
  int  noMessageCount;
  bool printCommands;
  
  // technical stuff
  
  int          sd;
  struct       sockaddr_in server_addr;
  bool         serverOk;

  sigset_t     alrmask;
  sigset_t     iomask;
  sigset_t     bothmask;
  sig_atomic_t alarmFlag;  
  sig_atomic_t ioFlag;

 
  int        getTimeExpression();
  comm_error createSocket (const char *host, const int port);
  comm_error sendToServer(char *msg);
  comm_error receiveFromServer(char *msg, long timeout_msec);
  void       generateCheckinMessage(char* outmsg, 
				    client_type type, int reconnectNumber);
  comm_error parseCheckinResponse(char* inmsg, client_type type, 
				  int reconnectNumber, playing_side coachSide);
  comm_error checkinAtServer (client_type type, 
			      int reconnectNumber, playing_side coachSide);
  comm_error blockSignals();
  comm_error unblockSignals();
  comm_error initSigHandlers();
  comm_error setAlarmInterval(int msec);
  void       removeSigHandlers();
  
  static void alarmHandler(int signo); // the static alarm handlers
  static void ioHandler   (int signo);  
  static Comm *theComm;
  // as the system signal can only call static signal handler methods, we
  // need a static handle to the single Comm-object, so we can call the 
  // actual signal handling methods:

  void alarmAction();    // executed on SIGALRM via alarmHandler()
  void ioAction();       // executed on SIGIO via ioHandler()

  // executed in main process, if using FLAGGING; else executed in sighandlers 
  void finishAlarm();    
  void finishIo();       


};

#endif
