//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "coordinates.h"

/************************************ common geometry ***********************/
Intersect
intersectLineCircle(const Vector2d &t, const Vector2d &s,
		    double d, double &l1, double &l2)
  /* |t| should be >> 0, the problem is instable for t->0.
     result: l1 and l2 are the borders of the stretching interval of t
             to fit into B_d(s). l1 <= l2.
     returns: Instable if t ~ 0, None if there is no intersection,
              Valid else
   */
{
  double nrmtq = t*t;
  if (nrmtq < 1e-5)
    return Instable;
  double mqhalf = (t*s) / nrmtq,
    discriminant = mqhalf*mqhalf - (s*s - d*d)/nrmtq;
  if (discriminant < 0)
    return None;

  // there are zero(s)
  discriminant = sqrt(discriminant);
  l1 = mqhalf - discriminant;
  l2 = mqhalf + discriminant;
  return Valid;
}

Intersect
intersectCircleCircle(double r1, const Vector2d &c2, double r2,
		      Vector2d &l1, Vector2d &l2)
{
  /* calculate the two intersecting points of two circles
     of radius r1 located at (0,0) and r2 located at c2.
     returns Valid if there are two intersection points, which are found as
     solavg +/- soldev, Instable for t ~ 0 and R1larger if the first Circle
     contains the second, R2larger if the second contains the first.
     None if there is no intersection.
     */
  
  double nrmc2q = c2*c2;
  if (nrmc2q < 1e-5) {
    if (r1 > r2)
      return R1larger;
    else
      return R2larger;
  }
  //cout << "c = " << norm(c2) << " " << r1 << " " << r2 << endl;
  Vector2d soldev = c2, solavg = c2;
  soldev.perp();
  double l = (nrmc2q + sqr(r1) - sqr(r2))/2.0/nrmc2q;
  //cout << "l = " << l << endl;
  double h = sqr(r1) - sqr(l)*nrmc2q;
  //cout << "h = " << h << endl;
  if (h < 0) {
    if (nrmc2q > sqr(r1 + r2))
      return None;
    else if (r1 > r2)
      return R1larger;
    else
      return R2larger;
  }
  h = sqrt(h);

  // stretch to represent the two solutions
  solavg.stretch(l);
  soldev.stretch(h/sqrt(nrmc2q));
  l1 = solavg + soldev;
  l2 = solavg - soldev;
  return Valid;
}

Intersect intersectLineLine( const Vector2d &l1dir,
			     const Vector2d &l2offset,
			     const Vector2d &l2dir,
			     double &l1, double &l2)
/*
  calculate the intersection of line1 starting at (0,0) and line2 starting
  at line2offset. Directions are line1/2dir. l1*line1dir = l2*line2dir +
  line2offset both give the intersection point if exactly one exists.
  Instable is returned if l1dir and l2dir are nearly parallel, else Valid.
 */
{
  // l1*l1dir = l2*l2dir + l2offset =>
  // l1*l2dperp*l1dir = l2dperp*l2offset

  Vector2d l2dperp = l2dir;
  l2dperp.perp();
  double detM = l2dperp*l1dir;

  if (fabs(detM) < 1e-5)
    return Instable;

  l1 = (l2dperp*l2offset)/detM;
  // 0=l2*l1dperp*l2dir + l1dperp*l2offset

  Vector2d l1dperp = l1dir;
  l1dperp.perp();
  l2 = -(l1dperp*l2offset)/(l1dperp*l2dir);
  return Instable;
}
