//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////// 
//   $Id: snapshot.h,v 1.19 2000/10/19 15:27:48 goetz Exp $
//   $Manager: cm$
//
//   snapshot.h
//
////////////////////////////////////////////////////////////////////////

#ifndef _SNAPSHOT_H_
#define _SNAPSHOT_H_

#include "sensor.h"         // sensor data types
#include "coordinates.h"    // angle class and vector class for 2-dimensional space
#include "constants.h"      // constants used by soccerserver
#include "minmax.h"         // min, max, sqr

// constants needed by worldmodel classes

const int UNKNOWN = -1;          // this is used for unknown player numbers
const int ETERNAL = 10000;       // age of objects we have never seen (->invalid objects)
                                 // the value should be larger than match length
const int MAXAGE  = 20;          // objects older than this are (by default) not
                                 // considered in the sorting methods

// Reports if there could have been a player<->ball collision when
// simulating a snapshot.

enum CollisionState
{
  NONE,        // There was no collision
  UNCERTAIN,   // It is not even clear if there was a collision
  CRITICAL,    // There MUST have been a collision (or two), at least according to our
               // simulated data, but simulation of this collision was impossible
               // (because we can not decide which object collides first)
  PROCESSED    // There WAS a collision AND it could be simulated
};


// State of a WorldObject (interesting for world model updates)

enum WorldObjectState
{
  UNSAFE,   // State might be undefined or unsafe; e.g. Object is a player and is 
            // simulated from visual data, but the number might be wrong
  FUZZY,    // Object is simulated from data that fitted to some older visual data;
            // e.g. object is a player without visible number but fitted to some
            // older visual information
  VISUAL,   // Object is simulated from visual data
  AUDIAL,   // Object is simulated from audial data
};


//////////////////////////////////////////////////////////////////////
//
// Class WorldObject
//
//////////////////////////////////////////////////////////////////////

class WorldObject {
public:
  WorldObject( const Vector2d& _pos = Vector2d() )
    : pos( _pos ), age( ETERNAL ), state( UNSAFE ) {}

  Vector2d  pos;   // absolute position on the field
  int       age;   // number of simulation steps that have passed since this object
                   // has been updated for the last time

  WorldObjectState state;   // tells which data type object was simulated from

  // calculate distance to a given object "obj"
  double    distTo( const WorldObject& obj ) const { return dist( pos, obj.pos ); }

  // check if object is in corridor as sketched below:
  //              ___________________________________ __         _
  //             /                                   |  \         |
  //            /                                    |   \        |
  // source -> /_ _ _ _ _ _ _ _ _ _ _ _ _ target -> _|_   |        \ d
  //           \                                     |    |        /
  //            \                                    |   /        |
  //             \___________________________________|__/        _|
  //
  // On the right there is a half circle with diameter d
  // (which is difficult to draw with ASCII characters...).
  // The angle of the cone on the left is 2 * alpha.
  // 
  // CAUTION: it is also checked if object is in a very small circle around source!
  bool      inCorridor( Vector2d source, Vector2d target, double d, Angle alpha = PI_2 ) const;
  bool      inOurPenaltyArea() const;
  bool      inOtherPenaltyArea() const;
};


//////////////////////////////////////////////////////////////////////
//
// Class WorldMovedObject
//
//////////////////////////////////////////////////////////////////////

class WorldPlayer;

class WorldMovedObject: public WorldObject {
public:
  WorldMovedObject( double _decay = 0.0,
		    double _vmax = 0.0,
		    double _size = 0.0,
		    const Vector2d& _pos = Vector2d( 0.0, 0.0 ),
		    const Vector2d& _v = Vector2d( 0.0, 0.0 ) )
    : WorldObject( _pos ), v( _v ), decay( _decay ), vmax( _vmax ), size( _size ), coneTime( 0 ) {}

  // accelerate object, but consider maximum velocity
  void      acc( const Vector2d& _a )
  { v += _a; if( v.getLength() > vmax ) v *= vmax / v.getLength(); }

  // simulate one moving step
  void      simulate() { pos += v; v *= decay; }

  // calculate object's absolute velocity if distChange and dirChange relative to 
  // relObject (usually player himself) are given
  void      calcVelocity( const WorldMovedObject& relObject, double distChange, double dirChange );

  // check if distance is less than the sum of the two sizes
  bool      overlaps( const WorldMovedObject& secondObject )
  { return distTo( secondObject ) < size + secondObject.size ? true : false; }

  // collide with "collObject" (as soccerserver does)
  void      collideWith( const WorldMovedObject& collObject );

  // update object position with sensor data (sensor data = relative position to "my")
  void      setPos( const SensorObject& object, const WorldPlayer& my );

  Vector2d  v;          // velocity vector
  double    decay;      // um... that's how we brake (decay factor)
  double    vmax;       // maximum LENGTH of velocity vector
  double    size;       // object radius
  int       coneTime;   // time object spent in a certain view cone
  // (does not make sense without connection to a player)
};


//////////////////////////////////////////////////////////////////////
//
// Class WorldBall
//
//////////////////////////////////////////////////////////////////////

class WorldBall : public WorldMovedObject {
public:
  WorldBall( const Vector2d& _pos = Vector2d( 0.0, 0.0 ),
	     const Vector2d& _v = Vector2d( 0.0, 0.0 ) )
    : WorldMovedObject( BALL_DECAY, BALL_SPEED_MAX, BALL_SIZE, _pos, _v ) {}

  // calculate ball position with sensor data (sensor data = relative position to a "player")
  void calcPos( const SensorObject& object, const WorldPlayer& my );

};


//////////////////////////////////////////////////////////////////////
//
// Class WorldPlayer
//
//////////////////////////////////////////////////////////////////////

class WorldPlayer : public WorldMovedObject {
public:
  WorldPlayer( const Vector2d& _pos = Vector2d( 0.0, 0.0 ),
	       const Vector2d& _v = Vector2d( 0.0, 0.0 ),
	       const Angle& _bodyDir = 0.0,
	       const Angle& _faceDir = 0.0,
	       int _number = UNKNOWN,
	       goalie_player _isGoalie = GOALIE_UNKNOWN )
    : WorldMovedObject( PLAYER_DECAY, PLAYER_SPEED_MAX, PLAYER_SIZE, _pos, _v ),
      bodyDir( _bodyDir ), faceDir( _faceDir ), number( _number ), isGoalie( _isGoalie ) {}

  // calculate angle between player's BODY direction and "obj"
  Angle      bodyAngleTo( const WorldObject& obj ) const
  { return ( obj.pos - pos ).getAngle() - bodyDir; }

  // calculate angle between player's FACE direction and "obj"
  Angle      faceAngleTo( const WorldObject& obj ) const
  { return ( obj.pos - pos ).getAngle() - faceDir; }

  // check if ball is in kickable area
  bool       ballKickable( const WorldBall& ball ) const;

  // check if ball is in catchable area (for a catch command with catchAngle)
  bool       ballCatchable( const WorldBall& ball, const Angle& catchAngle ) const;

  // calculate actual kick power from angle and distance to ball
  double     kickPowerRate( const WorldBall& ball ) const;

  // calculate inertia moment for turn command
  double     inertiaMoment() const { return v.getLength() * INERTIA_MOMENT + 1; }

  Angle      bodyDir;   // player's body-and-running direction
  Angle      faceDir;   // player's facing direction ( = bodyDir + headAngle )

  int        number;    // player's number ( -1 = undefined, 0 = coach )
  goalie_player isGoalie;

  // update player with sensor data ( sensor data = relative position to "my")
  void       setBySensorData( SensorPlayer player, const WorldPlayer& my );
};


//////////////////////////////////////////////////////////////////////
//
// Class WorldSense
//
//////////////////////////////////////////////////////////////////////

class WorldSense {
public:
  WorldSense() 
    : age( ETERNAL ), stamina( 0.0 ), effort( 1.0 ), v( 0.0, 0.0 ), headAngle( 0.0 ),
      kickCount( 0 ), dashCount( 0 ), turnCount( 0 ), sayCount( 0 ), turnNeckCount( 0 ),
      viewWidth( VIEW_WIDTH_NORMAL ), viewQuality( VIEW_QUALITY_HIGH ), recovery( 1.0 ) {}
  
  int          age;            // time that has passed since this information has been generated

  // all values soccerserver provides

  double       stamina;        // player's stamina
  double       effort;         // player's effort (which shows how "tired" player is)

  Vector2d     v;              // player's velocity
  Angle        headAngle;      // angle of player's head relative to the body

  int          kickCount;      // numbers of commands server has received
  int          dashCount;
  int          turnCount;
  int          sayCount;
  int          turnNeckCount;

  view_width   viewWidth;      // angle of view cone
  view_quality viewQuality;    // quality of visual information

  Angle        getViewAngle() const; // calculate view angle (= HALF of view cone!) from viewWidth

  double       recovery;       // recovery value for calculation of player's stamina. This value
                               // is not reported by soccerserver, but it makes sense to store
                               // it together with effort and stamina

  // calculates length of acceleration vector that a dash command with "power" would cause
  // (depends on stamina, effort, dashing backward/forward and multiplication with DASH_POWER_RATE)
  double       accRad( double power );

  // decrease stamina by "power" (which means the power for a dash command)
  void         decStamina( double power );

  void         simStamina();   // simulate stamina recovery (with effort and recovery)
};


//////////////////////////////////////////////////////////////////////
//
// Class Snapshot
//
//////////////////////////////////////////////////////////////////////

class Snapshot {
public:
  Snapshot();

  WorldPlayer      my;               // myself

  // get player with number "n" of our team
  const WorldPlayer      &ourTeam( int n ) const
  {
    if( n < 1 || n > 11 )
      cerr << "ERROR: ourTeam() call with parameter " << n
	   << ", my.number = " << my.number << endl;
    ensure( n >= 1 && n <= 11 );
    return ourTeamArray[n-1];
  }

  // get player with number "n" of other team
  const WorldPlayer      &otherTeam( int n ) const
  {
    if( n < 1 || n > 11 )
      cerr << "ERROR: otherTeam() call with parameter " << n
	   << ", my.number = " << my.number << endl;
    ensure( n >= 1 && n <= 11 ); return otherTeamArray[n-1];
  }

  // set player with number "n" of our team to "player"
  // CAUTION: do not use this to change snapshots in the worldmodel!
  void setOurTeam( int n, const WorldPlayer& player )
  { ensure( n >= 1 && n <= 11 ); ourTeamArray[n-1] = player; }

  // set player with number "n" of other team to "player"
  // CAUTION: do not use this to change snapshots in the worldmodel!
  void setOtherTeam( int n, const WorldPlayer& player )
  { ensure( n >= 1 && n <= 11 ); otherTeamArray[n-1] = player; }

  int              ourGoalieNo;      // number of our goalie (-1 = unknown)
  int              otherGoalieNo;    // number of opponents' goalie

  WorldBall        ball;
  
  CollisionState   collState;        // reports player<->ball collisions (after simulation)

  WorldObject      ourGoal;
  WorldObject      otherGoal;
  
  WorldSense       senseBody;        // information given by SenseBody (e.g. stamina)

  int              time;             // simulation time that this snapshot belongs to 
  int              lastCatchTime;    // time when last catch command has been sent

  playing_side     ourSide;
  play_mode        playMode;

  int              ourScore;         // number of goals we scored
  int              otherScore;       // number of goals scored by opponent team

  // the following 3 methods are used to set information invalid, i.e.
  // set its age to ETERNAL

  void    setInvalid();                 // visual and sense information
  void    setSeeInvalid();              // only visual information
  void    setSenseInvalid();            // only sense information

  // Get the number of that player of our resp. other team whose rank is "index" if the
  // players are sorted by distance to "target". Default for "index" is 1 (closest player to
  // "target"). Consider only players younger than "maxAge". Considering players we have never
  // seen before (which for example could mean that their pos is still marked as zero!) can be
  // avoided by setting maxAge < ETERNAL.
  // If no suitable player is found, UNKNOWN is returned. (!)
  // CAUTION: player himself might be returned by ourTeamClosestTo()!

  int     ourTeamClosestTo( const Vector2d& target, int maxAge = MAXAGE, int index = 1 ) const;
  int     otherTeamClosestTo( const Vector2d& target, int maxAge = MAXAGE, int index = 1 ) const;

  // Get the rank of player no. "number" if our resp. other team is sorted by distance to "target".
  // Only players with age <= maxAge are considered. If player with no. "number" is older than maxAge,
  // 0 is returned.

  int     ourTeamDistIndex( int number, const Vector2d& target, int maxAge = MAXAGE ) const;
  int     otherTeamDistIndex( int number, const Vector2d& target, int maxAge = MAXAGE ) const;
  
  // count players in our resp. other team that are in a certain corridor between "source" and
  // "target" (for details of the corridor see WorldObject::inCorridor())
  // Consider only players with age <= maxAge.
  // CAUTION: player himself will be in the corridor and thus counted by ourTeamInCorridor()!

  int     ourTeamInCorridor( const Vector2d& source, const Vector2d& target, double d,
			     const Angle& alpha = PI_2, int maxAge = MAXAGE ) const;
  int     otherTeamInCorridor( const Vector2d& source, const Vector2d& target, double d,
			       const Angle& alpha = PI_2, int maxAge = MAXAGE ) const;
  
  // calculate minimum x-position that is judged to be offside
  // Only opponent players with age <= maxAge are considered
  double  ourOffsideLine( int maxAge ) const;

  // calculate the opponent's offside line (of interest for our defenders)
  double  otherOffsideLine( int maxAge ) const;

  // write/read Snapshot as binary to/from stream
  bool    binaryWriteToStream( ostream &out ) const;
  bool    binaryReadFromStream( istream &in );

private:
  WorldPlayer         ourTeamArray[11];      // players of my team sorted by number
  WorldPlayer         otherTeamArray[11];    // players of other team sorted by number
  // CAUTION: access to player with number i is via our/otherTeamArray[i-1]

  bool    adjustSense( SensorData &data );   // fill in new sense data
  bool    adjustSee( SensorData &data );     // fill in new visual data
  bool    adjustExact( SensorData &data );   // fill in new exact data

  // Calculate my absolute position and face direction with new sensor data.
  // Face direction is calculated with a line, if only one line is seen
  // (this means we are inside the field), otherwise with flags.
  // Position is calculated with flags.
  // calcMyPos() is true if position could be calculated.
  bool    calcMyPos( SensorData& data );

  bool    adjustPlayers( SensorData &data ); // adjust other players to my (new) position
  bool    adjustBall( SensorData &data );    // same with ball
  bool    adjustGoals( SensorData &data );   // same with goals

  friend class Worldmodel;   // Worldmodel calls adjusting methods
};


class ostream& operator<< (class ostream &os, Snapshot& s);

class AudioMsg& operator<< (class AudioMsg& aum, const Snapshot& s);
class AudioMsg& operator>> (class AudioMsg& aum, Snapshot& s);

#endif
