//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////// 
//   $Id: worldmodel.C,v 1.39 2000/10/19 15:27:49 goetz Exp $	
//   $Manager: cm$
//
//   worldmodel.C
//
////////////////////////////////////////////////////////////////////////


#include "unistd.h"
#include "math.h"
#include "mytrace.h"
#include "worldmodel.h"
#include "Ensure.h"


// output generated when checking for collisions
#undef COLLISIONTRACE
#define COLLISIONTRACE(print_command)


// output generated when merging audial information into worldmodel
#undef AUDIALINPUTTRACE
#define AUDIALINPUTTRACE(print_command)


// output generated when spreading worldmodel to other players
#undef AUDIALOUTPUTTRACE
#define AUDIALOUTPUTTRACE(print_command)


//////////////////////////////////////////////////////////////////////
//
// Class Worldmodel
//
//////////////////////////////////////////////////////////////////////


Worldmodel::Worldmodel( int _numSnapshots, Comm* _comm )
{
  numSnapshots = _numSnapshots;
  comm = _comm;

  visualTime = -1;                    // no sensor information before kick off
  senseTime = -1;
  maxTime = -1;
  lastCatchTime = -CATCH_BAN_CYCLE;   // goalie is allowed to catch at once (i.e. in timestep 0)

  ensure( numSnapshots <= 30 );       // just to avoid out of memory (30 IS a lot!)
  
  snapshots = new Snapshot[numSnapshots];
  ensure( snapshots );
  
  actions = new WorldAction[numSnapshots];
  ensure( actions );

  for( int i = 0; i < numSnapshots; i++ )
    {
      snapshots[i].time = i - numSnapshots;   // set time into the past
      snapshots[0].time = 0;                  // ...but not for time step 0
      snapshots[i].lastCatchTime = -CATCH_BAN_CYCLE;
      snapshots[i].playMode = comm->getPlayMode();
      snapshots[i].ourSide = comm->getMySide();
      snapshots[i].my.number = comm->getMyNumber();
      snapshots[i].my.isGoalie = comm->amIGoalie() ? GOALIE_TRUE : GOALIE_FALSE;
      if( comm->amIGoalie() )
	{
	  snapshots[i].ourGoalieNo = comm->getMyNumber();
	  for( int j = 0; j < 11; j++ )
	  if( j != comm->getMyNumber() - 1 )
	    snapshots[i].ourTeamArray[j].isGoalie = GOALIE_FALSE;
	}
      snapshots[i].ourScore = comm->getOurScore();
      snapshots[i].otherScore = comm->getOtherScore();
      
      for( int j = 0; j < 11; j++)
	{
	  snapshots[i].ourTeamArray[j].number = j + 1;
	  snapshots[i].otherTeamArray[j].number = j + 1;
	}
    }
}


Worldmodel::Worldmodel( const Worldmodel& wm )
{
  numSnapshots = wm.numSnapshots;
  comm = wm.comm;

  visualTime = wm.visualTime;
  senseTime = wm.senseTime;
  maxTime = wm.maxTime;
  lastCatchTime = wm.lastCatchTime;

  snapshots = new Snapshot[numSnapshots];
  ensure( snapshots );

  actions = new WorldAction[numSnapshots];
  ensure( actions );
  
  for( int i = 0; i < numSnapshots; i++ )
    {
      snapshots[i] = wm.snapshots[i];
      actions[i] = wm.actions[i];
    }
}


Worldmodel& 
Worldmodel::operator = ( const Worldmodel& wm )
{
  numSnapshots = wm.numSnapshots;
  comm = wm.comm;

  visualTime = wm.visualTime;
  senseTime = wm.senseTime;
  maxTime = wm.maxTime;

  lastCatchTime = wm.lastCatchTime;

  delete [] snapshots;
  delete [] actions;

  snapshots = new Snapshot[numSnapshots];
  ensure( snapshots );

  actions = new WorldAction[numSnapshots];
  ensure( actions );

  for( int i = 0; i < numSnapshots; i++ )
    {
      snapshots[i] = wm.snapshots[i];
      actions[i] = wm.actions[i];
    }
  return *this;
}


Worldmodel::~Worldmodel() 
{
  delete [] snapshots;
  delete [] actions;
}


void
Worldmodel::getAction()
{
  CycleAction action;

  comm->getLastAction( action );

  int actionIndex = getTimeIndex( action.time );

  actions[actionIndex].time = action.time;
  actions[actionIndex].type = action.command;

  actions[actionIndex].moment = 0.0;    // cancel old data
  actions[actionIndex].power  = 0.0;
  actions[actionIndex].pos.setXY( 0.0, 0.0 );

  switch( action.command )
    {
    case NO_COMMAND:    // param1 = param2 = 0
      break;
      
    case TURN_COMMAND:  // param1 = moment_rad, param2 = 0
      actions[actionIndex].moment = action.param1;
      break;
      
    case DASH_COMMAND:  // param1 = power, param2 = 0
      actions[actionIndex].power = action.param1;
      break;
      
    case KICK_COMMAND:  // param1 = power, param2 = dir_rad
      actions[actionIndex].power  = action.param1;
      actions[actionIndex].moment = action.param2;
      break;
      
    case CATCH_COMMAND: // for goalie only; param1 = dir_rad, param2 = 0
      actions[actionIndex].moment = action.param1;
      break;
      
    case MOVE_COMMAND:  // only for coach or in certain playmodes; param1=x, param2=y
      actions[actionIndex].pos.setXY( action.param1, action.param2 );
      break;
      
    default:
      break;
    }
}


void
Worldmodel::simVisualData( int time )
{
  // simulate visual information time -> time+1

  int      i;
  double   turnMoment;

  WorldAction& action = actions[getTimeIndex( time )];
  
  Snapshot& oldSnap = snapshots[getTimeIndex( time )];
  Snapshot& newSnap = snapshots[getTimeIndex( time+1 )];
  
  // step 1 : copy old information

  newSnap.lastCatchTime = oldSnap.lastCatchTime;
  newSnap.my = oldSnap.my;                    // player himself
  newSnap.ball = oldSnap.ball;                // ball
  newSnap.otherGoal = oldSnap.otherGoal;      // goals
  newSnap.ourGoal = oldSnap.ourGoal;
  newSnap.ourScore = oldSnap.ourScore;
  newSnap.otherScore = oldSnap.otherScore;
  newSnap.ourGoalieNo = oldSnap.ourGoalieNo;
  newSnap.otherGoalieNo = oldSnap.otherGoalieNo;

  for( i = 0; i < 11; i++ )                   // players of both teams
    {
      newSnap.ourTeamArray[i] = oldSnap.ourTeamArray[i];
      newSnap.otherTeamArray[i] = oldSnap.otherTeamArray[i];
    }

  // step 2 : Birthday party for all objects

  if( newSnap.my.age < ETERNAL ) newSnap.my.age++;
  if( newSnap.ball.age < ETERNAL ) newSnap.ball.age++;
  if( newSnap.otherGoal.age < ETERNAL ) newSnap.otherGoal.age++;
  if( newSnap.ourGoal.age < ETERNAL ) newSnap.ourGoal.age++;
  for( i = 0; i < 11; i++ )
    {
      if( newSnap.ourTeamArray[i].age < ETERNAL ) newSnap.ourTeamArray[i].age++;
      if( newSnap.otherTeamArray[i].age < ETERNAL ) newSnap.otherTeamArray[i].age++;
    }
  
  // step 3 : if player has desired an action, simulate it now
  
  if( action.time == time )   // is action valid?
    {
      switch( action.type )
	{
	case DASH_COMMAND: 
	  // simulation of a dash command requires knowledge about the stamina value.
	  // That's why in getSnapshot() sense information is simulated first
	  newSnap.my.acc( radAng2Vector2d( oldSnap.senseBody.accRad( action.power ), newSnap.my.bodyDir ) );

	  // move ball if goalie dashes after a catch command
	  // caution: this does not check if goalie really sent a catch command!
	  if( newSnap.my.isGoalie == GOALIE_TRUE && ( newSnap.playMode == PLAY_MODE_FREE_KICK_US ) )
	    if( oldSnap.my.ballKickable( oldSnap.ball ) )
	      newSnap.ball.pos += newSnap.my.v;
	  break;
	  
	case TURN_COMMAND:
	  turnMoment = action.moment;
	  if( turnMoment > MAXMOMENT / 180.0 * M_PI ) turnMoment = MAXMOMENT / 180.0 * M_PI;
	  if( turnMoment < MINMOMENT / 180.0 * M_PI ) turnMoment = MINMOMENT / 180.0 * M_PI;
	  newSnap.my.bodyDir += turnMoment / newSnap.my.inertiaMoment();
	  newSnap.my.faceDir += turnMoment / newSnap.my.inertiaMoment();
	  break;
	  
	case MOVE_COMMAND:
	  // check if playMode allows a move command
	  switch( newSnap.playMode )
	    {
	    case PLAY_MODE_FREE_KICK_US:
	      if( newSnap.my.isGoalie != GOALIE_TRUE )
		break;
	      // caution: this does not check if goalie sent a catch command!
	      if( oldSnap.my.ballKickable( oldSnap.ball ) )
		newSnap.ball.pos = action.pos + oldSnap.ball.pos - oldSnap.my.pos;
	    case PLAY_MODE_BEFORE_KICK_OFF:
	    case PLAY_MODE_GOAL_US:
	    case PLAY_MODE_GOAL_THEM:
	      newSnap.my.pos = action.pos;
	      newSnap.my.v.setXY( 0.0, 0.0 );
	      break;
	    default:
	      break;
	    }
      
	case KICK_COMMAND:
	  // is the ball within kick radius?
	  if( newSnap.my.ballKickable( newSnap.ball ) )
	    {
	      newSnap.ball.acc( radAng2Vector2d( action.power * newSnap.my.kickPowerRate( newSnap.ball ),
						 newSnap.my.bodyDir + Angle( action.moment ) ) );	       
	    }
	  break;
	  
	case CATCH_COMMAND:
	  // check if player is goalie, if player is allowed to catch again and if ball is in catchable area
	  if( newSnap.my.isGoalie == GOALIE_TRUE ) 
	    if( action.time >= newSnap.lastCatchTime + CATCH_BAN_CYCLE )
	      {
		newSnap.lastCatchTime = action.time;
		if( action.time > lastCatchTime ) 
		  lastCatchTime = action.time;
		if( newSnap.my.ballCatchable( newSnap.ball, action.moment ) )
		  {
		    newSnap.ball.v.setXY( 0.0, 0.0 );
		    newSnap.ball.pos =
		      oldSnap.my.pos + oldSnap.my.v +
		      radAng2Vector2d( PLAYER_SIZE + BALL_SIZE,
				       ( newSnap.ball.pos - newSnap.my.pos ).getAngle() );
		  }
	      }
	  break;
	  
	default:
	  break;
	}
    }
  
  // step 4 : simulate movement of objects

  WorldBall oldBall = newSnap.ball;  // save ball before simulation for collision check
  WorldPlayer oldMy = newSnap.my;    // save player before simulation for collision check

  newSnap.my.simulate();             // simulate my movement  
  newSnap.ball.simulate();           // simulate ball movement
  for( i = 0; i < 11; i++ )          // simulate other players' movement
    {
      newSnap.ourTeamArray[i].simulate();
      newSnap.otherTeamArray[i].simulate();
    }

  // step 5 : check for collision with ball (we do NOT check player<->player collisions)

  newSnap.collState = NONE;

  WorldBall newBall = newSnap.ball;
  WorldPlayer newMy = newSnap.my;

  // check if there is a collision if player is simulated first

  if( newMy.overlaps( oldBall ) )
    {
      COLLISIONTRACE( cout << "If player is simulated first, there is a collision!" << endl; );
      newSnap.collState = UNCERTAIN;

      newMy.collideWith( oldBall );
      oldBall.simulate();
      if( oldBall.overlaps( newMy ) )   // maybe second collision?
	{
	  COLLISIONTRACE( cout << "...no, there are TWO collisions!" << endl; );
	  oldBall.collideWith( newMy );
	}
    }

  // check if there is a collision if ball is simulated first

  if( newBall.overlaps( oldMy ) )
    {
      COLLISIONTRACE( cout << "If ball is simulated first, there is a collision!" << endl; );
      if( newSnap.collState == UNCERTAIN )
	newSnap.collState = CRITICAL;
      else
	newSnap.collState = UNCERTAIN;

      newBall.collideWith( oldMy );
      oldMy.simulate();
      if( oldMy.overlaps( newBall ) )   // maybe second collision?
	{
	  COLLISIONTRACE( cout << "...no, there are TWO collisions!" << endl; );
	  oldMy.collideWith( newBall );
	}
    }

  // Now we hope that it doesn't matter which object is simulated first

  if( newSnap.collState == CRITICAL )
    {
      // Now, (newMy, oldBall) is what the world looks like after all
      // collisions if player is simulated first. (oldMy, newBall) is
      // the picture if ball is simulated first.

      // Compare relative positions of ball in both scenarios. If
      // these coincide, the collision can be simulated without critical errors.

      if( fabs( angleBetween( oldBall.pos - newMy.pos, newBall.pos - oldMy.pos ) < PI / 18.0 ) )   // 10 deg
	{
	  COLLISIONTRACE( cout << "Simulation can be done! (possible angle error: " <<
			  RAD_TO_DEG( fabs( angleBetween( oldBall.pos - newMy.pos, newBall.pos - oldMy.pos ) ) )
			  << ")" << endl; );

	  newSnap.collState = PROCESSED;
	  // Here we don't choose between the two different simulations,
	  // because we have no idea which one is better
	  newSnap.ball = oldBall;
	  newSnap.my = newMy;
	}
    }

  // if we received a sense body, use velocity vector and head angle
  // maybe the velocity vector could also be helpful to detect and handle collisions
  if( newSnap.senseBody.age == 0 && newSnap.time == time + 1 )
    {
      newSnap.my.v = newSnap.senseBody.v;
      // calculate faceDir with headAngle
      newSnap.my.faceDir = newSnap.my.bodyDir + newSnap.senseBody.headAngle;
      // Don't forget: v is relative...
      newSnap.my.v.turn( newSnap.my.faceDir );
    }

  // update coneTime for ball and players

  // check if ball is in my view cone or in my "feel radius"
  if( ( fabs( newSnap.my.faceAngleTo( newSnap.ball ) ) < newSnap.senseBody.getViewAngle() ) ||
      newSnap.my.distTo( newSnap.ball ) < VISIBLE_DISTANCE )
    newSnap.ball.coneTime++;

  // check if players are in my view cone or in my "feel radius"
  if( ( fabs( newSnap.my.faceAngleTo( newSnap.ourTeamArray[i] ) ) < newSnap.senseBody.getViewAngle() &&
	newSnap.my.distTo( newSnap.ourTeamArray[i] ) < TEAM_FAR_LENGTH ) ||
      newSnap.my.distTo( newSnap.ourTeamArray[i] ) < VISIBLE_DISTANCE )
    newSnap.ourTeamArray[i].coneTime++;

  if( ( fabs( newSnap.my.faceAngleTo( newSnap.otherTeamArray[i] ) ) < newSnap.senseBody.getViewAngle() &&
	newSnap.my.distTo( newSnap.otherTeamArray[i] ) < TEAM_FAR_LENGTH ) ||
      newSnap.my.distTo( newSnap.otherTeamArray[i] ) < VISIBLE_DISTANCE )
    newSnap.otherTeamArray[i].coneTime++;

  // copy player into team array (can be done faster)
  // don't try this for the coach!
  if( newSnap.my.number > 0 )
    newSnap.ourTeamArray[newSnap.my.number-1] = newSnap.my;

  // finally set time of new snapshot
  newSnap.time = time + 1;
}


void
Worldmodel::simSenseData( int time )
{
  // simulate sense information time -> time+1
  
  Snapshot& oldSnap = snapshots[getTimeIndex( time )];
  Snapshot& newSnap = snapshots[getTimeIndex( time+1 )];
  
  WorldAction& action = actions[getTimeIndex( time )];
  
  // step 1 : copy old information
  
  newSnap.senseBody = oldSnap.senseBody;
  newSnap.time = time + 1;
  
  // step 2 : Birthday party
  
  newSnap.senseBody.age++;
  
  // step 3 : decrease stamina if agent desired a dash command
  
  if( action.time == time )   // is action valid?
    if( action.type == DASH_COMMAND )
      newSnap.senseBody.decStamina( action.power );

  // step 4 : increase stamina (considering effort and recovery parameters)

  newSnap.senseBody.simStamina();
}


Snapshot *
Worldmodel::getSnapshot( int time )
{
  int    t;

  getAction();      // get latest Action sent by agent (we might need that for our simulation)

  int    minTime= min( visualTime, senseTime );

  // desired time too far back into the past
  if( time < maxTime - numSnapshots + 1 ) return NULL;
  
  // no sensor data received at all
  if( ( visualTime < 0 ) && ( senseTime < 0 ) ) return NULL;
  
  // desired time too far away from minTime. The largest time allowed depends
  // on the size of the "snapshots" array (i.e. numSnapshots)
  if( time - minTime > numSnapshots ) return NULL;
  
  // simulation of sense information (from senseTime to time)
  if( senseTime >= 0 )
    for( t = senseTime; t < time; t++ )
      simSenseData( t );
  else
    snapshots[getTimeIndex( time )].setSenseInvalid(); // no sense information available
  
  // simulation of visual information (from visualTime to time)
  if( visualTime >= 0 )
    for( t = visualTime; t < time; t++ )
      simVisualData( t );
  else
    snapshots[getTimeIndex( time )].setSeeInvalid();   // no visual information available

  if( time > maxTime ) maxTime = time;

  return &snapshots[getTimeIndex( time )];
}


bool
Worldmodel::adjust()
{
  SensorData   data;              // new data received from soccerserver
  bool         newData = false;   // value of adjust()

  // as long as there is data in the buffer - get it

  while ( comm->getNextData( data ) )
    {
      Snapshot &snap = snapshots[getTimeIndex( data.time )];

      switch( data.type ) 
	{
	case SENSOR_TYPE_SEE:      
	  
	  // only care about newer data
	  if( data.time >= visualTime )
	    {
	      // simulate world until data.time, then update with newer information
	      getSnapshot( data.time );
	      
	      snap.playMode = comm->getPlayMode();
	      snap.ourSide = comm->getMySide();
	      snap.ourScore = comm->getOurScore();
	      snap.otherScore = comm->getOtherScore();
	      
	      if( comm->amIGoalie() && snap.my.isGoalie != GOALIE_TRUE )
		{
		  snap.ourGoalieNo = comm->getMyNumber();
		  for( int j = 0; j < 11; j++ )
		    if( j != comm->getMyNumber() - 1 )
		      snap.ourTeamArray[j].isGoalie = GOALIE_FALSE;
		}
	      snap.my.isGoalie = comm->amIGoalie() ? GOALIE_TRUE : GOALIE_FALSE;
	      
	      snap.adjustSee( data );
	      
	      // try to keep up with headAngle changes
	      // this works because we always process senseBody messages first
	      snap.my.bodyDir = snap.my.faceDir - snap.senseBody.headAngle;
	      
	      // copy player into team array (can be done faster)
	      // don't try this for the coach!
	      if( snap.my.number > 0 )
		snap.ourTeamArray[snap.my.number-1] = snap.my;
  
	      snap.collState = NONE;   // just forget simulation - we have got visual data!
	      
	      visualTime = data.time;
	      newData = true;
	    }
	  break;
	  
	case SENSOR_TYPE_SENSE_BODY:
	  
	  // only care about new data
	  if ( data.time >= senseTime )
	    {
	      // simulate senseBody-informaton until data.time, then update with newer information
	      // simulation is needed for recovery parameter only (and to fill in missing snapshots)
	      getSnapshot( data.time );
	      
	      snap.playMode = comm->getPlayMode();
	      snap.ourSide = comm->getMySide();
	      snap.ourScore = comm->getOurScore();
	      snap.otherScore = comm->getOtherScore();
	      
	      if( comm->amIGoalie() && snap.my.isGoalie != GOALIE_TRUE )
		{
		  snap.ourGoalieNo = comm->getMyNumber();
		  for( int j = 0; j < 11; j++ )
		    if( j != comm->getMyNumber() - 1 )
		      snap.ourTeamArray[j].isGoalie = GOALIE_FALSE;
		}
	      snap.my.isGoalie = comm->amIGoalie() ? GOALIE_TRUE : GOALIE_FALSE;
	      
	      snap.adjustSense( data );

	      snap.my.faceDir = snap.my.bodyDir + Angle( data.head_angle );
	      
	      // update player velocity (caution: velocity is relative!)
	      snap.my.v = radAng2Vector2d( data.speed, Angle( data.speed_angle ) + snap.my.faceDir );
	      
	      // copy player into team array (can be done faster)
	      // don't try this for the coach!
	      if( snap.my.number > 0 )
		snap.ourTeamArray[snap.my.number-1] = snap.my;
	      
	      senseTime = data.time;
	      newData = true;
	    }
	  break;
	  
	case SENSOR_TYPE_EXACT:
	  
	  // only care about newer data
	  if( data.time >= visualTime )
	    {
	      //simulate world until data.time, then overwrite with exact information
	      getSnapshot( data.time );
	      
	      snap.playMode = comm->getPlayMode();
	      snap.ourSide = comm->getMySide();
	      snap.ourScore = comm->getOurScore();
	      snap.otherScore = comm->getOtherScore();
	      
	      snap.adjustExact( data );
	      
	      visualTime = data.time;
	      
	      // As the coach never receives sense_body, we set them invalid.
	      // But we have to set senseTime up, so snapshots will be 
	      // returned nevertheless.
	      senseTime = data.time;
	      snap.setSenseInvalid();
	      newData = true;
	    }
	  break;
	  
	default:
	  break;
	}
    }
  return newData;
}


bool
Worldmodel::feedbackAudio( AudioMsg &audioMsg )
{
  bool ok;
  audioMsg >> ok;
  
  if( ok )
    {
      Snapshot audioSnap;
      audioMsg >> audioSnap;

      AUDIALINPUTTRACE( cout << "Audial information from no. " << audioMsg.getSource() << " :" << endl; );

      ok |= updateWithSnapshot( audioSnap, audioMsg.getSource() );
    }

  return ok;
}


bool
Worldmodel::updateWithSnapshot( Snapshot audioSnap, int sender )
{
  int   timeDiff   = visualTime - audioSnap.time;
  bool  changed    = false;   // true if any data was updated
  bool  feltPlayer = false;   // did I "feel" players around me?

  if( timeDiff > 0 )
    {
      // simulate audio information until visualTime
      if( audioSnap.ball.age < ETERNAL )
	{
	  for( int i = 0; i < timeDiff; i++ )
	    audioSnap.ball.simulate();
	  audioSnap.ball.age += timeDiff;
	}
      for( int i = 0; i < 11; i++ )
	{
	  if( audioSnap.ourTeamArray[i].age < ETERNAL )
	    audioSnap.ourTeamArray[i].age += timeDiff;
	  if( audioSnap.otherTeamArray[i].age < ETERNAL )
	    audioSnap.otherTeamArray[i].age += timeDiff;
	}
    }

  // get the snapshot that is compared with audioSnap
  Snapshot& worldSnap = snapshots[getTimeIndex( max( visualTime, audioSnap.time ) )];

  // check for felt players
  for( int i = 0; i < 11; i++ )
    {
      if( i != worldSnap.my.number - 1 )
	if( dist( worldSnap.ourTeamArray[i].pos, worldSnap.my.pos ) < VISIBLE_DISTANCE )
	  {
	    feltPlayer = true;
	    break;
	  }
      if( dist( worldSnap.otherTeamArray[i].pos, worldSnap.my.pos ) < VISIBLE_DISTANCE )
	{
	  feltPlayer = true;
	  break;
	}		
    }

  AUDIALINPUTTRACE( cout << "Using friends : "; );
  
  // update players of my team
  for( int i = 0; i < 11; i++ )
    {
      // only update with valid information
      if( audioSnap.ourTeamArray[i].age >= ETERNAL )
	continue;

      // never update player himself
      if( i == worldSnap.my.number - 1 )
	continue;

      // use of audial information depends on its quality and age
      bool updatePlayer = false;
      bool checkAge = false;
      WorldObjectState state = UNSAFE;

      switch( worldSnap.ourTeamArray[i].state )
	{
	case UNSAFE :
	  switch( audioSnap.ourTeamArray[i].state )
	    {
	    case UNSAFE : updatePlayer = true;  checkAge = true;  state = UNSAFE; break;
	    case FUZZY  : updatePlayer = true;  checkAge = false; state = FUZZY;  break;
	    case VISUAL : updatePlayer = true;  checkAge = false; state = AUDIAL; break;
	    case AUDIAL : updatePlayer = true;  checkAge = false; state = AUDIAL; break;
	    }
	  break;

	case FUZZY :
	  switch( audioSnap.ourTeamArray[i].state )
	    {
	    case UNSAFE : updatePlayer = false; checkAge = false; state = UNSAFE; break;
	    case FUZZY  : updatePlayer = true;  checkAge = true;  state = FUZZY;  break;
	    case VISUAL : updatePlayer = true;  checkAge = false; state = AUDIAL; break;
	    case AUDIAL : updatePlayer = true;  checkAge = false; state = AUDIAL; break;
	    }
	  break;

	case VISUAL :
	  switch( audioSnap.ourTeamArray[i].state )
	    {
	    case UNSAFE : updatePlayer = false; checkAge = false; state = UNSAFE; break;
	    case FUZZY  : updatePlayer = true;  checkAge = true;  state = FUZZY;  break;
	    case VISUAL : updatePlayer = true;  checkAge = true;  state = AUDIAL; break;
	    case AUDIAL : updatePlayer = true;  checkAge = true;  state = AUDIAL; break;
	    }
	  break;

	case AUDIAL :
	  switch( audioSnap.ourTeamArray[i].state )
	    {
	    case UNSAFE : updatePlayer = false; checkAge = false; state = UNSAFE; break;
	    case FUZZY  : updatePlayer = true;  checkAge = true;  state = FUZZY;  break;
	    case VISUAL : updatePlayer = true;  checkAge = true;  state = AUDIAL; break;
	    case AUDIAL : updatePlayer = true;  checkAge = true;  state = AUDIAL; break;
	    }
	  break;
	}

      // if player is very close I should have felt him (player might be my shadow!)
      if( updatePlayer )
	if( dist( audioSnap.ourTeamArray[i].pos, worldSnap.my.pos ) < VISIBLE_DISTANCE )
	  if( !feltPlayer )
	    updatePlayer = false;

      // if I felt that player I have to be careful with updates
      if( updatePlayer )
	if( dist( worldSnap.ourTeamArray[i].pos, worldSnap.my.pos ) < VISIBLE_DISTANCE )
	  checkAge = true;

      // update player if age is not to be checked or he is young enough
      if( updatePlayer )
	if( ( checkAge && audioSnap.ourTeamArray[i].age < worldSnap.ourTeamArray[i].age ) ||
	    ( !checkAge && audioSnap.ourTeamArray[i].age - 5 < worldSnap.ourTeamArray[i].age ) )
	  {
	    worldSnap.ourTeamArray[i].pos = audioSnap.ourTeamArray[i].pos;
	    worldSnap.ourTeamArray[i].age = audioSnap.ourTeamArray[i].age;
	    worldSnap.ourTeamArray[i].state = state;
	    worldSnap.ourTeamArray[i].coneTime = 0;

	    changed = true;
	    AUDIALINPUTTRACE( cout << i + 1 << " "; );
	  }
    }
  
  AUDIALINPUTTRACE( cout << endl << "Using opponents : "; );

  // update players of opponent team
  for( int i = 0; i < 11; i++ )
    {
      // only update with valid information
      if( audioSnap.otherTeamArray[i].age >= ETERNAL )
	continue;

      // use of audial information depends on its quality and age
      bool updatePlayer = false;
      bool checkAge = false;
      WorldObjectState state = UNSAFE;

      switch( worldSnap.otherTeamArray[i].state )
	{
	case UNSAFE :
	  switch( audioSnap.otherTeamArray[i].state )
	    {
	    case UNSAFE : updatePlayer = true;  checkAge = true;  state = UNSAFE; break;
	    case FUZZY  : updatePlayer = true;  checkAge = false; state = FUZZY;  break;
	    case VISUAL : updatePlayer = true;  checkAge = false; state = AUDIAL; break;
	    case AUDIAL : updatePlayer = true;  checkAge = false; state = AUDIAL; break;
	    }
	  break;

	case FUZZY :
	  switch( audioSnap.otherTeamArray[i].state )
	    {
	    case UNSAFE : updatePlayer = false; checkAge = false; state = UNSAFE; break;
	    case FUZZY  : updatePlayer = true;  checkAge = true;  state = FUZZY;  break;
	    case VISUAL : updatePlayer = true;  checkAge = false; state = AUDIAL; break;
	    case AUDIAL : updatePlayer = true;  checkAge = false; state = AUDIAL; break;
	    }
	  break;

	case VISUAL :
	  switch( audioSnap.otherTeamArray[i].state )
	    {
	    case UNSAFE : updatePlayer = false; checkAge = false; state = UNSAFE; break;
	    case FUZZY  : updatePlayer = true;  checkAge = true;  state = FUZZY;  break;
	    case VISUAL : updatePlayer = true;  checkAge = true;  state = AUDIAL; break;
	    case AUDIAL : updatePlayer = true;  checkAge = true;  state = AUDIAL; break;
	    }
	  break;

	case AUDIAL :
	  switch( audioSnap.otherTeamArray[i].state )
	    {
	    case UNSAFE : updatePlayer = false; checkAge = false; state = UNSAFE; break;
	    case FUZZY  : updatePlayer = true;  checkAge = true;  state = FUZZY;  break;
	    case VISUAL : updatePlayer = true;  checkAge = true;  state = AUDIAL; break;
	    case AUDIAL : updatePlayer = true;  checkAge = true;  state = AUDIAL; break;
	    }
	  break;
	}

      // if player is very close I should have felt him (player might be my shadow!)
      if( updatePlayer )
	if( dist( audioSnap.otherTeamArray[i].pos, worldSnap.my.pos ) < VISIBLE_DISTANCE )
	  if( !feltPlayer )
	    updatePlayer = false;

      // if I felt that player I have to be careful with updates
      if( updatePlayer )
	if( dist( worldSnap.otherTeamArray[i].pos, worldSnap.my.pos ) < VISIBLE_DISTANCE )
	  checkAge = true;

      // update player if age is not to be checked or he is young enough
      if( updatePlayer )
	if( ( checkAge && audioSnap.otherTeamArray[i].age < worldSnap.otherTeamArray[i].age ) ||
	    ( !checkAge && audioSnap.otherTeamArray[i].age - 5 < worldSnap.otherTeamArray[i].age ) )
	  {
	    worldSnap.otherTeamArray[i].pos = audioSnap.otherTeamArray[i].pos;
	    worldSnap.otherTeamArray[i].age = audioSnap.otherTeamArray[i].age;
	    worldSnap.otherTeamArray[i].state = state;
	    worldSnap.otherTeamArray[i].coneTime = 0;

	    changed = true;
	    AUDIALINPUTTRACE( cout << i + 1 << " "; );
	  }
    }

  AUDIALINPUTTRACE( cout << endl; );

  // update ball if audial information is younger than my visual information
  // CAUTION: ball.v may be bad if sending player is far away from the ball
  if( audioSnap.ball.age < worldSnap.ball.age )
    {
      // be careful: do not overwrite good information with audio information
      // from players that are far from the ball
      if( audioSnap.ball.age < worldSnap.ball.age - 3 ||
	  sender == 0 ||   // coach!
	  sqrnorm( audioSnap.ourTeam( sender ).pos - worldSnap.ball.pos ) < sqrnorm( worldSnap.my.pos - worldSnap.ball.pos ) )
	{
	  worldSnap.ball = audioSnap.ball;
	  worldSnap.ball.state = AUDIAL;
	  worldSnap.ball.coneTime = 0;

	  changed = true;
	  AUDIALINPUTTRACE( cout << "Using Ball; pos = " << worldSnap.ball.pos << " , v = " << worldSnap.ball.v << endl; );
	}
    }

  // update goalie numbers (if sending player knows them but we do not)
  if( audioSnap.ourGoalieNo != UNKNOWN && worldSnap.ourGoalieNo == UNKNOWN )
    {
      worldSnap.ourGoalieNo = audioSnap.ourGoalieNo;
      for( int i = 0; i < 11; i++ )
	if( i == worldSnap.ourGoalieNo - 1 )
	  worldSnap.ourTeamArray[i].isGoalie = GOALIE_TRUE;
	else
	  worldSnap.ourTeamArray[i].isGoalie = GOALIE_FALSE;
    }
  if( audioSnap.otherGoalieNo != UNKNOWN && worldSnap.otherGoalieNo == UNKNOWN )
    {
      worldSnap.otherGoalieNo = audioSnap.otherGoalieNo;
      for( int i = 0; i < 11; i++ )
	if( i == worldSnap.otherGoalieNo - 1 )
	  worldSnap.otherTeamArray[i].isGoalie = GOALIE_TRUE;
	else
	  worldSnap.otherTeamArray[i].isGoalie = GOALIE_FALSE;
    }

  if( changed && audioSnap.time > visualTime )
    visualTime = audioSnap.time;

  // maxTime = audioSnap.time;

  return changed;
}


bool
Worldmodel::getSayMsg( AudioMsg &audioMsg )
{
  Snapshot audioSnap;
  Snapshot& worldSnap = snapshots[getTimeIndex( maxTime )];
  bool ok = true;

  AUDIALOUTPUTTRACE( cout << "Sending worldmodel : "; );

  // send ball
  audioSnap.ball = worldSnap.ball;

  AUDIALOUTPUTTRACE( cout << "ball with pos = " << audioSnap.ball.pos << " , v = " << audioSnap.ball.v << endl; );
  
  // send teams
  for( int i = 0; i < 11; i++ )
    {
      audioSnap.ourTeamArray[i] = worldSnap.ourTeamArray[i];
      audioSnap.otherTeamArray[i] = worldSnap.otherTeamArray[i];
    }

  audioSnap.ourGoalieNo = worldSnap.ourGoalieNo;
  audioSnap.otherGoalieNo = worldSnap.otherGoalieNo;

  audioMsg.markWrite();
  audioMsg << true;
  audioMsg << audioSnap;
      
  // check if error occurred
  if( !audioMsg.getState() )
    {
      audioMsg.rollbackWrite().clearState();
      audioMsg << false;
      ok = false;
    }

  AUDIALOUTPUTTRACE( if( !ok ) cout << "Didn't send anything!" << endl; );

  return ok;
}
