//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

// $Id: ballhandlingAct.C,v 1.59 2000/10/19 16:30:10 eri Exp $

/////////////////////////////////////////////////////////////////////////
//
// ACT
//
/////////////////////////////////////////////////////////////////////////


bool 
BallHandlingModule::act( AdvSkills* skills,
			 const Snapshot* snap, 
			 TacticalSnapshot* tsnap ) {
  // main action method, called by master control to execute action
  // previously evaluated

  if (!skills) {

    CERR ("act -> failed as skills==NULL" );
    return false;

  } else if (! checkDataStorage(snap, tsnap)) {

    // in some way corrupted, check master control for correct order of calls
    // (feedback .. -> .. evaluate .. -> act)
    return false;    
  } else {

    switch (action) {
      
      // in each case:
      //  - report the tactical action to the tactical module
      //  - (re)set view angle and head angle
      //  - output via trace macros
      //  - execute ONE from ( kick | dash | turn | catch )

    case WAIT:
      // no TacticalAction to report
      skills->changeView( snap, VIEW_WIDTH_NORMAL); // reset view width

      // turn head towards current body dir, i.e. straight
      if(turnHead ( skills, snap, my.bodyDir )) {
	COUT ("act -> nothing to do, straightening head");
      }
      
      turnParams = skills->turn( snap, (ball.pos - my.pos).getAngle());
      if ( (turnParams.state == AdvSkills::ROK) ||
	   (turnParams.state == AdvSkills::RANGLE) ) {
	COUT ("act -> nothing to do, turning body to ball");
	skills->exec( snap, turnParams);
      } else {
	COUT ("act -> nothing to do, waiting");
	skills->noAction();
      }
      break;
            
    case FIND_BALL:
      // no TacticalAction to report
      skills->changeView( snap, VIEW_WIDTH_NARROW);
      if(turnHead ( skills, snap, my.bodyDir)) {
	COUT ("act -> straightening head when looking for ball");
      }
      findBall(skills, snap, tsnap);
      break;
      
    case GOALIE_ATTACK:
      tsnap->reportOurTacticalAction (TA_REACT);
      TACTICALTRACE("goalie attack, TA_REACT reported");
      skills->changeView( snap, VIEW_WIDTH_NORMAL);
      // [!] goalie attack is not really ok
      goalieAttack( skills, snap, tsnap );
      break;
      
    case GOALIE_INTERCEPT:
      tsnap->reportOurTacticalAction (TA_CONTROL,my.number);
      TACTICALTRACE("goalie intercept, TA_CONTROL reported");
      // if near to ball, maximize number of visual infos
      if (my.distTo(ball) < 2*CATCHABLE_AREA_L &&
	  my.inOurPenaltyArea())
	skills->changeView(snap, VIEW_WIDTH_NARROW);
      else
	skills->changeView(snap, VIEW_WIDTH_NORMAL);
      goalieIntercept( skills, snap, tsnap );
      break;
            
    case GOALIE_CATCH:
      tsnap->reportOurTacticalAction (TA_CONTROL,my.number);
      skills->changeView( snap, VIEW_WIDTH_NORMAL);
      goalieCatch( skills, snap, tsnap );
      break;

    case GOALIE_DRIBBLE:
      tsnap->reportOurTacticalAction (TA_CONTROL,my.number);
      skills->changeView( snap, VIEW_WIDTH_NORMAL);
      fastdribble( skills, snap, tsnap );
      break;
            
    case GOALIE_PLACE:
      tsnap->reportOurTacticalAction (TA_CONTROL,my.number);
      skills->changeView( snap, VIEW_WIDTH_NORMAL);
      place( skills, snap, tsnap );
      break;
      
    case PLAYER_ATTACK: 
      tsnap->reportOurTacticalAction (TA_REACT);
      TACTICALTRACE("player attack, TA_REACT reported");
      skills->changeView( snap, VIEW_WIDTH_NORMAL);
      playerAttack( skills, snap, tsnap);
      break;

    case PLAYER_INTERCEPT:
      skills->changeView( snap, VIEW_WIDTH_NORMAL);
      playerIntercept( skills, snap, tsnap);

      // check if we safely reached the ball, or if we lost it
      // and return the appropriate tacticalAction
      if ((tsnap->ourAction.state != TA_PASS) ||
	  (tsnap->firstResponsiblePlayer == my.number)) {
        tsnap->reportOurTacticalAction (TA_CONTROL, my.number);
        TACTICALTRACE("player intercept, I am first responsible, "
		      << "TA_CONTROL reported");
	break;
      }

      if ((lastDistToBall > distToBall) &&
	  ((tsnap->firstDangerousOpponent == -1) ||
	   (tsnap->ourTeamReachBallTime(my.number) <
	    tsnap->otherTeamReachBallTime(tsnap->firstDangerousOpponent)))) {
	TACTICALTRACE("player intercept, getting closer and opponent "
		      << "is further away, TA_PASS kept");
	break;
      }

      tsnap->reportOurTacticalAction (TA_REACT, my.number);
      TACTICALTRACE("player intercept, will not be first at ball, "
		    << "TA_REACT reported");
      break;

    case PLAYER_FAST_DRIBBLE:
      tsnap->reportOurTacticalAction (TA_CONTROL,my.number);
      TACTICALTRACE("player fast dribble, TA_CONTROL reported");
      skills->changeView( snap, VIEW_WIDTH_NORMAL);
      fastdribble( skills, snap, tsnap);
      break;
      
    case PLAYER_DRIBBLE:
      tsnap->reportOurTacticalAction (TA_CONTROL,my.number);
      TACTICALTRACE("player dribble, TA_CONTROL reported");
      skills->changeView( snap, VIEW_WIDTH_NORMAL);
      dribble( skills, snap, tsnap );
      break;

    case PLAYER_PLACE:
      tsnap->reportOurTacticalAction (TA_CONTROL,my.number);
      TACTICALTRACE("place, TA_CONTROL reported");
      skills->changeView( snap, VIEW_WIDTH_NORMAL);
      place( skills, snap, tsnap);
      break;

    default:
      COUT ("act -> dont know what to do with action = " << (int)action);
      skills->noAction();
      return false;
      
    }
    return true; 
  }
}

void
BallHandlingModule::findBall( AdvSkills        *skills,
			      const Snapshot   *snap,
			      TacticalSnapshot *) {
  // turn continously to find ball on field (quite simple algorithm)

  if (!lookingForBall) { // we just started the search this timestep
    lookingForBall = true;
    
    // turn left or right to find ball ? (take last ball dir into account)
    turnAngle = (snap->my.bodyAngleTo(snap->ball ) > 0) ?
      DEG_TO_RAD(+40.0) : DEG_TO_RAD(-40.0);

    FINDTRACE("started ball search, bodyAngleToBall = " 
	      << my.bodyAngleTo(ball) 
	      << ", turnDir = " << turnAngle);

  } else {
    FINDTRACE("continuing ball search");
  }
  
  lastBallAge = ball.age;
  
  COUT ("act -> turning slowly to find ball");
  skills->basicTurn(turnAngle);
}

void
BallHandlingModule::playerIntercept( AdvSkills        *skills,
				     const Snapshot   *snap, 
				     TacticalSnapshot *tsnap) {
  // calculate earliest position where rolling ball can be reached, 
  // and run to that position (takes slight changes of this position 
  // into account)

  const int maxReachTime = 4; // maximum time delay allowed when intercepting
  const double secBorder = 2.0; // safety distance from edge of field
  WorldBall meet;

  // see what controlBall will report
  turtleParams = skills->controlBall(snap, availableStamina,
				     skills->kickRange, meet);

  if (turtleParams.state != AdvSkills::TFINISHED) {
    // not at the ball yet

    if ((turtleParams.time > 2) && (meet.v.getX() > 0.2)) {
      // We will need some time (more than 2 timesteps) until we get the ball.
      // If ball rolls toward other goal, get it later, if ball will be safe.

      // [!!] this intercept point calculation does not quite suit me...
      // [!] check if controlBall is doing its job correctly !
      
      int closestTime = turtleParams.time;
      int timeToReach = (tsnap->firstDangerousOpponent == UNKNOWN) ?
	maxReachTime : 
	// opponent may be closer than we think ("6" is hand-guessed)
	(tsnap->otherTeamReachBallTime(tsnap->firstDangerousOpponent) -
	 snap->otherTeam(tsnap->firstDangerousOpponent).age - 6);
     
      if (timeToReach > maxReachTime)
	timeToReach = maxReachTime;

      // test for maximum safe delay before opponent will also be at meet pos:

      // move finalball to meet pos of opponent
      Vector2d finalball = meet.pos + (1 - power(BALL_DECAY, timeToReach))/
	(1 - BALL_DECAY)*meet.v;
      // [!!!] bug: should this rather not be (timeToReach-closestTime) ??
      Vector2d finalv = power(BALL_DECAY, timeToReach)*meet.v;
     
      // adjust intercept point to some time later, according to safe delay,
      // because ball is rolling in right direction and we can afford to make
      // a safer intercept:
      while (timeToReach > closestTime) {
	// only use simulated position if ball safely inside field 
	// and not too far in the front
	if ((finalball.getX() > -FIELD_LENGTH/2 + secBorder) &&
	    (finalball.getX() < FIELD_LENGTH/5) &&
	    (fabs(finalball.getY()) < FIELD_WIDTH/2 - secBorder)) {
	  turtleParams =
	    skills->goToPositionTimed(snap, finalball, availableStamina,
				      skills->kickRange);
	  if (turtleParams.time < timeToReach)
	    // found a better position
	    break;
	}
	timeToReach--;
	finalv *= 1/BALL_DECAY;
	finalball -= finalv;
      }
      
      if (timeToReach > closestTime)
	ICPTTRACE("intercepting with a delay of " <<
		  timeToReach - closestTime);
      DRAWTRACE(CIRCLE(finalball, 1));
      
    }
    skills->exec(snap, turtleParams);
   
 
    if (turnHeadToBall(skills, snap, turtleParams))
      COUT("act->turningHead for controlBall");
    ICPTTRACE ("this is an intercept in " << turtleParams.time );
    return;
  }

  // finished 
  if (turnHead(skills, snap, snap->my.bodyDir))
    ICPTTRACE("straightening head");
  turnParams = skills->turn(snap,
			    (snap->ball.pos - snap->my.pos).getAngle());
  skills->exec(snap, turnParams);
  ICPTTRACE ("turning to ball");
}

void
BallHandlingModule::goalieIntercept( AdvSkills        *skills,
				     const Snapshot   *snap, 
				     TacticalSnapshot *) {

  const double goaliePenPrec = 0.8*CATCHABLE_AREA_L;
  WorldBall meet;

  // try controlBall first
  turtleParams =
    skills->controlBall(snap, availableStamina, goaliePenPrec, meet);
  // if we can't catch, we need the ball nearer
  if (!meet.inOurPenaltyArea())
    turtleParams = skills->controlBall(snap, availableStamina, skills->kickRange, meet);

  // in all cases use controlBall
  if (turtleParams.state != AdvSkills::TFINISHED) {
    skills->exec(snap, turtleParams);
    if (turnHeadToBall(skills, snap, turtleParams))
      COUT("act->turningHead for controlBall");
    ICPTTRACE ("this is a closest intercept in " << turtleParams.time );
    return;
  }

  // intercept finished
  if (turnHead(skills, snap, snap->my.bodyDir))
    ICPTTRACE("straightening head");
  turnParams = skills->turn(snap,
			    (snap->ball.pos - snap->my.pos).getAngle());
  skills->exec(snap, turnParams);
  ICPTTRACE ("turning to ball");
}


void
BallHandlingModule::playerAttack( AdvSkills*        skills,
				  const Snapshot*   snap, 
				  TacticalSnapshot* tsnap ) {

  if (tsnap->firstDangerousOpponent == UNKNOWN) {
    CERR ("act -> attack failed, because firstDangerousOpponent is not set !");
    skills->noAction();
    return;
  }

  const double distToOpponent = 3 * PLAYER_SIZE; 
  // distance to keep when going around opponent in close range

  WorldPlayer opponent = snap->otherTeam(tsnap->firstDangerousOpponent);
  WorldPlayer closestFriend;
  
  // is there a closest friend? ("close" means in opponent's kickrange)
  bool closestFriendExists = false;
  if (my.number == tsnap->firstResponsiblePlayer) {
    if (tsnap->secondResponsiblePlayer != UNKNOWN) {
      closestFriend = snap->ourTeam(tsnap->secondResponsiblePlayer);
      closestFriendExists = ( closestFriend.distTo( opponent.pos ) < KICKRANGE );
    } 
  } else if (tsnap->firstResponsiblePlayer != UNKNOWN ) {
    closestFriend = snap->ourTeam(tsnap->firstResponsiblePlayer);
    closestFriendExists = ( closestFriend.distTo( opponent.pos ) < KICKRANGE );
  }

  // ball is not kickable right now (otherwise we wouldnt attack, but dribble)

  int       simTime = 0;      // time opponent will need to reach the ball
  WorldBall simBall = ball;   // ball where opponent will reach it
  double    stam    = availableStamina;
  
  simBall = tsnap->otherTeamReachBallPos( tsnap->firstDangerousOpponent );
  simTime = tsnap->otherTeamReachBallTime( tsnap->firstDangerousOpponent );

  AdvSkills::TurtleCommand goToParams = skills->goToPositionTimed( snap, simBall.pos, stam, skills->kickRange );

  if( simTime == 0 )  // i.e. ball is in opponent's kickrange
    {
      // if opponent is not in my way, try to intercept directly
      if( goToParams.time <= 1 &&
	  my.distTo(simBall.pos) < my.distTo(opponent)) {
	skills->exec( snap, goToParams );
	turnHeadToBall( skills, snap, goToParams );
	return;
      }
    }

  if ( (closestFriendExists) &&
       (tsnap->myTacticalNo < tsnap->ourTeamTacticalNo(closestFriend.number)) ) {

    // I am second attacker:
    
    // if first attacker is between opponent and ball, block opponent
    if( closestFriend.inCorridor( opponent.pos, snap->ball.pos, skills->kickRange ) )
      {
	goToParams = skills->goToPositionTimed( snap, opponent.pos, stam, 2 * PLAYER_SIZE );
	skills->exec( snap, goToParams );
	turnHeadToBall( skills, snap, goToParams );
	return;
      }

    // else try to directly intercept ball
    ATTACKTRACE ("act -> second attacker; intercepting ball");
    goToParams = goAroundOpponent( skills, simBall.pos, opponent, distToOpponent, stamina, snap );
    DRAWTRACE ( POSITION ( simBall.pos ) );
    
  } else {
    
    // I am first attacker:
    Vector2d target;

    // if I am between opponent and ball, just intercept
    if( snap->my.inCorridor( opponent.pos, snap->ball.pos, skills->kickRange ) )
      {
	ATTACKTRACE("act -> between ball and player, intercepting ball");
	turtleParams = skills->controlBall( snap, stam );
	skills->exec( snap, turtleParams );
	turnHeadToBall( skills, snap, turtleParams );
	return;
      }

    if (simBall.pos.getX() > 0) {
      // it's likely the ball is kicked, so attack ball directly
      target = simBall.pos;
    }
    else if ((my.pos.getX() > simBall.pos.getX())  &&
	     (my.pos.getX() > -0.5*FIELD_LENGTH + PENALTY_AREA_LENGTH)) {
      // go to intersection of player to goal line and penalty area border
      target = Vector2d(-0.5*FIELD_LENGTH + PENALTY_AREA_LENGTH,
			PENALTY_AREA_LENGTH*simBall.pos.getY()/(0.5*FIELD_LENGTH + simBall.pos.getX()));
    }
    else {
      // else try to block opponent
      double offsetDist = min( 3 * skills->kickRange,
			       max( 0.0, MAXPOWER * DASH_POWER_RATE * (goToParams.time - simTime ) ) );
      if( snap->my.inCorridor( simBall.pos, snap->ourGoal.pos, 2 * skills->kickRange ) )
	offsetDist = 0.0;
      target = simBall.pos + radAng2Vector2d( offsetDist, ( snap->ourGoal.pos - simBall.pos ).getAngle() );
    }
    goToParams = goAroundOpponent( skills, target, opponent, distToOpponent, stamina, snap );
    ATTACKTRACE("act -> first attacker; blocking opponent");
    DRAWTRACE ( POSITION ( target ) );
  }
 
  // goToParams from one of the goAroundOpponent() above
  if( goToParams.state == AdvSkills::TFINISHED )
    {
      ATTACKTRACE("act -> reached target, intercepting ball");
      goToParams = skills->goToPositionTimed( snap, snap->ball.pos, stam, skills->kickRange );
      if( goToParams.state == AdvSkills::TFINISHED )
	{
	  ATTACKTRACE("act -> ??? already reached ball ???");
	  skills->noAction();
	  return;
	}
      skills->exec( snap, goToParams );
      turnHeadToBall( skills, snap, goToParams );
      return;
    }
  
  ATTACKTRACE("act -> going to attack position");
  skills->exec( snap, goToParams );
  turnHeadToBall( skills, snap, goToParams );

  return;
}


AdvSkills::TurtleCommand 
BallHandlingModule::goAroundOpponent(AdvSkills *skills, Vector2d target,
				     WorldPlayer opponent,
				     double dist,
				     double stamina,
				     const Snapshot* snap ) {

  // find a (non-flickering) way around an opponent, either left or right
  
  Vector2d relopp = opponent.pos - my.pos;
  
  // if opponent is in this corridor, we could collide and should evade
  double corrWidth = dist;
  if (!opponent.inCorridor(snap->my.pos, target, corrWidth))
    return skills->goToPositionTimed(snap, target, stamina, dist/2);
  Vector2d orthVector = relopp.perp();
  if (norm(orthVector) > 1e-5)
    orthVector *= dist/norm(orthVector);
  Vector2d target_alt1 = opponent.pos + orthVector;
  Vector2d target_alt2 = opponent.pos - orthVector;

  // choose from the two orthogonals, consider the time used to reach those points
  // this should avoid flickering
  AdvSkills::TurtleCommand c_alt1   = skills->goToPositionTimed(snap, target_alt1, stamina, dist/2);
  AdvSkills::TurtleCommand c_alt2   = skills->goToPositionTimed(snap, target_alt2, stamina, dist/2);

  if (c_alt1.time < c_alt2.time)
    return skills->goToPosition(snap, target_alt1, dist/2, min( MAXPOWER, stamina ));
  if (c_alt1.time > c_alt2.time)
    return skills->goToPosition(snap, target_alt2, dist/2, min( MAXPOWER, stamina ));
  return (sqrnorm(target_alt1 - my.pos) < sqrnorm(target_alt2 - my.pos)) ?
    skills->goToPosition(snap, target_alt1, dist/2, min( MAXPOWER, stamina )) :
    skills->goToPosition(snap, target_alt2, dist/2, min( MAXPOWER, stamina ));
}


void
BallHandlingModule::goalieAttack( AdvSkills*        skills,
				  const Snapshot*   snap, 
				  TacticalSnapshot* tsnap ) {
  // general description of behaviour:
  // - stand in the way of first opponent's goal kick
  // - when there, use strafing (moving sideways) to continously block opponent
  // - eventually go for the ball when it is close

  // [!!] goalie has to consider more than one opponent, and friends as well
  
  if (tsnap->firstDangerousOpponent == UNKNOWN) {
    CERR ("act -> goalie attack failed, firstDangerousOpponent is unknown !");
    skills->noAction();
    return;
  }
  
  // threshold dist for changing intercept position completely
  // (going from 'strafe-mode' back to 'move into goal kick path')
  double attackError = 3.0;
  
  WorldPlayer opponent = snap->otherTeam(tsnap->firstDangerousOpponent);
  
  WorldBall simBall     = ball;
  double    leftStamina = availableStamina;
  int       simTime     = 0;

  // [!] is safety dist concept ok? 
  // or should we rather stay inside goal area (instead of penalty area)
  double safetyDist = setSafetyDistance(snap); // min. dist to edge of p. area

  bool canReach = simulateOpponentGoalkick ( my, opponent, simTime, simBall, 
					     leftStamina, safetyDist );
  // true if ball can be reached if opponent does an optimal goal kick
  
  DRAWTRACE ( POSITION (simBall.pos) );
  if (canReach) {
    DRAWTRACE ( CTEXT ("goalie ball reach pos") );
  } else {
    DRAWTRACE ( CTEXT ("goalie minimum ball dist intercept pos") );
  }
  
  double power = min ( 2*MAXPOWER, availableStamina-2*MAXPOWER );
  
  turtleParams = skills->goToPosition( snap, simBall.pos, attackError, power, 
				       true, (ball.pos-my.pos).getAngle());
  // first of all go near the blocking position
  
  if (turtleParams.state == AdvSkills::TGOINGBACKWARDS) {
    ATTACKTRACE("going backwards when attacking opponent");
  }

  if (turtleParams.state == AdvSkills::TFINISHED) {
    // when close to blocking position, start strafing
    // [!!] strafing code needs some cleanup, eventually move it into skills
    
    // maximum Angle error allowed when strafing (otherwise turn first)
    // (if this is too small, strafe will continuosly 
    // adapt by turning instead of strafing!)
    Angle maxStrafeError = DEG_TO_RAD(45.0);
    
    // strafeDir is orthogonal to ball->ourGoal vector
    Angle strafeDir = (ourGoal.pos - ball.pos).getAngle() + PI_2;
    
    if ( fabs ( my.bodyDir-strafeDir ) > maxStrafeError ) {
      // turn to new strafeDir
      COUT ("act -> near ball intercept pos, turning to strafeDir="
	    << strafeDir);
      turnParams = skills->turn ( snap, strafeDir );
      skills->exec( snap, turnParams ); // regardless of skills->minTurnError
      if (turnHeadToBall ( skills, snap, turnParams )) {
	COUT ("turning head to next ball pos when turning for strafing");
      }
    } else {
      
      // calculate strafe distance to correct blocking position
      Vector2d ballgoalperp = (ourGoal.pos-ball.pos).perp();
      double p = ballgoalperp*ball.pos - ballgoalperp*my.pos;
      double q = ballgoalperp*radAng2Vector2d(1.0, my.bodyDir);
      
      if (fabs(q) < EPS) {
	// cannot happen, as angle is at least maxStrafeError degrees off strafeDir !
	CERR ("unexpeced strafe error, skipping goalie command");
	skills->noAction();
	return;
      }
      
      double strafeDistance = p/q;

      double power = strafeDistance * (1.0/DASH_POWER_RATE); // power needed
      
      ATTACKTRACE ("strafe: myDistToInterceptPos=" << my.distTo(simBall.pos)
		   << ", strafeDist=" << strafeDistance
		   << ", strafePower=" << power);
    
      // take backwards movement into account
      if (power < 0) {
	power = max ( power, MINPOWER );
	if (-2.0*power > availableStamina) power = -availableStamina/2.0;
      } else {
	power = min ( power, MAXPOWER );
	power = min ( power, availableStamina );
      }
      
      skills->basicDash (power);
      if (turnHead ( skills, snap, (ball.pos-my.pos).getAngle()) ) {
	COUT ("turning head to ball when strafing");
      }
      
    }
    
  } else {
    
    COUT ("act -> going for ball intercept pos, distToPos=" 
	  << my.distTo(simBall.pos) << ", stam=" << availableStamina);
    
    skills->exec(snap, turtleParams);
  }
  
  if( turnHeadToBall ( skills, snap, turtleParams ) ) {
    COUT ("act -> turning head towards ball when attacking");
  }
  
}


double BallHandlingModule::setSafetyDistance( const Snapshot* snap) {
  double oppMinDistToGoal;
  double safetyDist;

  // [!] personally I dont like this safety dist concept 
  // to keep goalie near goal

  // a very simple 'first try':

  double maxSafetyDistance = PENALTY_AREA_LENGTH - 4.0;

  if ( snap->otherTeamClosestTo(snap->ourGoal.pos) > 0 ) {
    oppMinDistToGoal = 
      snap->otherTeam(snap->otherTeamClosestTo(snap->ourGoal.pos)).
      distTo(snap->ourGoal.pos);
    safetyDist = max (3.0, PENALTY_AREA_LENGTH-oppMinDistToGoal);
    safetyDist = min (safetyDist, maxSafetyDistance);
    ATTACKTRACE("SafetyDistance = " << safetyDist);
  } else {
    safetyDist = 3.0;
  }
  return safetyDist;
}  


void
BallHandlingModule::fastdribble( AdvSkills        *skills,
				 const Snapshot   *snap,
				 TacticalSnapshot *) {
  // [!] check if fastDribble is safe with respect to player's body dir
  // (fastDribble is only safe within fd_numSteps if body dir is correct)

  Vector2d neededBallVel; // speed the ball will initially need

  double dist = fd_dashPower/(1 - PLAYER_DECAY)*
    (fd_numSteps - PLAYER_DECAY*(1 - power(PLAYER_DECAY, fd_numSteps))/(1 - PLAYER_DECAY));
  // dist to kick ball ahead to (projected so that numSteps dashing 
  // with dashPower will let us reach ball after that dist)

  neededBallVel.setRadAng((1 - BALL_DECAY)/(1 - power(BALL_DECAY, fd_numSteps))*dist, dribbleDir);
  // initial ball velocity to achieve that fast dribble
  

  DRIBBLETRACE(" initial ball speed=" << neededBallVel
	       << " dribble dist=" << dist);
    
  kickParams = skills->kick( snap, neededBallVel );
    
  if (kickParams.state == AdvSkills::KRANGE) {
    skills->noAction();
    COUT("fd: kicking out of range");
  }
  else
    skills->exec ( snap, kickParams );

  if (turnHead(skills, snap, my.bodyDir)) {
    COUT("fd: looking forward into fd dir");
  }
}


void
BallHandlingModule::dribble( AdvSkills*        skills,
			     const Snapshot*   snap, 
			     TacticalSnapshot *) {
  // keep ball under control while running

  dribbleParams = skills->dribble( snap, dribbleDir, 
				   holdDir, dribbleStamina );
  COUT ("act -> dribbling");
  skills->exec( snap, dribbleParams );
  DRAWTRACE ( LINE ( my.pos, my.pos + radAng2Vector2d ( 10.0, dribbleDir ) ) );
  DRAWTRACE ( CTEXT ("dribble dir") );

  if ( turnHeadForDribble ( skills, snap, dribbleParams, dribbleDir ) ) {
    COUT ("act -> turning head for dribble");
  }
  
}


void
BallHandlingModule::place( AdvSkills*        skills,
			   const Snapshot*   snap, 
			   TacticalSnapshot* tsnap ) {
  // move ball to placeDir on a safe circle around us.

  if (tsnap);
  
  Vector2d placePos;
  placePos.setRadAng( skills->safeCircle, placeDir );
  
  kickParams = skills->placeBall( snap, placePos );
  if (kickParams.state == AdvSkills::KPLACING) {

    COUT ("act -> placing ball to placeDir=" << RAD_TO_DEG(placeDir) );
    skills->exec( snap, kickParams);

    WorldBall simBall = ball;
    simBall.pos += kickParams.v; // pos after kicking
    Angle simBallDir = (simBall.pos - my.pos - my.v).getAngle();
    if(turnHead ( skills, snap, simBallDir )) {
      COUT ("act -> turning head to next ball pos when placing, dir=" 
	    << simBallDir);
    }

  } else {
    CERR ("act -> trying to place ball, but failed!");
    skills->noAction();
  }

}


void
BallHandlingModule::goalieCatch ( AdvSkills*        skills,
				  const Snapshot*   snap, 
				  TacticalSnapshot* tsnap ) {
  
  // try to catch ball directly

  if (tsnap);

  WorldBall simBall = ball;
  simBall.simulate(); // ball pos when catch fails
  Angle simBallDir = (simBall.pos - my.pos).getAngle();
  if(turnHead ( skills, snap, simBallDir )) {
    COUT("act -> turning head towards ball at dir=" 
	 << simBallDir << " when catching");
  }
  
  COUT ("act -> catching ball");
  skills->exec(snap, catchParams); // params computed in evaluate()
}


// some turnHead auxiliary methods...
/////////////////////////////////////////////////////////////////

bool
BallHandlingModule::reportTHState(AdvSkills *skills, const Snapshot *snap,
				  const AdvSkills::TurnHeadCommand &params)
{
  switch (params.state) {
  case (AdvSkills::THOK):
    TURNHEADTRACE("can make the turn head");
    skills->exec(snap, params);
    return true;
  case (AdvSkills::THANGLE):
    TURNHEADTRACE("cannot make it completely, truncating angle");
    skills->exec(snap, params);
    return true;
  case (AdvSkills::THBAN):
    TURNHEADTRACE("turn head still banned");
    return false;
  case (AdvSkills::THFINISHED):
    TURNHEADTRACE("head angle is within accepted range");
    return false;
  default:
    TURNHEADTRACE("what turnHeadState is this ??");
    return false;
  } // switch
}

bool
BallHandlingModule::turnHead( AdvSkills* skills, const Snapshot* snap, 
			      Angle targetDir)
{
  AdvSkills::TurnHeadCommand params = skills->turnHead (snap, targetDir);
  return reportTHState(skills, snap, params);
}


bool
BallHandlingModule::turnHeadToBall( AdvSkills* skills, const Snapshot* snap)
{
  Angle turnHeadDir = (ball.pos - my.pos).getAngle();
  return turnHead (skills, snap, turnHeadDir);
}

bool
BallHandlingModule::turnHeadToBall( AdvSkills* skills, const Snapshot* snap, 
				    AdvSkills::TurtleCommand tc)
{
  AdvSkills::TurnHeadCommand params = skills->turnHeadToBall (snap, tc);
  return reportTHState(skills, snap, params);
}


bool
BallHandlingModule::turnHeadToBall( AdvSkills* skills, const Snapshot* snap, 
				    AdvSkills::TurnCommand tc)
{
  AdvSkills::TurnHeadCommand params = skills->turnHeadToBall (snap, tc);
  return reportTHState(skills, snap, params);
}



bool
BallHandlingModule::turnHeadToBall( AdvSkills* skills, const Snapshot* snap, 
				    AdvSkills::KickCommand kc) {

  Angle turnHeadDir = (snap->ball.pos + kc.v - snap->my.pos - snap->my.v).getAngle();
  
  return turnHead (skills, snap, turnHeadDir );
}


bool
BallHandlingModule::turnHeadForDribble( AdvSkills* skills, 
					const Snapshot* snap, 
					AdvSkills::DribbleCommand dc,
					Angle dribbleDir)
{
  AdvSkills::TurnHeadCommand params = skills->turnHeadForDribble (snap, dc, dribbleDir);
  return reportTHState(skills, snap, params);
}

// simulate how the opponent performs an optimal goal kick:
bool
BallHandlingModule::simulateOpponentGoalkick( WorldPlayer simMy, 
					      WorldPlayer simOpponent,
					      int &simTime, 
					      WorldBall &simBall,
					      double &leftStamina,
					      double safetyDist) {
  
  // simulates opponent intercept and immediate kick towards goal
  // changes simTime, simBall (interceptPos)  and leftStamina
  // returns true, if ball can be intercepted before reaching the goal, 
  // false otherwise
  
  double      simBallSpeed           = BALL_SPEED_MAX;
  double      opponentInterceptSpeed = PLAYER_SPEED_MAX;
  
  double oppRange = PLAYER_SIZE + BALL_SIZE + KICKABLE_MARGIN;
  double range    = 0; // for safety. CATCHABLE_AREA_L would be possible, 
  // but we want to make absolutely sure the goalie will be first at the ball/
  // at the position in time. So rather start him with a handikap.
  
  // minimum distance to ball in Penalta Area during simulation 
  Vector2d minDistSimBallPosInPA = simBall.pos;
  double minDistToSimBallInPA = simMy.distTo(simBall);
  
  Vector2d oppReachBallPos = simOpponent.pos;
  
  ATTACKTRACE ("*** start of simOpponent goalkick simulation ***");
  
  SIMTRACE("SimTime=" << simTime << ", oppRange=" 
	   << oppRange << ", oppDistToBall=" 
	   << simOpponent.distTo(simBall)-oppRange 
	   << ", ourDistToBall=" << simMy.distTo(simBall)-range);
  
  // simulate until opponent reaches ball, ball in goal or we reach ball
  while ( ( oppRange < simOpponent.distTo(simBall) ) &&
	  ( range < simMy.distTo(simBall) ) ) {
    
    oppRange += opponentInterceptSpeed;
    
    double pow = min (leftStamina, MAXPOWER);
    
    // use limit of player speed as simplification:
    range += pow * DASH_POWER_RATE * ( 1 / (1 - PLAYER_DECAY) ) ;
    leftStamina -= pow;
    
    simBall.simulate();
    
    leftStamina += STAMINA_INC_MAX;
    simTime += 1;
    
    // update minimum distance to ball in penalty area
    if ( (simMy.distTo(simBall) < minDistToSimBallInPA) &&
	 simBall.inOurPenaltyArea() ){
      minDistToSimBallInPA = simMy.distTo(simBall);
      minDistSimBallPosInPA = simBall.pos;
    }
    
    // check if ball aready in goal
    if (simBall.pos.getX() < -0.5 * FIELD_LENGTH) {
      ATTACKTRACE( "goalkick sim: ball will be behind goal line at simTime="
		   << simTime);
      ATTACKTRACE ("*** end goalkick simulation ***");
      simBall.pos = minDistSimBallPosInPA;
      return false;
    }
    
    SIMTRACE("SimTime=" << simTime << ", oppRange=" 
	     << oppRange << ", oppDistToBall=" 
	     << simOpponent.distTo(simBall)-oppRange 
	     << ", ourDistToBall=" << simMy.distTo(simBall)-range);  
  }
  
  if ( range > simMy.distTo(simBall) ) {
    ATTACKTRACE( "goalkick sim: we reach ball first at simTime=" << simTime);
    ATTACKTRACE ("*** end goalkick simulation ***");
    return true;
  } else {
    
    oppReachBallPos = simBall.pos;    
    
    ATTACKTRACE ("goalkick sim: opponent kicks ball at simTime=" << simTime);

    // from here the ball is moving in our goal's direction    
    simBall.v.setRadAng( simBallSpeed,
			 (ourGoal.pos - simBall.pos).getAngle() );
    
    SIMTRACE("SimTime=" << simTime << ", range=" << range
	     << ", ourDistToBall=" 
	     << simMy.distTo(simBall)-range
	     << ", stam=" << leftStamina);
    
    while ( range < simMy.distTo(simBall) ) {
      
      double pow = min (leftStamina, MAXPOWER);

      // use limit of player speed as simplification:
      range += pow * DASH_POWER_RATE * ( 1 / (1 - PLAYER_DECAY) ) ;      
      leftStamina -= pow;

      leftStamina += STAMINA_INC_MAX;
      simTime += 1;

      simBall.simulate();
      
      // update minimum dist to ball in PA
      if (simMy.distTo(simBall) < minDistToSimBallInPA) {
	minDistToSimBallInPA = simMy.distTo(simBall);
	minDistSimBallPosInPA = simBall.pos;
      }

      // check for ball behind goal line
      if (simBall.pos.getX() < -0.5 * FIELD_LENGTH) {
	ATTACKTRACE( "goalkick sim: ball will be behind goal line at simTime="
		     << simTime);
	simBall.pos = minDistSimBallPosInPA;
	ATTACKTRACE ("*** end goalkick simulation ***");
	return false;
      }
      
      SIMTRACE("SimTime=" << simTime << ", range=" << range
	       << ", ourDistToBall=" 
	       << simMy.distTo(simBall)-range
	       << ", stam=" << leftStamina);
      
    } //while

    if ( ( fabs (simBall.pos.getY()) < 0.5*PENALTY_AREA_WIDTH - safetyDist ) &&
	 ( simBall.pos.getX() < -0.5*FIELD_LENGTH+PENALTY_AREA_LENGTH - safetyDist ) ) {
      ATTACKTRACE ("goalkick sim: can reach kicked ball in safe penalty area "
		   << "at simTime=" << simTime);
      ATTACKTRACE ("*** end goalkick simulation ***");
      return true;
    } else {
      ATTACKTRACE ("goalkick sim: earliest intercept point outside safe penalty "
		   << "area at simTime=" << simTime);
      double x = 0;
      double y = 0; 
      double m = (ourGoal.pos.getY() - oppReachBallPos.getY()) /
	(ourGoal.pos.getX() -oppReachBallPos.getX());
      if (m!=0) {
	x = -0.5*FIELD_LENGTH+PENALTY_AREA_LENGTH - safetyDist;
	y = (x - oppReachBallPos.getX()) * m + oppReachBallPos.getY();
	if ( fabs(y) > 0.5 * PENALTY_AREA_WIDTH ) {
	  if (y > 0) {
	    y = 0.5 * PENALTY_AREA_WIDTH - safetyDist;
	    ATTACKTRACE("using point on bottom penalty area line");
	  } else {
	    y = -0.5 * PENALTY_AREA_WIDTH + safetyDist;
	    ATTACKTRACE("using point on top penalty area line");
	  }
	  x = oppReachBallPos.getX() + (1/m) * (y - oppReachBallPos.getY());
	} else {
	  ATTACKTRACE("using point on right penalty area line");
	}
      } else
	{
	  ATTACKTRACE("ball on line in center of field, " <<
		      "using point on center line");
	  y = oppReachBallPos.getY();
	  x = -0.5*FIELD_LENGTH+PENALTY_AREA_LENGTH - safetyDist;
	}
      simBall.pos.setX(x);
      simBall.pos.setY(y); 
      ATTACKTRACE ("*** end goalkick simulation ***");
      return true;
    } // outside

  }
  
}

double
BallHandlingModule::avgPower( double totalStamina, int time ) {

  if (time == 0) return totalStamina;
  else return ( totalStamina ) / time;
  
}
