//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

// $Id: ballhandlingEvaluate.C,v 1.52 2000/10/20 12:57:46 goetz Exp $

/////////////////////////////////////////////////////////////////////////
//
// EVALUATE
//
/////////////////////////////////////////////////////////////////////////


Grade 
BallHandlingModule::evaluate(AdvSkills* skills, 
			     const Snapshot* snap, 
			     TacticalSnapshot* tsnap) {
  // main evaluate method, called by master control to evaluate
  // current situation by module and return a grade of responsibility
  // and desired activity. Naturally plans some of the future actions
  // already in its computations.
  
  Grade g;

  if (!skills) {

    CERR ("eval -> failed as skills==NULL" );
    action = WAIT;
    g.set_IMPOSSIBLE();

  } else if (! checkDataStorage(snap, tsnap)) {

    // data store is corrupt, same snap was not fed into fedback before !
    action = WAIT;
    g.set_IMPOSSIBLE();

  } else {

    // [!] does standard situation want to use findBall as well? -> switch the two following ifs around.

    // check for standard situation
    if ( playMode != PLAY_MODE_PLAY_ON ) {

      action = WAIT;
      g.set_IMPOSSIBLE();           // standard sit must jump in here
      COUT ("eval -> planning to let standard sit have control,"
	    <<" grade = " << g);

    } else if ( needToLookForBall(snap, tsnap) ) {

      action = FIND_BALL;
      g.set_POOR();   
      COUT ("eval -> ball too old, planning to look for ball, grade = " << g);
      
    } else if (goalie) {
      
      g = evaluateGoalie(skills, snap, tsnap);

    } else {
      
      g = evaluatePlayer(skills, snap, tsnap);

    }

  }

  return g;

}


Grade
BallHandlingModule::evaluateGoalie( AdvSkills* skills,
				    const Snapshot* snap,
				    TacticalSnapshot* tsnap ) {
  // evaluate goalie's possible actions
  
  Grade g;
  
  WorldBall ballIn3Steps = snap->ball;
  ballIn3Steps.simulate();
  ballIn3Steps.simulate();
  ballIn3Steps.simulate();

  // see if there is anything dangerous out there first:
  bool danger = 
    (snap->ball.pos.getX() < -FIELD_LENGTH/6.0) && // ball near our goal
    (tsnap->ourAction.state != TA_PASS ||   // do not run after kickoffs!
     (tsnap->ourAction.state == TA_PASS && 
      tsnap->ourAction.performer != snap->my.number));
  // [!] if our players lose a pass near our goal, we are in trouble 
  // (TA_PASS has to be reset as soon as possible!)


  // if goalie kicks the ball away, there is no direct danger
  if( snap->ball.inOurPenaltyArea() && !ballIn3Steps.inOurPenaltyArea())
     danger = false;
  
  // only get active in dangerous situations:
  if (danger) {

    // try to catch ball, if possible -> PERFECT
    catchParams = skills->goalieCatch(snap);
    if (catchParams.state == AdvSkills::COK) {
      // check if near border of catchable area
      if (my.distTo(ball) > 0.9*CATCHABLE_AREA_L) {
	WorldBall wb;
	// ok, catch is unsafe, check if we can wait
	turtleParams = skills->controlBall(snap, 2*MAXPOWER,
					   0.8*CATCHABLE_AREA_L, wb);
	// one dash helps and no opponent to near
	if ((turtleParams.time == 1) &&
	    ((tsnap->firstDangerousOpponent == UNKNOWN) ||
	     (tsnap->otherTeamReachBallTime(tsnap->firstDangerousOpponent) > 1))) {
	  // check ball rolling into goal
	  WorldBall simBall = ball;
	  simBall.simulate();
	  // check for ball safely in penalty area next step
	  if ((simBall.pos.getX() > -0.5*FIELD_LENGTH + BALL_SIZE) &&
	      (simBall.pos.getX() < -0.5*FIELD_LENGTH + PENALTY_AREA_LENGTH - BALL_SIZE) &&
	      (fabs(simBall.pos.getY()) < 0.5*PENALTY_AREA_WIDTH - BALL_SIZE)) {
	    // another dash makes catching safer, so intercept
	    action = GOALIE_INTERCEPT;
	    g.set_PERFECT();
	    COUT ("eval -> catch unsafe, going near ball, grade = " << g);
	    return g;
	  }
	}
	// here for some reason a security dash was impossible, so better try unsafe catch
      }
      action = GOALIE_CATCH;
      g.set_PERFECT();
      COUT ("eval -> planning to catch ball, grade = " << g);
      return g;
    } // catching was ok
    
    // try to dribble ball back into penalty area or to place it
    if ( skills->kickSafe(snap) ) { 
      // ball is kickable

      g = goalieEvaluateDribble(snap, tsnap);
      action = GOALIE_DRIBBLE;
      
      COUT ("eval -> cannot catch ball, planning to dribble/place,"
	    << " grade = " << g);
      if ( g.is_BAD() )
	COUT ("dribbling/placing is BAD -> passmodule should do something!");

      return g;
    }

    // I am most responsible 
    if ( tsnap->firstResponsiblePlayer == my.number ) {

      // there is no dangerous opponent around
      if ( (tsnap->firstDangerousOpponent == UNKNOWN) ||
	   (tsnap->ourTeamReachBallTime(my.number) + 8 <
	    tsnap->otherTeamReachBallTime(tsnap->firstDangerousOpponent)) ){
	
	// go for the ball
	action = GOALIE_INTERCEPT;
	g.set_PERFECT();
	COUT ("eval -> closest to ball, planning to intercept it,  "
	      << "grade = " << g);
	return g;
      }

      // if there is a goalkick, intercept
      Vector2d ballEndPos = snap->ball.pos + snap->ball.v * ( 1.0 / ( 1.0 - BALL_DECAY ) );
      if( ballEndPos.getX() < -0.5 * FIELD_LENGTH )
	{
	  double factor = - ( 0.5 * FIELD_LENGTH + snap->ball.pos.getX() ) /
	    ( ballEndPos.getX() - snap->ball.pos.getX());
	  if( fabs( snap->ball.pos.getY() + factor *
		    ( ballEndPos.getY() - snap->ball.pos.getY() ) ) <= 0.5 * GOAL_WIDTH )
	    {
	      COUT ("eval-> ball will cross goal line, planning to intercept it");
	      action = GOALIE_INTERCEPT;
	      g.set_PERFECT();
	      return g;
	    }
	}

      // opponent controls the ball


      
      // [!!] here more than one opponent for the goalie has to be taken into account
      // do that in goalie_attack, or alternatively have a new action type 
      // to do this, or let positioning so it
      action = GOALIE_ATTACK;
      g.set_PERFECT();
      COUT ("eval -> close to ball, but opponent ist closer, "
	    << "planning to attack, grade = " << g);

      return g;

    }
    
    // I am not most responsible but still close to the ball
    if (( tsnap->secondResponsiblePlayer == my.number ) ||
	( tsnap->thirdResponsiblePlayer == my.number )) {

      // [!] same things as a few lines above: more than one opp!
      action = GOALIE_ATTACK;
      if( tsnap->firstDangerousOpponent == UNKNOWN )
	g.set_BAD();
      else
	g.set_PERFECT();

      COUT ("eval -> close to ball, but I am not firstResponsible and opponent is closer, "
	    << "planning to attack, grade = " << g);

      return g;
    }

    action = WAIT;
    g.set_BAD();
    COUT ("eval -> potential danger, but no special situation, grade = " << g);

    return g;
  }

  action = WAIT;
  g.set_BAD();
  COUT ("eval -> no danger, grade = " << g);
  
  return g;
}


Grade
BallHandlingModule::goalieEvaluateDribble ( const Snapshot* snap,
					    TacticalSnapshot* tsnap ) {
  // possibly place ball away from opponent, or dribble back into penalty area 
  // to have ball safe

  Grade g;
  
  if (ball.inOurPenaltyArea()) {
    action = GOALIE_PLACE;
    // placing is perfect if we can be sure that the opponent can not reach us
    // before we can catch again
    if ( (tsnap->firstDangerousOpponent == UNKNOWN) ||
	 (tsnap->otherTeamReachBallTime(tsnap->firstDangerousOpponent) >
	  max( 0, CATCH_BAN_CYCLE - tsnap->time + snap->lastCatchTime ) + 3)){
      placeDir = 0;
      DRIBBLETRACE(" placing ball away from goal (opponent is far away, placing should be safe)");
      g.set_PERFECT();
    } else {
      WorldPlayer first = snap->otherTeam(tsnap->firstDangerousOpponent);

      if ( (tsnap->secondDangerousOpponent != UNKNOWN) &&
	   (tsnap->otherTeamReachBallTime(tsnap->secondDangerousOpponent) < 5)) {
	WorldPlayer second = snap->otherTeam(tsnap->secondDangerousOpponent);
	placeDir = bisect ( (my.pos - first.pos).getAngle(),
			    (my.pos - second.pos).getAngle() );
	DRIBBLETRACE(" placing opposite of both closest opponents (placing is dangerous!)");
      }
      else {
	placeDir = (my.pos - first.pos).getAngle();
	DRIBBLETRACE(" placing opposite of closest opponent (placing is dangerous!)");
      }

      // placing is always bad if opponent is close!
      g.set_BAD();
    }
  } else {
    // dribble into penalty area
    action = GOALIE_DRIBBLE;
    Vector2d point;
    if (ball.pos.getX() > - 0.5 * FIELD_LENGTH + PENALTY_AREA_LENGTH - 1) {
      point.setX(- 0.5 * FIELD_LENGTH + PENALTY_AREA_LENGTH - 1);
      if (ball.pos.getY() < - 0.5 * PENALTY_AREA_WIDTH + 1) {
	point.setY(- 0.5 * PENALTY_AREA_WIDTH + 1);
      } else if (ball.pos.getY() > 0.5 * PENALTY_AREA_WIDTH - 1) {
	point.setY(0.5 * PENALTY_AREA_WIDTH - 1); 
      } else {
	point.setY(my.pos.getY());
      }
    } else {
      point.setX(ball.pos.getX());
      if (ball.pos.getY() < - 0.5 * PENALTY_AREA_WIDTH + 1) {
	point.setY( - 0.5 * PENALTY_AREA_WIDTH + 1);
      } else {
	point.setY( 0.5 * PENALTY_AREA_WIDTH - 1);
      }
    }
    Vector2d target = point - my.pos;
    
    dribbleDir = target.getAngle();
    
    bool dribbleSafe = ( snap->otherTeamInCorridor( my.pos, point, max( 10.0, 2 * target.getLength() ) ) == 0 &&
			 ( ( tsnap->firstDangerousOpponent == UNKNOWN ) ||
			   ( tsnap->otherTeamReachBallTime( tsnap->firstDangerousOpponent ) > 3 ) ) );

    if ( dribbleSafe )
      g.set_GOOD();
    else
      g.set_BAD();

    double tlen = target.getLength();
    fd_dashPower = MAXPOWER;
    // calculation power lower to avoid kicking to far
    for (fd_numSteps = 1; fd_numSteps < 20; fd_numSteps++) {
      double dist = fd_dashPower*DASH_POWER_RATE/(1 - PLAYER_DECAY)*
	(fd_numSteps - PLAYER_DECAY*(1 - power(PLAYER_DECAY, fd_numSteps))/(1 - PLAYER_DECAY));
      if (dist > tlen)
	break;
    }

    if (fd_numSteps*(MAXPOWER - STAMINA_INC_MAX) < availableStamina ||
	(fd_numSteps > 10)) { 
      g.set_BAD();
      CERR ("cannot kick ball back into penalty area safely, stamina too low!");
      CERR ("need passmodule to do something useful !");
    }
  }

  return g;
  
}


Grade 
BallHandlingModule::evaluatePlayer( AdvSkills* skills,
				    const Snapshot* snap,
				    TacticalSnapshot *tsnap)
{
  
  Grade g;


  // update our tactical action depending on current state and ball control

  guessedState = tsnap->ourAction.state;
  guessedPerformer = tsnap->ourAction.performer;
  guessedReceiver = tsnap->ourAction.receiver;

  if ( (guessedState != TA_PASS) &&
       (guessedState != TA_GOALKICK) ) {
    if (tsnap->ourBallControl == BC_SAFE) { 
      guessedState = TA_CONTROL;
      guessedPerformer = tsnap->firstResponsiblePlayer;
    }
    if ( // (tsnap->ourBallControl == BC_DANGER) ||
	(tsnap->ourBallControl == BC_LOST) )
      guessedState = TA_REACT;
  }

  TACTICALTRACE("guessed state = " << guessedState <<
		", guessed performer = " << guessedPerformer <<
		", guessed receiver = " << guessedReceiver);

  if (guessedState == TA_PASS) 
    COUT("Pass from player " << guessedPerformer 
	 << " to player " << guessedReceiver);

  // depending on currently guessed game state, decide
  switch (guessedState) {
  case TA_GOALKICK: // ball is being kicked into goal
  case TA_CONTROL : // ball is under control by our team
    if (skills->kickSafe(snap)) {
      g = evaluateDribble(snap, tsnap); // try to dribble, if possible
      break;
    }
    if (tsnap->firstResponsiblePlayer == my.number) {
      action = PLAYER_INTERCEPT; // otherwise intercept the ball
      g.set_PERFECT();
      COUT ("eval -> TA_CONTROL & firstResponsible & cannot kick, planning to intercept, "<< "grade = " << g);
      break;
    }
    action = WAIT; // or wait, if not responsible
    g.set_BAD();  // positioning must jump in here with a better grade!
    COUT ("eval -> TA_CONTROL & not responsible, planning to let positioning have control, "<< "grade = " << g);
    break;
  case TA_PASS : // ball is being passed in our team
    if (guessedReceiver == my.number) {
      action = PLAYER_INTERCEPT; // go for a pass directed at me
      g.set_PERFECT();
      COUT ("eval -> TA_PASS & receiver, planning to intercept pass, " 
	    << "grade = " << g);
    } else if ( (tsnap->firstResponsiblePlayer == my.number) &&
		(tsnap->ourTeamReachBallTime(my.number) < 4) &&
		(guessedPerformer != my.number)) {
      action = PLAYER_INTERCEPT; // ball is really close, so intercept anyway
      g.set_GOOD();
      COUT ("eval -> TA_PASS & close to ball, planning to intercept pass, " 
	    << "grade = " << g);
    } else {
      action = WAIT; // otherwise let the pass roll on
      g.set_BAD();
      COUT ("eval -> TA_PASS & not receiver, planning to have positioning have control, " << "grade = " << g);
    }
    break;  
  case TA_UNKNOWN : 
  case TA_REACT :   // opponent controls ball
    if( skills->kickSafe( snap ) ) {
      g = evaluateDribble( snap, tsnap ); // dribble, if possible
      break;
    }
    if (tsnap->firstResponsiblePlayer == my.number) {
      action = PLAYER_ATTACK; // attack opponent, if responsible
      g.set_PERFECT();
      COUT ("eval -> TA_REACT & firstResponsible, planning to attack, " 
	    << "grade = " << g);
      break;
    }
    if (tsnap->secondResponsiblePlayer == my.number) {
      action = PLAYER_ATTACK; // second responsible always supports the attack
      g.set_PERFECT();
      COUT ("eval -> TA_REACT & secondResponsible, planning to attack, " 
	    << "grade = " << g);
      break;
    }
    if ( (tsnap->thirdResponsiblePlayer == my.number) && 
	 ( (tsnap->firstResponsiblePlayer == snap->ourGoalieNo) ||
	   (tsnap->secondResponsiblePlayer == snap->ourGoalieNo)) ) {
      action = PLAYER_ATTACK; // third responsible only supports goalie
      g.set_PERFECT();
      COUT ("eval -> TA_REACT & ThirdResponsible & Goalie in trouble, planning to attack, " << "grade = " << g);
      break;
    }
    action = WAIT; // otherwise let the other team play on, and move to a better position
    g.set_BAD();
    COUT ("eval -> TA_REACT & not responsible, planning to have positioning have control, " << "grade = " << g);
    break; 
  }
  
  return g;
  
}

Grade 
BallHandlingModule::evaluateDribble(const Snapshot* snap,
				    TacticalSnapshot* tsnap)
{
  const double sqrt2 = sqrt(0.5);

  Grade g;

  // minimum dist to field border watched while dribbling
  const double kickRange = PLAYER_SIZE + BALL_SIZE + KICKABLE_MARGIN;
  const double distToBorder = 3*kickRange;
  const double dribbleCorrW = 2*kickRange;
  
  // preferred dribble directions (directions to be checked for dribbling)
  int  numDirs;
  Vector2d dirs[5];
  Grade    grades[5];
  bool     dribbles[5];

  const double sideBorder(5.0);
  const double backBorder(8.0);
  const double cornerDist((FIELD_WIDTH - GOAL_WIDTH)/2);
  const double targetingDist(cornerDist + GOAL_WIDTH/2);

  // the dribble directions depend on the player's position on the field:

  // 1. right opponent corner-triangle
  if (1.5*(FIELD_LENGTH/2 - my.pos.getX()) +
      (FIELD_WIDTH/2 - my.pos.getY()) < cornerDist) {
    numDirs = 3;
    dirs[0] = Vector2d(0.0, -1.0);     // upwards toward center of penalty area
    dirs[1] = Vector2d(-sqrt2, -sqrt2);// diagonally back and up
    dirs[2] = Vector2d(-1.0, 0.0);     // straight backwards
    grades[0].set_GOOD();
    grades[1].set_GOOD(); // would be generally a good idea
    grades[2].set_POOR(); // rather poor direction
    dribbles[0] = true;
    dribbles[1] = true;
    dribbles[2] = true;
  }
  // 2. left opponent corner-triangle
  else if (1.5*(FIELD_LENGTH/2 - my.pos.getX()) +
	   (FIELD_WIDTH/2 + my.pos.getY()) < cornerDist) {
    numDirs = 3;
    dirs[0] = Vector2d(0.0, 1.0); // etc.
    dirs[1] = Vector2d(-sqrt2, sqrt2);
    dirs[2] = Vector2d(-1.0, 0.0);
    grades[0].set_GOOD();
    grades[1].set_GOOD();
    grades[2].set_POOR();
    dribbles[0] = true;
    dribbles[1] = true;
    dribbles[2] = true;
  }
  // 3. right border
  else if (my.pos.getY() > FIELD_WIDTH/2 - sideBorder) {
    numDirs = 2;
    dirs[0] = Vector2d(sqrt2, -sqrt2); // diagonally forward and up
    dirs[1] = Vector2d(1.0, 0.0);      // straight forward
    grades[0].set_GOOD();
    grades[1].set_AVERAGE();
    dribbles[0] = false;
    dribbles[1] = true;
  }
  // 4. left border
  else if (my.pos.getY() < -FIELD_WIDTH/2 + sideBorder) {
    numDirs = 2;
    dirs[0] = Vector2d(sqrt2, sqrt2);
    dirs[1] = Vector2d(1.0, 0.0);
    grades[0].set_GOOD();
    grades[1].set_AVERAGE();
    dribbles[0] = false;
    dribbles[1] = true;
  }
  // 5. our backfield
  else if (my.pos.getX() < -FIELD_LENGTH/2 + backBorder) {
    numDirs = 3;
    if (my.pos.getY() > 0) {
      dirs[0] = Vector2d(sqrt2, sqrt2);
      dirs[1] = Vector2d(0.0, 1.0);
      dirs[2] = Vector2d(1.0, 0.0);
    }
    else {
      dirs[0] = Vector2d(sqrt2, -sqrt2);
      dirs[1] = Vector2d(0.0, -1.0);
      dirs[2] = Vector2d(1.0, 0.0);
    }
    grades[0].set_AVERAGE();
    grades[1].set_AVERAGE();
    grades[2].set_POOR();
    dribbles[0] = true;
    dribbles[1] = true;
    dribbles[2] = true;
  }
  else if (1.0*(FIELD_LENGTH/2 - my.pos.getX()) +
	   (FIELD_WIDTH/2 - my.pos.getY()) < targetingDist) {
    // 6. right front flank (Axel's "Leitplanke")
    numDirs = 3;
    dirs[0] = Vector2d(sqrt2, -sqrt2);
    dirs[1] = Vector2d(0.0, -1.0);
    dirs[2] = Vector2d(-1.0, 0.0);
    grades[0].set_PERFECT();
    grades[1].set_GOOD();
    grades[2].set_POOR();
    dribbles[0] = true;
    dribbles[1] = true;
    dribbles[2] = true;
  }
  else if (1.0*(FIELD_LENGTH/2 - my.pos.getX()) +
	   (FIELD_WIDTH/2 + my.pos.getY()) < targetingDist) {
    // 7. left front flank
    numDirs = 3;
    dirs[0] = Vector2d(sqrt2, sqrt2);
    dirs[1] = Vector2d(0.0, 1.0);
    dirs[2] = Vector2d(-1.0, 0.0);
    grades[0].set_PERFECT();
    grades[1].set_GOOD();
    grades[2].set_POOR();
    dribbles[0] = true;
    dribbles[1] = true;
    dribbles[2] = true;
  }
  else {
    // 8. somewhere else in the body of the field
    numDirs = 5;
    if (my.pos.getY() > 0) {
      dirs[0] = Vector2d(1.0, 0.0);
      dirs[1] = Vector2d(sqrt2, sqrt2);
      dirs[2] = Vector2d(sqrt2, -sqrt2);
      dirs[3] = Vector2d(0.0, 1.0);
      dirs[4] = Vector2d(0.0, -1.0);
    }
    else {
      dirs[0] = Vector2d(1.0, 0.0);
      dirs[1] = Vector2d(sqrt2, -sqrt2);
      dirs[2] = Vector2d(sqrt2, sqrt2);
      dirs[3] = Vector2d(0.0, -1.0);
      dirs[4] = Vector2d(0.0, 1.0);
    }
    grades[0].set_PERFECT();
    grades[1].set_PERFECT();
    grades[2].set_GOOD();
    grades[3].set_AVERAGE();
    grades[4].set_POOR();
    dribbles[0] = false;
    dribbles[1] = false;
    dribbles[2] = false;
    dribbles[3] = false;
    dribbles[4] = false;
  }

  double savedStamina = min (2*MAXPOWER, availableStamina);
  // keep back some stamina in case we lose the ball while dribbling, so we can
  // intercept the ball again within 2-3 steps.
  dribbleStamina = availableStamina - savedStamina;
  int nDashes = 0;
  bool dribbleonly = false; // fastdribble/dribble switch
  double dist = 0.0, dashPower = 0.0;
  Grade temp_g;
  temp_g.set_BAD();
  int temp_dir = 0;;
  for (int dir = 0; dir < numDirs; dir++) {
    // check if dribble direction orthogonal to a field side
    // when we are near this side (danger of dribbling out of the field!)
    if ((my.pos.getX() > FIELD_LENGTH/2 - distToBorder) &&
	(dirs[dir].getX() >= 0.0)) // front
      continue;
    else if ((my.pos.getX() < -FIELD_LENGTH/2 + distToBorder) &&
	     (dirs[dir].getX() <= 0.0)) // back
      continue;
    if ((my.pos.getY() > FIELD_WIDTH/2 - distToBorder) &&
	(dirs[dir].getY() >= 0.0)) // right
      continue;
    else if ((my.pos.getY() < -FIELD_WIDTH/2 + distToBorder) &&
	     (dirs[dir].getY() <= 0.0)) //left
      continue;

    DRAWTRACE(LINE(my.pos, my.pos + 2*dirs[dir]));

    if (dribbles[dir]) {
      // [!!] dribble corridor angle is wrong:
      // PI/4 only when fastdribbling (a cone, because opponents from the
      // sides or back wont reach the kicked ball anyway)
      // PI/2 when dribbling only (a rectangle, because opponents may attack us
      // directly fron the sides; maybe we need even some room to the back!)
      if (snap->otherTeamInCorridor(my.pos, my.pos + 3.0*dirs[dir], dribbleCorrW, PI_4) == 0) {
	// if corridor for dribbling is free of opponents
	if(grades[dir].is_better_than(temp_g)){
	  dribbleonly = true;
	  temp_g = grades[dir];
	  temp_dir = dir;
	}
      }
    }
    else {
      for (nDashes = 11; nDashes >= 7; nDashes -= 2) {
	// check several fastdribble distances 
	// (iterating number of dashes needed for intercept)
	dashPower = DASH_POWER_RATE*min(MAXPOWER, STAMINA_INC_MAX + dribbleStamina/nDashes);
	dist = dashPower/(1 - PLAYER_DECAY)*
	  (nDashes - PLAYER_DECAY*(1 - power(PLAYER_DECAY, nDashes))/(1 - PLAYER_DECAY));
	COUT("evalDribble: dist = " << dist);

	// check for opponents in dribble corridor
	if (snap->otherTeamInCorridor(my.pos, my.pos + 2.0*dist*dirs[dir], 2.0*dist, PI_4) == 0) {
	  if(grades[dir].is_better_than(temp_g)){
	    dribbleonly = false;
	    temp_g = grades[dir];
	    temp_dir = dir;
	    break;
	  }
	}
      }

      // [!!] dribble corridor angle is wrong here as well
      if (snap->otherTeamInCorridor(my.pos, my.pos + 3.0*dirs[dir], dribbleCorrW, PI_4) == 0) {
	grades[dir].downgrade();
	if(grades[dir].is_better_than(temp_g)){
	  dribbleonly = true;
	  temp_g = grades[dir];
	  temp_dir = dir;
	}
      }
    } 
  }

  if(!temp_g.is_BAD()) {
    
    // we have got a very nice and handsome dribble direction
    g = temp_g;
    dribbleDir = dirs[temp_dir].getAngle();
    action = dribbleonly ? PLAYER_DRIBBLE : PLAYER_FAST_DRIBBLE;
    
    if (dribbleonly) {
      COUT("eval: dribble slow");
      DRAWTRACE(CORRIDOR(my.pos, (my.pos + 3.0*dirs[temp_dir]), dribbleCorrW, PI_4));
    }
    
    else {
      COUT("eval: dribble fast");
      DRAWTRACE(CORRIDOR(my.pos, (my.pos + 2.0*dist*dirs[temp_dir]), 2.0*dist, PI_4));
      fd_dashPower = dashPower;
      fd_numSteps  = nDashes;
      if (fabs(my.bodyDir - dribbleDir) > DEG_TO_RAD(5.0))
	fd_numSteps -= 1; // to be safe, we can kick again either
    }
    
    // direction where to place ball when no opponent in sight
    // (if near a border, place to inside of field)
    double defaultdir = dribbleDir + (((ball.pos - my.pos).getAngle() - dribbleDir > 0) ? PI_3 : -PI_3);
    COUT("defdir = " << defaultdir);
    if (my.pos.getY() > FIELD_WIDTH/2 - sideBorder)
      defaultdir = -PI_2;
    else if (my.pos.getY() < -FIELD_WIDTH/2 + sideBorder)
      defaultdir = PI_2;
    else if (my.pos.getX() > FIELD_LENGTH/2 - sideBorder)
      defaultdir = PI;
    else if (my.pos.getX() < -FIELD_LENGTH/2 + sideBorder)
      defaultdir = 0;
    COUT("defdir = " << defaultdir);
    
    // place to defaultdir if firstDangerous is far away
    if ((tsnap->firstDangerousOpponent == UNKNOWN) ||
	(my.distTo(snap->otherTeam(tsnap->firstDangerousOpponent)) > 4.0*kickRange))
      holdDir = defaultdir;
    else {
      // also place to defaultdir if secondDangerous is also really dangerous
      // [!] place dir if we have two dangerous opponents is not ok
      if ((tsnap->secondDangerousOpponent != UNKNOWN) &&
	  (my.distTo(snap->otherTeam(tsnap->secondDangerousOpponent)) < 4.0*kickRange))
	holdDir = defaultdir;
      else {
	// place away from first dangerous opponent
	Angle opponentDir = (snap->otherTeam(tsnap->firstDangerousOpponent).pos - my.pos).getAngle();
	holdDir = dribbleDir + ((opponentDir - dribbleDir > 0) ? -PI_3 : PI_3);
      }
    }
    COUT("holddir = " << holdDir);
    return g;
  }

  // no direction for dribbling found
  action = PLAYER_PLACE;

  // place somewhere if firstDangerous is far away

  // [!!] firstDangerousOpponent is often unknown?
  // [!] placing happens quite rarely (is this good or bad? should we take mor time, like cmu99?)
  // [!] placing should not just be opposite side, but 
  // could even try to trick opponent (place to one orthogonal side of my->mytargetdir vector while opp is medium far away, when he gets too close, move ball around back to make him pass us on one side, and way is free...)

  if ((tsnap->firstDangerousOpponent == UNKNOWN) ||
      (my.distTo(snap->otherTeam(tsnap->firstDangerousOpponent)) 
       > 4.0*kickRange)) {
    placeDir = (ball.pos - my.pos).getAngle();
    // opponent far away, therefore just take some breath and keep ball where it is.
  } else {
    if ((tsnap->secondDangerousOpponent != UNKNOWN) &&
	(my.distTo(snap->otherTeam(tsnap->secondDangerousOpponent)) < 4.0*kickRange)) {
      placeDir = (ball.pos - my.pos).getAngle();
      // [!] find a better place pos if two opponents are close
    } else {
      Angle opponentDir = (snap->otherTeam(tsnap->firstDangerousOpponent).pos - my.pos).getAngle();
      // opposite side of closest opponent
      holdDir = PI + opponentDir;
    }
  }

  g.set_BAD();
  return g;
}

bool
BallHandlingModule::needToLookForBall( const Snapshot*   snap, 
				       TacticalSnapshot* ) {

  if (lookingForBall)
    return true;
  
  if ( snap->ball.coneTime > maxViewBallAge || snap->ball.age > maxBallAge ) 
    return true;
  
  return false;
}
