//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// MRB player module ballhandling
//
// $Id: ballhandlingmodule.h,v 1.50 2000/10/19 16:30:11 eri Exp $
// $manager: fel_und_eri$
//
// handles all situations in which a more or less responsible player
// or goalie has to go for the ball or to keep the ball to himself
//
//////////////////////////////////////////////////////////////////////


#ifndef __BALL_HANDLING_MODULE_H__
#define __BALL_HANDLING_MODULE_H__


#include "grade.h"
#include "drawMacros.h"

#include "worldmodel.h"
#include "tacticalmodule.h"
#include "audiomgr.h"
#include "AdvSkills.h"

/*
  
  short description:

  When called for feedback(), snap and tsnap are stored locally and some
  auxiliary variables (my, ball, etc) are set. A timestamp of snap and tsnap 
  is saved to be compared against when called later for eval() and act()

  In evaluate() we first test for some simple special cases, and then call the
  more complex evaluatePlayer() and evaluateGoalie() routines. They do the rest
  of the evaluation, and make use of the further singled out 
  dribble evaluation. At the end of the eval() method, the variables action and
  g (for grade) have to be set to the desired values to be returnd to 
  master control. Also take care to set any local temporary variables (e.g. 
  for values already computed for the later execution of the command). All
  this is implemented in ballhandlingEvaluate.C.

  In act() we split up the different cases in a switch{}. Auxiliary routines
  then do the rest of the action execution in most cases. All this is 
  implemented in ballhandlingAct.C. Some simulation routines and specific 
  turnHead routines (taking care of the players current action) are 
  included in this file as well.

  Things that need some more serious thinking are marked with [!] (or more
  !!s for urgency).

  Known bugs:

  - playerAttack() does not go in between the opponent and our goal, but 
    selects the opposite point!
  - goalie control is much too careless about leaving the goal, and cannot cope
    with two attacking players. Either solve this here soon, or leave the 
    goalieAttack() situation for the PositionModule. The strafing should be
    used more extensively (maybe even exclusively?) when an attacker is near.
    Turning 90 degrees and attacking directly is still always possible within
    2 simulator cycles.
  
  major ToDo:

  - goalie control rethinking
  - can we do the kickForward now (to trick a single approaching opponent), 
    or not?
  - is the modified fastDribble still robust enough?
  - personal decision of the person maintaining the code is needed: do we want
    closed-form computations (for efficiency) or simulations (for adaptability)
    of intercepts, attacks, opponent kicks, etc. ?
  - see the [!..!]s

  minor ToDo:

  - could fastDribble (or something like kickAhead()) not go into the skills?
  - should these turnHead-routines not go into the skills?
  - does it make sense to put strafing into the skills?
  - code hygiene: replace data store by something better, remove global kick, dash, turnParams etc.

  
*/


class BallHandlingModuleInfo; // forward decl. needed


/////////////////////////////////////////////////////////////////////
//
// Class Ballhandling Module
//
/////////////////////////////////////////////////////////////////////


class BallHandlingModule {
  
 public:
  
  BallHandlingModule(bool _goalie);
  
  ~BallHandlingModule();



  // feedback of communicated information
  bool feedbackWorld( const Snapshot* snap, 
		      TacticalSnapshot* tsnap );
  
  bool feedbackAudio( AudioMsg &audioMsg );
  
  

  // evaluation of given situation into a Grade
  Grade evaluate( AdvSkills* skills, 
		  const Snapshot* snap, 
		  TacticalSnapshot* tsnap);
  
  // perform action according to last evaluation
  bool act( AdvSkills* skills,
            const Snapshot* snap, 
	    TacticalSnapshot* tsnap);
  
  // pipe information to be communicated into the communication stream
  bool getSayMsg(AudioMsg &audioMsg);
  

  
 private:
  
  // action decision
  ///////////////////////////

  enum planned_action {
    
    WAIT = 0,            // do nothing 
    FIND_BALL,           // turn to get ball in sight
    
    GOALIE_INTERCEPT,    // go for a rolling ball
    GOALIE_ATTACK,       // attack an oponent with the ball
    GOALIE_CATCH,        // catch ball
    GOALIE_DRIBBLE,      // dribble to penalty area
    GOALIE_PLACE,        // keep ball away from opponent

    PLAYER_INTERCEPT,    // go for a rolling ball
    PLAYER_ATTACK,       // attack an opponent that got the ball
    PLAYER_FAST_DRIBBLE, // kick ball ahead a few steps
    PLAYER_DRIBBLE,      // keep ball close, but move
    PLAYER_PLACE,        // just keep ball away from opponent
  };

  // the decision made in evaluate(), to be executed in act()
  planned_action action;
  

  ///////////////////////////////////////////////////////////////////////////

    
  // local data storage
  ////////////////////////

  // the actual local data
  int         time;
  play_mode   playMode;
  WorldPlayer my;
  double      stamina;          // total
  double      availableStamina; // total - recovery threshold
  WorldBall   ball;
  double      distToBall;
  Angle       faceAngleToBall;
  Angle       bodyAngleToBall;
  double      lastDistToBall;
  WorldObject ourGoal;

  void initDataStorage  ();

  bool updateDataStorage( const Snapshot*   snap, 
			  TacticalSnapshot* tsnap);
  
  bool checkDataStorage ( const Snapshot*   snap, 
			  TacticalSnapshot* tsnap);

  // eval methods operating on internal data storage
  Grade evaluateGoalie( AdvSkills*        skills,
			const Snapshot*   snap, 
			TacticalSnapshot* tsnap );

  Grade evaluatePlayer( AdvSkills*        skills,
			const Snapshot*   snap, 
			TacticalSnapshot* tsnap );

  Grade evaluateDribble(const Snapshot* snap,
			TacticalSnapshot* tsnap ); 

  Grade goalieEvaluateDribble(const Snapshot* snap,
			TacticalSnapshot* tsnap ); 
  
  // internal: to compare snap and tsnap against when called to check
  int snapTime;
  int tsnapTime; 
  
  ///////////////////////////////////////////////////////////////////////////


  // pseudo constants
  /////////////////////////////////////
  int                     maxBallAge;
  int                     maxViewBallAge;
  int                     maxPlayerAge;
  double                  safeKickRange;
  double                  maxDribbleStamina;
  


  // are we a module of a goalie?
  /////////////////////////////////////
  bool                    goalie;   
  


  // communication
  /////////////////////////////////////
  bool                    sendSomething; // send flag
  BallHandlingModuleInfo* sendInfo;      // send info



  // findBall
  /////////////////////////////////////
  bool   lookingForBall; // flag set while still turning
  int    lastBallAge;    // latest age of ball seen
  Angle  turnAngle;      // angle by which to turn
  bool   needToLookForBall( const Snapshot*   snap, 
			    TacticalSnapshot* tsnap );
  // checks if ball is too old

  void   findBall         ( AdvSkills*        skills,
			    const Snapshot*   snap, 
			    TacticalSnapshot* tsnap );
  // turns to look for the ball



  // intercept & attack
  /////////////////////////////////////
  bool   simulateOpponentGoalkick( WorldPlayer simMy,
				   WorldPlayer simOpponent,
				   int &time, 
				   WorldBall &simBall,
				   double &leftStamina,
				   double safetyDist);
  // simulate intercept position of the ball when opponent does an optimal
  // goal kick
  
  double avgPower          ( double totalStamina,
			     int time );
  // averages power consumption

  void    playerIntercept  ( AdvSkills*        skills,
			     const Snapshot*   snap, 
			     TacticalSnapshot* tsnap );
  // intercepts the freely rolling ball at the computed intercept point

  void    goalieIntercept  ( AdvSkills*        skills,
			     const Snapshot*   snap, 
			     TacticalSnapshot* tsnap );
  // intercepts the freely rolling ball at the computed intercept point


  void    playerAttack ( AdvSkills*        skills,
			 const Snapshot*   snap, 
			 TacticalSnapshot* tsnap );
  // attacks an opponent holding the ball or going for the ball before us.
  // When 2 players are close, the first one goes for the opponent directly, 
  // the second moves into the opponent's way.

  AdvSkills::TurtleCommand goAroundOpponent( AdvSkills *skills,
					     Vector2d target,
					     WorldPlayer opponent,
					     double dist,
					     double stamina,
					     const Snapshot* snap );
  // close range attack to get around a blocking opponent

  void   goalieAttack      ( AdvSkills*        skills,
			     const Snapshot*   snap, 
			     TacticalSnapshot* tsnap );
  // highly questionable: does everything when opponent with the ball comes 
  // close to our goal.

  double setSafetyDistance ( const Snapshot*   snap );
  //the distance by which the goalie will stay away from the edge of the 
  // penalty area.
 

  // dribble
  /////////////////////////////////////
  double fd_dashPower;
  int    fd_numSteps;
  Angle dribbleDir; // for fast dribble and dribble
  Angle holdDir;    // for fast dribble and dribble
  Angle placeDir;   // for place
  Angle kickDir;    // for kickaway
  double dribbleStamina;

  void   fastdribble       ( AdvSkills*        skills,
			     const Snapshot*   snap, 
			     TacticalSnapshot* tsnap );
  // kick ball ahead slightly

  void   dribble           ( AdvSkills*        skills,
			     const Snapshot*   snap, 
			     TacticalSnapshot* tsnap );
  // run and keep ball under control at all times

  void   place             ( AdvSkills*        skills,
			     const Snapshot*   snap, 
			     TacticalSnapshot* tsnap );
  // move ball around ourselves away from opponent

  void   playerKickaway    ( AdvSkills*        skills,
			     const Snapshot*   snap, 
			     TacticalSnapshot* tsnap );
  

  
  // tactical state
  /////////////////////////////////////
  TacticalActionState guessedState; // what we guess about the tactical action
  int guessedPerformer;
  int guessedReceiver;
  
  // goalie catch
  /////////////////////////////////////
  void   goalieCatch      ( AdvSkills*        skills,
			    const Snapshot*   snap, 
			    TacticalSnapshot* tsnap );
  // catch the ball immediately
  
  
  // turnHead to ball pos after execution of the current command,
  // with tracing output
  /////////////////////////////////////
  bool turnHead           ( AdvSkills*         skills,
			    const Snapshot*   snap,
			    Angle             faceDir);
  bool turnHeadToBall     ( AdvSkills*        skills, 
			    const Snapshot*   snap);
  bool turnHeadToBall     ( AdvSkills*        skills, 
			    const Snapshot*   snap, 
			    AdvSkills::TurtleCommand     tc);
  bool turnHeadToBall     ( AdvSkills*        skills, 
			    const Snapshot*   snap, 
			    AdvSkills::TurnCommand       tc);
  bool turnHeadToBall     ( AdvSkills*        skills, 
			    const Snapshot*   snap, 
			    AdvSkills::KickCommand       kc);
  bool turnHeadForDribble ( AdvSkills*        skills, 
			    const Snapshot*   snap, 
			    AdvSkills::DribbleCommand    dc,
			    Angle             dribbleDir);
  bool reportTHState(AdvSkills *skills, const Snapshot *shot,
		     const AdvSkills::TurnHeadCommand &params);


  ///////////////////////////////////////////////////////////////////////////


  
  // temporary store for commands to check them before they are executed
  AdvSkills::CatchCommand    catchParams;
  AdvSkills::KickCommand     kickParams;
  AdvSkills::TurtleCommand   turtleParams;
  AdvSkills::TurnCommand     turnParams;
  AdvSkills::DribbleCommand  dribbleParams;
  AdvSkills::TurnHeadCommand turnHeadParams;



}; // class






//////////////////////////////////////////////////////////////////////
//
// Class to keep information for modules of other agents.
//
//////////////////////////////////////////////////////////////////////




class BallHandlingModuleInfo {

  // internal communication between ballhandlingModules of different players
  // of our team is not used yet.
  
 public:

  BallHandlingModuleInfo();
  ~BallHandlingModuleInfo();
  
}; // class


/** put information to AudioMsg-Stream */
class AudioMsg& operator<< (AudioMsg &aumsg, const BallHandlingModuleInfo &info);


/** fill with information from AudioMsg-Stream */
class AudioMsg& operator>> (AudioMsg &aumsg, BallHandlingModuleInfo &info);

#endif
