//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////
//
// MRB player module GoalKickModule
//
// \$Id: goalkickmodule.C,v 1.33 2000/10/19 15:27:53 goetz Exp $
// \$manager: $schappel
//
//////////////////////////////////////////////////////////////////////

#include "goalkickmodule.h"

////////////////////////////
//
// Constructor
//
///////////////////////////

GoalKickModule::GoalKickModule(bool isGoalie, GoalKickEval *_goalkickeval):
  goalie(isGoalie),
  goalkickeval(_goalkickeval),
  sendInfo(NULL),
  mustRekick(false),
  finishedKick(false),
  lastSnapshotTime(-1),       // module has never seen a snaphshot
  availableKickTime(0),
  goalTarget(Vector2d(FIELD_LENGTH /2, 0)),
  useSuperKick(false),
  useFixedKick(false),
  turnBeforeKick(false)
{
  if(goalkickeval)
    maxAge = goalkickeval->maxAge;
  else
    maxAge = 20;
};

////////////////////////////
//
// Destructor
//
///////////////////////////

GoalKickModule::~GoalKickModule()
{
};


////////////////////////////////////////////////////////////////////////////////////
//
// public methods
//
////////////////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////
//
// bool GoalKickModule::feedbackWorld
//
/////////////////////////////////////////

bool GoalKickModule::feedbackWorld(const Snapshot* worldSnap, TacticalSnapshot* tacticalSnap)

{
  // process new snapshot 
  updateDataStorage(worldSnap);

  // must the player report a goalkick?
  if(!tacticalSnap)
    cerr << "GOAL: feedbackWorld -> tacticalSnap not valid" << endl;
  else
    if(finishedKick)
      {
	GOALTRACE("feedbackWorld-> ther is a finished goal  kick");
	
	finishedKick=false;
	// did the player get a see-Information about the ball?
	if(worldSnap->ball.age == 0)
	  {
	    // will the ball stop behind the goalline?
	    if( (worldSnap->ball.pos + worldSnap->ball.v *(1/(1-BALL_DECAY))).getX() >= FIELD_LENGTH /2.)
	      {
		// calculate the parameter t, when the line from ball.pos to ball.pos + ball.v*(1/(1-BALL_DECAY))
		// intersects the goalline
		double t = FIELD_LENGTH /2. - worldSnap->ball.pos.getX();
		t /= worldSnap->ball.v.getX();
		// does the ball pass the goalline within the goal?
		if (fabs(worldSnap->ball.pos.getY() + worldSnap->ball.v.getY() * t ) < GOAL_WIDTH / 2.) 
		  // report a goalkick
		  {
		    tacticalSnap ->reportOurTacticalAction(TA_GOALKICK, me.number);
		    GOALTRACE("feedbackWorld -> goalkick reported");
		  }
	      }
	  }
		
      }
  
  return true;
}; 
  
//////////////////////////////////////////
//
// bool GoalKickModule::feedbackAudio
//
/////////////////////////////////////////

bool GoalKickModule::feedbackAudio(AudioMsg &audioMsg)

{
  GoalKickModuleInfo receivedInfo(false);
  audioMsg >> receivedInfo;
  if (receivedInfo.valid)
    {
      GOALTRACESHORT("got valid info");
      // evaluate
    }
  else
    {
      GOALTRACESHORT("got invalid info");
    }

  return true;
}; 

//////////////////////////////////////
//
// Grade GoalKickModule::evaluate
//
/////////////////////////////////////

Grade GoalKickModule::evaluate(AdvSkills* skills, 
			       const Snapshot* worldSnap, 
			       TacticalSnapshot* )
{
  Grade grade;

  delete (sendInfo);
  sendInfo = new GoalKickModuleInfo(false);
  
  // do we have a valid snapshot?
  if(worldSnap == NULL)
    {
      cerr << "GOAL: act -> snapshot is NULL" << endl;
      grade.set_IMPOSSIBLE();
      return grade;
    }

  // are the skills available?
  if(skills == NULL)
    {
      cerr << worldSnap->time << "GOAL: eval -> no skills available" << endl;
      grade.set_IMPOSSIBLE();
      return grade;
    }
 
  // if we have a newer snapshot, process it
  if (worldSnap->time > lastSnapshotTime) 
    { 
      updateDataStorage(worldSnap);
      GOALTRACE("eval -> worldsnapshot changed from feedback to evaluate");
    }  


  // the player is standing in his own half of the field, so goalkick is
  // impossible
  ///////////////////////////////////////////////////////////////////////

  if(me.pos.getX() < 0)
    {
      grade.set_IMPOSSIBLE();
      GOALTRACE("eval -> player is in his own half, grade = "
		<< grade);
      return grade;
    }

  // ball not in the kick range
  /////////////////////////////////
  if (!skills->kickSafe(worldSnap))
    {
      grade.set_IMPOSSIBLE();
      mustRekick = false;
      GOALTRACE ("eval -> ball not kickable; grade = "
	     << grade);
      return grade;
    }

  // do we need a rekick?
  ///////////////////////////////
  if (mustRekick)
    {
      grade.set_ABSOLUTE();
      GOALTRACE ("eval -> rekick must be done; grade = "
	     << grade);
      return grade;
    }

  // now we are able to kick the ball 
  ////////////////////////////////////////

  // calculate the minimum of the players distance to the left edge
  // the middle or the right edge of the goal for calculating a first
  // goal target

  goalkickeval->minDistToGoal(me.pos, goalTarget);


  // how many time does the player have for kicking ?
  availableKickTime = goalkickeval->availableTimeForKick(worldSnap, me.pos);
  GOALTRACE("eval -> available time for kick = " << availableKickTime);

  turnBeforeKick = false;

  // searching for the best corridor and evaluating it
  grade = searchAndEvalBestCorridor(worldSnap, skills, goalTarget, availableKickTime);

  return grade;
};

////////////////////////////
//
// GoalKickModule::act
//
///////////////////////////

bool GoalKickModule::act(AdvSkills* skills,
			 const Snapshot* worldSnap, TacticalSnapshot* tacticalSnap)
{
  // are the skills available?
  if(skills == NULL)
    {
      cerr << "GOAL: act -> no skills available" << endl;
      return false;
    }

  // do we have a valid snapshot?
  if(worldSnap == NULL)
    {
      cerr << "GOAL: act -> snapshot is NULL" << endl;
      skills->noAction();
      return true;
    }

  
  // if we have a newer snapshot, process it
  if (worldSnap->time > lastSnapshotTime) 
    { 
      updateDataStorage(worldSnap);
      GOALTRACE("act -> worldsnapshot changed from evaluate to act");
    }  

  // the act-routine was called also the ball was not kickable,
  // so do nothing
  if(!skills->kickSafe(worldSnap))
    {
      GOALTRACE (" act -> ball not kickable, send no Action");
      mustRekick = false;
      skills->noAction();
      return true;
    }

  // tell the tactical snapshot that player is doing a goalkick
  if(!tacticalSnap)
    // tactical snapshot is not valid => error message
    cerr << "GOAL: act -> tactical snapshot is NULL!" << endl;

  // player started a superkick and has not finished it yet
  if(mustRekick)
    {
      GOALTRACE ("act -> rekicking");
      kickParams = skills->rekick(worldSnap, kickParams);
      skills->exec(worldSnap, kickParams);
      if (kickParams.state == AdvSkills::KFINISHING)
	// superkick is finished after this kick
	{
	  mustRekick = false;
	  finishedKick = true;
	  GOALTRACE("act -> mustRekick has finished");
	}
      return true;
    }

  bool reallyKicked = true;

  if(turnBeforeKick)
    // player has enough time to turn his body before kicking
    // so consider carefully if this is really necessary
    {
      // testing if
      //  - angle between the players bodyDir and the ball is greater than PI_2,
      //  - the ball is visible,
      //  - the ball is still kickable in the next time step, and if
      //  - the ball will not collide with the player in th next time step
      // only if all these conditions are fullfilled, the player sends a turn
      // command to turn his boddy towards the ball
      
      double helpdist = dist(me.pos +me.v, 
			     worldSnap->ball.pos + worldSnap->ball.v);
      double angleToBall = me.bodyAngleTo(worldSnap->ball);
      
      if( (angleToBall > PI_2) 
	  && ( me.faceAngleTo(worldSnap->ball) < worldSnap->senseBody.viewWidth )
	  && ( helpdist < skills->kickRange )
	  && ( helpdist > PLAYER_SIZE + BALL_SIZE ) )
	// all conditions are fullfilled
	// so turn towards the ball
	{
	  reallyKicked = false;
	  GOALTRACE ("act -> turn to ball");
	  AdvSkills::TurnCommand turnParams
	    = skills->turn(worldSnap, angleToBall + me.bodyDir);
	  skills->exec(worldSnap, turnParams);
	}
      else
	// prepare a kick
	{
	  if(dist(me.pos, goalTarget) < GOAL_WIDTH /2)
	    // do not use too much time for a kick
	    availableKickTime = 2;
	  GOALTRACE ("act -> not turned, immediate kick with maxTime " << availableKickTime);
	  kickParams = skills->kick( worldSnap, BALL_SPEED_MAX, 
				     goalTarget,
				     availableKickTime);
	  skills->exec(worldSnap, kickParams);
	}
    }
  else
    {
      if(dist(me.pos, goalTarget) < GOAL_WIDTH /2)
	// do not use too much time for a kick
	availableKickTime = 2;

      // if ball can be very fast with just 2 kicks, do not perform a superkick
      if(availableKickTime > 2)
	{
	  kickParams = skills->kick(worldSnap, BALL_SPEED_MAX, 
				    goalTarget, 2);
	  if(kickParams.v.getLength() >= 2.0)
	    {
	      GOALTRACE ("act -> no superkick; ball is fast enough");
	      availableKickTime = 2;
	    }
	}

      GOALTRACE ("act -> kick with maxTime " << availableKickTime);
      kickParams = skills->kick( worldSnap, BALL_SPEED_MAX, 
				 goalTarget,
				 availableKickTime);
      skills->exec(worldSnap, kickParams);
    }
  
  if(reallyKicked)
    // player sends a kick command
    if(kickParams.state != AdvSkills::KFINISHING)
      {
	mustRekick = true;
	finishedKick = false;
      }
    else
      {
	finishedKick = true;
	GOALTRACE("act -> kick finished");
      }

  return true; 
};  


/////////////////////////////////
//
// GoalKickModule::getSayMsg
//
////////////////////////////////


bool GoalKickModule::getSayMsg(AudioMsg &audioMsg)
{
  audioMsg.markWrite();
  audioMsg << sendInfo;

  // check if error occurred
  if( !audioMsg.getState() )
    {
      cerr << "GOAL: Stream overflow: " << audioMsg << endl;
      audioMsg.rollbackWrite().clearState();
      
      audioMsg << false;
      
    }

  return true;
};


////////////////////////////////////////////////////////////////////////////////////
//
// private methods
//
////////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////
//
// GoalKickModule::updateDataStorage
//
////////////////////////////////////////

bool GoalKickModule::updateDataStorage(const Snapshot *worldSnap)
{
  // is the given snapshot ok?
  if(worldSnap == NULL)
    {
      cerr << "GOAL: updateDataStorage -> snapshot is NULL" << endl;
      return false;
    }

  // get new snapshot-time
  lastSnapshotTime = worldSnap->time;

  // the player himself
  me = worldSnap->my;

  return true;
}

///////////////////////////////////////////
//
// bool GoalKickModule::targetReachable
//
//////////////////////////////////////////

bool GoalKickModule::targetReachable(const Snapshot *worldSnap, AdvSkills *skills, 
				     Vector2d target, int availableTime)
{
  AdvSkills::ExtKickCommand kickParams;

  kickParams = skills->kick(worldSnap, BALL_SPEED_MAX, target, availableTime);
  
  if (kickParams.max_time <= availableTime)
    return true;
  else
    return false;
}



///////////////////////////////////////////////
//
// GoalKickModule::searchAndEvalBestCorridor
//
//////////////////////////////////////////////

Grade GoalKickModule::searchAndEvalBestCorridor(const Snapshot *worldSnap, 
						AdvSkills *skills, Vector2d &target,
						int availableTime)
{
  Grade grade;

  // setting the important positions in the goal

  Vector2d leftEdgeOfGoal;
  Vector2d middleOfGoal;
  Vector2d rightEdgeOfGoal;

  leftEdgeOfGoal = Vector2d(FIELD_LENGTH /2., -GOAL_WIDTH /2. + 1);
  middleOfGoal = Vector2d(FIELD_LENGTH / 2., 0);
  rightEdgeOfGoal = Vector2d(FIELD_LENGTH / 2., GOAL_WIDTH / 2. - 1);
      

  // now the player looks for the best corridor towards the goal
  // by first deciding which corridors(s) are possible by the given availableTime for
  // a kick

  bool useLeftCorridor = true;
  bool useMiddleCorridor = true;
  bool useRightCorridor = true;
  
  useLeftCorridor = targetReachable(worldSnap, skills, leftEdgeOfGoal, availableTime);
  useMiddleCorridor = targetReachable(worldSnap, skills, middleOfGoal, availableTime);
  useRightCorridor = targetReachable(worldSnap, skills, rightEdgeOfGoal, availableTime);

  if(!useLeftCorridor )
    {
      GOALTRACE("searchAndEval->left Edge not reachable; left corridor unusable");
    }
  if(!useMiddleCorridor )
    {
      GOALTRACE("searchAndEval->middle of goal not reachable; middle corridor unusable");
    }
  if(!useRightCorridor)
    {
      GOALTRACE("searchAndEval->right edge not reachable; right corridor unusable");
    }

  grade = goalkickeval->baseSearchAndEvalForCorridor(worldSnap, me.pos,
						     target, useLeftCorridor,
						     useMiddleCorridor,
						     useRightCorridor);
  return grade;
}



/////////////////////////////////////////////
//
// GoalKickModuleInfo::GoalKickModuleInfo
//
/////////////////////////////////////////////

GoalKickModuleInfo::GoalKickModuleInfo(bool _valid)
{
  valid = _valid;
}

/////////////////////////////////////////////
//
// GoalKickModuleInfo::~GoalKickModuleInfo
//
/////////////////////////////////////////////

GoalKickModuleInfo::~GoalKickModuleInfo()
{
}


/////////////////////////////////////////////
//
// AudioMsg &operator<<
//
/////////////////////////////////////////////

class AudioMsg &operator<< (AudioMsg &aumsg, const GoalKickModuleInfo &info)
{
  aumsg << info.valid;
  if (info.valid);
  // aumsg << //further data structures to write ;"
  return aumsg;
}


/////////////////////////////////////////////
//
// AudioMsg &operator>>
//
/////////////////////////////////////////////

class AudioMsg &operator>> (AudioMsg &aumsg, GoalKickModuleInfo &info)
{
  aumsg >> info.valid;
  if (info.valid);
  // aumsg >> // further data structures to read ;"
  return aumsg;
}
