//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#include "tacticalmodule.h"


// output after determining responsible players and most dangerous opponents
#undef RESPONSIBLETRACE
#define RESPONSIBLETRACE(print_command)

// output for simulation of opponent stamina
#undef STAMSIMTRACE
#define STAMSIMTRACE(print_command)

//////////////////////////////////////////////////////////////////////
//
// Class TacticalSystem
//
//////////////////////////////////////////////////////////////////////

TacticalSystem::TacticalSystem( int _numDef, int _numMid, int _numOff )
{
  ensure( _numDef >= 0 && _numMid >= 0 && _numOff >= 0 );
  ensure( _numDef + _numMid + _numOff == 10 );

  numDefenders = _numDef;
  numMidfield = _numMid;
  numOffenders = _numOff;
}


//////////////////////////////////////////////////////////////////////
//
// Class TacticalSnapshot
//
//////////////////////////////////////////////////////////////////////

TacticalSnapshot::TacticalSnapshot()
{
  time = 0;

  ourBallControl = BC_UNKNOWN;
  otherBallControl = BC_UNKNOWN;

  ourSituation = SI_UNKNOWN;
  otherSituation = SI_UNKNOWN;

  myTacticalNo = UNKNOWN;
  ourGoalieFound = false;
  otherGoalieFound = false;

  ourSystem.numDefenders   = Config::getValueAsInt( "ourDefenders" );
  ourSystem.numMidfield    = Config::getValueAsInt( "ourMidfield" );
  ourSystem.numOffenders   = Config::getValueAsInt( "ourOffenders" );

  otherSystem.numDefenders = Config::getValueAsInt( "otherDefenders" );
  otherSystem.numMidfield  = Config::getValueAsInt( "otherMidfield" );
  otherSystem.numOffenders = Config::getValueAsInt( "otherOffenders" );

  ourEdge   = Config::getValueAsInt( "tac_ourEdge" );
  otherEdge = Config::getValueAsInt( "tac_otherEdge" );

  firstResponsiblePlayer  = UNKNOWN;
  secondResponsiblePlayer = UNKNOWN;
  thirdResponsiblePlayer  = UNKNOWN;

  firstDangerousOpponent  = UNKNOWN;
  secondDangerousOpponent = UNKNOWN;
  thirdDangerousOpponent  = UNKNOWN;

  for( int i = 0; i < 11; i++ )
    {
      ourTeamRBTime[i]   = -1;
      otherTeamRBTime[i] = -1;

      ourTeamRBPos[i].setXY( 0, 0 );
      otherTeamRBPos[i].setXY( 0,0 );

      ourTeamTNo[i]   = UNKNOWN;
      otherTeamTNo[i] = UNKNOWN;

      ourTeamStam[i]   = STAMINA_MAX;
      otherTeamStam[i] = STAMINA_MAX;

      otherTeamSimStamPossible[i] = false;
    }
}


bool
TacticalSnapshot::reportOurTacticalAction( TacticalActionState state, int perPlayer, int recPlayer )
{
  bool newAction = false;

  switch( state )
    {
    case TA_CONTROL :
      if( perPlayer >= 1 && perPlayer <= 11 )
	{
	  newAction = true;
	  reportedTAction.performer = perPlayer;
	  reportedTAction.receiver = UNKNOWN;
	}
      break;

    case TA_PASS :
      if( perPlayer >= 1 && perPlayer <= 11 )
	if( recPlayer >= 1 && recPlayer <= 11 )
	{
	  newAction = true;
	  reportedTAction.performer = perPlayer;
	  reportedTAction.receiver = recPlayer;
	}
      break;
     
    case TA_GOALKICK :
      if( perPlayer >= 1 && perPlayer <= 11 )
	{
	  newAction = true;
	  reportedTAction.performer = perPlayer;
	  reportedTAction.receiver = UNKNOWN;
	}
      break;

    case TA_REACT :
      newAction = true;
      reportedTAction.performer = UNKNOWN;
      reportedTAction.receiver = UNKNOWN;
      break;

    case TA_UNKNOWN :
    default :
      break;
    }

  if( newAction )
    {
      reportedTAction.state = state;
      reportedTAction.quality = TQ_REPORTED;
      reportedTAction.startingTime = time;
      return true;
    }

  return false;
}


void
TacticalSnapshot::updateTacticalNumbers( Snapshot *snap )
{
  // try to determine goalie numbers

  if( !ourGoalieFound )
    if( snap->ourGoalieNo != UNKNOWN )
      {
	ourGoalieFound = true;

	// set tactical numbers of our team
	for( int i = 1; i <= 11; i++ )
	  {
	    if( i < snap->ourGoalieNo )
	      {
		ourTeamTNo[i-1] = i + 1;
	      }
	    else if( i > snap->ourGoalieNo )
	      {
		ourTeamTNo[i-1] = i;
	      }
	    else
	      {
		ourTeamTNo[i-1] = 1;
	      }
	  }
      }

  if( !otherGoalieFound )
    if( snap->otherGoalieNo != UNKNOWN )
      {
	otherGoalieFound = true;

	// set tactical numbers of other team
	for( int i = 1; i <= 11; i++ )
	  {
	    if( i < snap->otherGoalieNo )
	      {
		otherTeamTNo[i-1] = i + 1;
	      }
	    else if( i > snap->otherGoalieNo )
	      {
		otherTeamTNo[i-1] = i;
	      }
	    else
	      {
		otherTeamTNo[i-1] = 1;
	      }
	  }
      }

  // set my tactical number if it is still UNKNOWN
  if( myTacticalNo == UNKNOWN && ourGoalieFound && snap->my.number != UNKNOWN )
    {
      if( snap->my.number < snap->ourGoalieNo )
	{
	  myTacticalNo = snap->my.number + 1;
	}
      else if( snap->my.number > snap->ourGoalieNo )
	{
	  myTacticalNo = snap->my.number;
	}
      else
	{
	  myTacticalNo = 1;
	}
    }
}


bool
TacticalSnapshot::ourTeamDefender( int tn ) const
{
  ensure( tn >= 1 && tn <= 11 );
  if( tn < 1 || tn > 11 )
    return false;
  if( tn >= 2 && tn <= ourSystem.numDefenders + 1 )
    return true;
  return false;
}


bool
TacticalSnapshot::ourTeamMidfield( int tn ) const
{
  ensure( tn >= 1 && tn <= 11 );
  if( tn < 1 || tn > 11 )
    return false;
  if( tn > ourSystem.numDefenders + 1 &&
      tn <= ourSystem.numMidfield + ourSystem.numDefenders + 1 )
    return true;
  return false;
}


bool
TacticalSnapshot::ourTeamOffender( int tn ) const
{
  ensure( tn >= 1 && tn <= 11 );
  if( tn < 1 || tn > 11 )
    return false;
  if( tn > ourSystem.numMidfield + ourSystem.numDefenders + 1 )
    return true;
  return false;
}


bool
TacticalSnapshot::otherTeamDefender( int tn ) const
{
  ensure( tn >= 1 && tn <= 11 );
  if( tn < 1 || tn > 11 )
    return false;
  if( tn >= 2 && tn <= otherSystem.numDefenders + 1 )
    return true;
  return false;
}


bool
TacticalSnapshot::otherTeamMidfield( int tn ) const
{
  ensure( tn >= 1 && tn <= 11 );
  if( tn < 1 || tn > 11 )
    return false;
  if( tn > otherSystem.numDefenders + 1 &&
      tn <= otherSystem.numMidfield + otherSystem.numDefenders + 1 )
    return true;
  return false;
}


bool
TacticalSnapshot::otherTeamOffender( int tn ) const
{
  ensure( tn >= 1 && tn <= 11 );
  if( tn < 1 || tn > 11 )
    return false;
  if( tn > otherSystem.numMidfield + otherSystem.numDefenders + 1 )
    return true;
  return false;
}


int
TacticalSnapshot::ourTeamLeftNeighbour( int tn ) const
{
  ensure( tn >= 1 && tn <= 11 );

  // these players do not have left neighbours
  if( tn == 1 ||                             // goalie
      tn == 2 ||                             // left defender
      tn == 2 + ourSystem.numDefenders ||    // left midfield
      tn == 12 - ourSystem.numOffenders ||   // left offender
      tn < 1 ||                              // should not happen!
      tn > 11 )                              // should not happen!
    return UNKNOWN;

  for( int i = 1; i <= 11; i++ )
    if( ourTeamTacticalNo( i ) == tn - 1 )
      return i;
  
  return UNKNOWN;   // no neighbour found
}


int
TacticalSnapshot::ourTeamRightNeighbour( int tn ) const
{
  ensure( tn >= 1 && tn <= 11 );

  // these players do not have right neighbours
  if( tn == 1 ||                             // goalie
      tn == 1 + ourSystem.numDefenders ||    // right defender
      tn == 11 - ourSystem.numOffenders ||   // right midfield
      tn == 11 ||                            // right offender
      tn < 1 ||                              // should not happen!
      tn > 11 )                              // should not happen!
    return UNKNOWN;

  for( int i = 1; i <= 11; i++ )
    if( ourTeamTacticalNo( i ) == tn + 1 )
      return i;
  
  return UNKNOWN;   // no neighbour found
}


int
TacticalSnapshot::otherTeamLeftNeighbour( int tn ) const
{
  ensure( tn >= 1 && tn <= 11 );

  // these players do not have left neighbours
  if( tn == 1 ||                               // goalie
      tn == 2 ||                               // left defender
      tn == 2 + otherSystem.numDefenders ||    // left midfield
      tn == 12 - otherSystem.numOffenders ||   // left offender
      tn < 1 ||                                // should not happen!
      tn > 11 )                                // should not happen!
    return UNKNOWN;

  for( int i = 1; i <= 11; i++ )
    if( otherTeamTacticalNo( i ) == tn - 1 )
      return i;
  
  return UNKNOWN;   // no neighbour found
}


int
TacticalSnapshot::otherTeamRightNeighbour( int tn ) const
{
  ensure( tn >= 1 && tn <= 11 );

  // these players do not have right neighbours
  if( tn == 1 ||                               // goalie
      tn == 1 + otherSystem.numDefenders ||    // right defender
      tn == 11 - otherSystem.numOffenders ||   // right midfield
      tn == 11 ||                              // right offender
      tn < 1 ||                                // should not happen!
      tn > 11 )                                // should not happen!
    return UNKNOWN;

  for( int i = 1; i <= 11; i++ )
    if( otherTeamTacticalNo( i ) == tn + 1 )
      return i;
  
  return UNKNOWN;   // no neighbour found
}


void
TacticalSnapshot::determineResponsibleFriends( Snapshot *snap )
{
  WorldBall simBall = snap->ball;      // simulated ball
  int       simTime = 0;               // number of simulated steps

  double    useStamina[11];            // guessed stamina used for simulation
  double    myTeamDist[11];            // radius of players' movements
  double    myTeamV[11];               // players' velocity
  bool      sorted[11];                // has player "reached the ball"?
  int       numSimulatedPlayers = 0;   // number of players that have "reached the ball"
  bool      playerLeft = true;         // is there a player left to be sorted?

  // players' radius offset (at least kick range)
  double    interError = KICKRANGE;

  // if ball is close to a player we can not rely on its velocity
  // so we set it to zero
  for( int i = 1; i <= 11; i++ )
    {
      if( snap->ourTeam( i ).distTo( simBall ) < interError ||
	  snap->otherTeam( i ).distTo( simBall ) < interError )
	{
	  simBall.v.setXY( 0.0, 0.0 );
	  break;
	}
    }

  for( int i = 0; i < 11; i++ )
    {
      // use simulated stamina for team members if it is not too old
      // player's own stamina is well known
      if( i == snap->my.number - 1 )
	useStamina[i] = ourTeamStam[i].value - EFFORT_DEC_THR * STAMINA_MAX;
      else
	useStamina[i] = max( 5 * MAXPOWER,
			     ourTeamStam[i].value - EFFORT_DEC_THR * STAMINA_MAX -
			     ( MAXPOWER - STAMINA_INC_MAX ) * ourTeamStam[i].age );


      sorted[i] = false;
      myTeamDist[i] = 0.0;
      myTeamV[i] = 0.0;
    }

  // only find the (up to) three "most responsible" players
  while( playerLeft && numSimulatedPlayers < 3 && simTime < 50 )
    {
      playerLeft = false;
      for( int i = 0; i < 11; i++ )
	{
	  if( snap->ourTeam( i + 1 ).age <= MAXAGE && !sorted[i] && numSimulatedPlayers < 3 )
	    {
	      playerLeft = true;
	      // has player already reached the ball?
	      if( dist( snap->ourTeam( i + 1 ).pos, simBall.pos ) < myTeamDist[i] + interError )
		{
		  sorted[i] = true;
		  ourTeamRBTime[i] = simTime;
		  ourTeamRBPos[i] = simBall.pos;
		  switch( numSimulatedPlayers )
		    {
		    case 0  : firstResponsiblePlayer = i + 1; break;
		    case 1  : secondResponsiblePlayer = i + 1; break;
		    case 2  : thirdResponsiblePlayer = i + 1; break;
		    default : break;
		    }
		  numSimulatedPlayers++;
		}
	      else   // if not: simulate player's movement
		{
		  myTeamV[i] = min( PLAYER_SPEED_MAX,
				    ( STAMINA_INC_MAX +
				      max( 0.0, min( MAXPOWER - STAMINA_INC_MAX, useStamina[i] ) ) )
				    * DASH_POWER_RATE + myTeamV[i] * PLAYER_DECAY );
		  
		  useStamina[i] -= max( 0.0, min( MAXPOWER - STAMINA_INC_MAX, useStamina[i] ) );
		  useStamina[i] = max( 0.0, useStamina[i] );
		  myTeamDist[i] += myTeamV[i];
		}
	    }
	}

      simBall.simulate();
      simTime++;
    }

  // check responsible players for their age (maybe throw away old simulated information!)
  while( firstResponsiblePlayer != UNKNOWN && snap->ourTeam( firstResponsiblePlayer ).age > MAXAGE )
    {
      firstResponsiblePlayer = secondResponsiblePlayer;
      secondResponsiblePlayer = thirdResponsiblePlayer;
      thirdResponsiblePlayer = UNKNOWN;
    }
  while( secondResponsiblePlayer != UNKNOWN && snap->ourTeam( secondResponsiblePlayer ).age > MAXAGE )
    {
      secondResponsiblePlayer = thirdResponsiblePlayer;
      thirdResponsiblePlayer = UNKNOWN;      
    }
  if( thirdResponsiblePlayer != UNKNOWN && snap->ourTeam( thirdResponsiblePlayer ).age > MAXAGE )
    thirdResponsiblePlayer = UNKNOWN;

  // check for double players (mix of simulated and now calculated information)
  if( firstResponsiblePlayer == thirdResponsiblePlayer )
    thirdResponsiblePlayer = UNKNOWN;
  if( firstResponsiblePlayer == secondResponsiblePlayer )
    {
      secondResponsiblePlayer = thirdResponsiblePlayer;
      thirdResponsiblePlayer = UNKNOWN;
    }
  if( secondResponsiblePlayer == thirdResponsiblePlayer )
    thirdResponsiblePlayer = UNKNOWN;

  RESPONSIBLETRACE( cout << snap->time << " : Closest friends: " << firstResponsiblePlayer;
		    cout << " (" << ourTeamReachBallTime( firstResponsiblePlayer ) << ") ";
		    cout << secondResponsiblePlayer;
		    cout << " (" << ourTeamReachBallTime( secondResponsiblePlayer ) << ") ";
		    cout << thirdResponsiblePlayer;
		    cout << " (" << ourTeamReachBallTime( thirdResponsiblePlayer ) << ")" << endl; );
}


void
TacticalSnapshot::determineDangerousOpponents( Snapshot *snap )
{
  WorldBall simBall = snap->ball;      // simulated ball
  int       simTime = 0;               // number of simulated steps

  double    useStamina[11];            // players' simulated stamina
  double    otherTeamDist[11];         // radius of players' movements
  double    otherTeamV[11];            // players' velocity  
  bool      sorted[11];                // has player "reached the ball"?
  int       numSimulatedPlayers = 0;   // number of players that have "reached the ball"
  int       numValidPlayers = 0;       // number of players that are young enough
  bool      playerLeft = true;         // is there a player left to be sorted?

  // players' radius offset (at least kick range)
  double    interError = KICKRANGE;

  // if ball is close to a player we can not rely on its velocity
  // so we set it to zero
  for( int i = 1; i <= 11; i++ )
    {
      if( snap->ourTeam( i ).distTo( simBall ) < 2 * interError ||
	  snap->otherTeam( i ).distTo( simBall ) < 2 * interError )
	{
	  simBall.v.setXY( 0.0, 0.0 );
	  break;
	}
    }

  for( int i = 0; i < 11; i++ )
    {
      sorted[i] = false;
      useStamina[i] = STAMINA_MAX - EFFORT_DEC_THR * STAMINA_MAX;   // (almost) worst case scenario
      otherTeamDist[i] = 0.0;
      otherTeamV[i] = 0.0;
    }

  for( int i = 1; i <= 11; i++ )
    if( snap->otherTeam( i ).age <= MAXAGE )
      numValidPlayers++;

  // only find the (up to) three most dangerous opponents
  while( playerLeft && numSimulatedPlayers < 3 && simTime < 50 )
    {
      playerLeft = false;
      for( int i = 0; i < 11; i++ )
	{
	  if( snap->otherTeam( i + 1 ).age < MAXAGE && !sorted[i] && numSimulatedPlayers < 3 )
	    {
	      playerLeft = true;
	      // has player already reached the ball?
	      if( dist( snap->otherTeam( i + 1 ).pos, simBall.pos ) < otherTeamDist[i] + interError )
		{
		  sorted[i] = true;
		  otherTeamRBTime[i] = simTime;
		  otherTeamRBPos[i] = simBall.pos;
		  switch( numSimulatedPlayers )
		    {
		    case 0  : firstDangerousOpponent = i + 1; break;
		    case 1  : secondDangerousOpponent = i + 1; break;
		    case 2  : thirdDangerousOpponent = i + 1; break;
		    default : break;
		    }
		  numSimulatedPlayers++;
		}
	      else   // if not: simulate player's movement
		{
		  otherTeamV[i] = min( PLAYER_SPEED_MAX,
				       ( STAMINA_INC_MAX +
					 max( 0.0, min( MAXPOWER - STAMINA_INC_MAX, useStamina[i] ) ) )
				       * DASH_POWER_RATE + otherTeamV[i] * PLAYER_DECAY );
		  
		  useStamina[i] -= max( 0.0, min( MAXPOWER - STAMINA_INC_MAX, useStamina[i] ) );
		  useStamina[i] = max( 0.0, useStamina[i] );
		  otherTeamDist[i] += otherTeamV[i];
		}
	    }
	}

      simBall.simulate();
      simTime++;
    }

  // check responsible players for their age (maybe throw away old simulated information!)
  while( firstDangerousOpponent != UNKNOWN && snap->otherTeam( firstDangerousOpponent ).age > MAXAGE )
    {
      firstDangerousOpponent = secondDangerousOpponent;
      secondDangerousOpponent = thirdDangerousOpponent;
      thirdDangerousOpponent = UNKNOWN;
    }
  while( secondDangerousOpponent != UNKNOWN && snap->otherTeam( secondDangerousOpponent ).age > MAXAGE )
    {
      secondDangerousOpponent = thirdDangerousOpponent;
      thirdDangerousOpponent = UNKNOWN;      
    }
  if( thirdDangerousOpponent != UNKNOWN && snap->otherTeam( thirdDangerousOpponent ).age > MAXAGE )
    thirdDangerousOpponent = UNKNOWN;

  // check for double players (mix of simulated and now calculated information)
  if( firstDangerousOpponent == thirdDangerousOpponent )
    thirdDangerousOpponent = UNKNOWN;
  if( firstDangerousOpponent == secondDangerousOpponent )
    {
      secondDangerousOpponent = thirdDangerousOpponent;
      thirdDangerousOpponent = UNKNOWN;
    }
  if( secondDangerousOpponent == thirdDangerousOpponent )
    thirdDangerousOpponent = UNKNOWN;

  RESPONSIBLETRACE( cout << snap->time << " : Closest opponents: " << firstDangerousOpponent;
		    cout << " (" << otherTeamReachBallTime( firstDangerousOpponent ) << ") ";
		    cout << secondDangerousOpponent;
		    cout << " (" << otherTeamReachBallTime( secondDangerousOpponent ) << ") ";
		    cout << thirdDangerousOpponent;
		    cout << " (" << otherTeamReachBallTime( thirdDangerousOpponent ) << ")" << endl; );
}


bool
TacticalSnapshot::determineSituation( Snapshot *snap )
{
  bool useSitChannel = false;   // use situation channel for broadcast?

  // determine situation for both teams according to ball position and ball control (very simple)

  SituationState ourSituationNew = SI_UNKNOWN;
  SituationState otherSituationNew = SI_UNKNOWN;

  if( snap->ball.pos.getX() < -FIELD_LENGTH / 6.0 )
    {
      if( ourBallControl == BC_SAFE )
	{
	  ourSituationNew = SI_DEFENSIVE;
	  otherSituationNew = SI_MIDFIELD;
	}
      else
	{
	  ourSituationNew = SI_DEFENSIVE;
	  otherSituationNew = SI_OFFENSIVE;	  
	}
    }
  else if( snap->ball.pos.getX() < 0.0 )
    {
      if( ourBallControl == BC_SAFE )
	{
	  ourSituationNew = SI_MIDFIELD;
	  otherSituationNew = SI_MIDFIELD;
	}
      else
	{
	  ourSituationNew = SI_DEFENSIVE;
	  otherSituationNew = SI_MIDFIELD;
	}
    }
  else if( snap->ball.pos.getX() < FIELD_LENGTH / 6.0 )
    {
      if( ourBallControl == BC_SAFE )
	{
	  ourSituationNew = SI_MIDFIELD;
	  otherSituationNew = SI_DEFENSIVE;
	}
      else
	{
	  ourSituationNew = SI_MIDFIELD;
	  otherSituationNew = SI_MIDFIELD;
	}
    }
  else
    {
      if( ourBallControl == BC_SAFE )
	{      
	  ourSituationNew = SI_OFFENSIVE;
	  otherSituationNew = SI_DEFENSIVE;
	}
      else
	{
	  ourSituationNew = SI_MIDFIELD;
	  otherSituationNew = SI_DEFENSIVE;
	}
    }

  ourSituation = ourSituationNew;
  otherSituation = otherSituationNew;

  // decide if to use situation channel
  // firstResponsiblePlayer should not use that channel but emergency channel instead!

  switch( ourSituation )
    {
    case SI_UNKNOWN :
      // better do not send
      break;

    case SI_DEFENSIVE :
      // send if these conditions are fulfilled:
      // - I am not responsible but close to the ball
      //   (distance depends on if I am the goalie)
      // - I can see the ball
      if( ( snap->my.number != firstResponsiblePlayer ) &&
	  ( snap->my.distTo( snap->ball ) < ( ( snap->my.number == snap->ourGoalieNo ) ? 40.0 : 20.0 ) ) &&
	  ( snap->ball.age <= 4 ) &&
	  ( fabs( snap->my.faceAngleTo( snap->ball ) ) < snap->senseBody.getViewAngle() ) )
	{
	  useSitChannel = true;
	}
      break;

    case SI_MIDFIELD :
    case SI_OFFENSIVE :
      // send if these conditions are fulfilled:
      // - I am not responsible but close to the ball
      // - I am a little behind the most responsible player
      // - I can see the ball
      if( ( snap->my.number != firstResponsiblePlayer ) &&
	  ( snap->my.distTo( snap->ball ) < 20.0 ) &&
	  ( snap->ball.age <= 4 ) &&
	  ( fabs( snap->my.faceAngleTo( snap->ball ) ) < snap->senseBody.getViewAngle() ) &&
	  ( firstResponsiblePlayer != UNKNOWN ) &&
	  ( snap->my.pos.getX() < snap->ourTeam( firstResponsiblePlayer ).pos.getX() - 5.0 ) )
	{
	  useSitChannel = true;
	}
      break;

    default :
      break;
    }

  return useSitChannel;
}


void
TacticalSnapshot::determineBallControl( Snapshot *snap,
					BallControlState ourBallControlAudio,
					BallControlState otherBallControlAudio )
{
  BallControlState ourBallControlNew = BC_UNKNOWN;     // new value for ourBallControl
  BallControlState otherBallControlNew = BC_UNKNOWN;   // new value for otherBallControl

  int firstFriendReachTime = -1;     // time when firstResponsiblePlayer will reach the ball
  int firstOpponentReachTime = -1;   // time when firstDangerousOpponent will reach the ball

  switch( snap->playMode )
    {
    case PLAY_MODE_UNKNOWN :
    case PLAY_MODE_BEFORE_KICK_OFF :
    case PLAY_MODE_TIME_OVER :
    case PLAY_MODE_HALF_TIME :
    case PLAY_MODE_TIME_EXTENDED :
      
      // there is no game running
      ourBallControlNew = BC_UNKNOWN;
      otherBallControlNew = BC_UNKNOWN;
      break;

    case PLAY_MODE_KICK_OFF_US :
    case PLAY_MODE_KICK_IN_US :
    case PLAY_MODE_FREE_KICK_US :
    case PLAY_MODE_CORNER_KICK_US :
    case PLAY_MODE_GOAL_KICK_US :
    case PLAY_MODE_GOAL_THEM :
    case PLAY_MODE_OFFSIDE_THEM :

      // in a standard situation for us ball control is obvious
      ourBallControlNew = BC_SAFE;
      otherBallControlNew = BC_LOST;
      break;

    case PLAY_MODE_KICK_OFF_THEM :
    case PLAY_MODE_KICK_IN_THEM :
    case PLAY_MODE_FREE_KICK_THEM :
    case PLAY_MODE_CORNER_KICK_THEM :
    case PLAY_MODE_GOAL_KICK_THEM :
    case PLAY_MODE_GOAL_US :
    case PLAY_MODE_OFFSIDE_US :

      // in a standard situation for the opponent team ball control is obvious
      ourBallControlNew = BC_LOST;
      otherBallControlNew = BC_SAFE;
      break;

    case PLAY_MODE_DROP_BALL :
    case PLAY_MODE_PLAY_ON :

      // is there a player of our team? (there should be at least one...)
      if( firstResponsiblePlayer != UNKNOWN )
	{
	  firstFriendReachTime = ourTeamReachBallTime( firstResponsiblePlayer );

	  // is there a player of the opponent team?
	  if( firstDangerousOpponent != UNKNOWN )
	    {
	      firstOpponentReachTime = otherTeamReachBallTime( firstDangerousOpponent );

	      // opponent is faster
	      if( firstOpponentReachTime + otherEdge < firstFriendReachTime )
		{
		  ourBallControlNew = ( ourBallControl == BC_SAFE ? BC_DANGER : BC_LOST );
		  otherBallControlNew = ( ourBallControl == BC_SAFE ? BC_DANGER : BC_SAFE );
		}
	      // we are faster
 	      else if( firstFriendReachTime + ourEdge < firstOpponentReachTime )
		{
		  ourBallControlNew = ( ourBallControl == BC_LOST ? BC_DANGER : BC_SAFE );
		  otherBallControlNew = ( ourBallControl == BC_LOST ? BC_DANGER : BC_LOST );
		}
	      // dangerous situation
	      else
		{
		  ourBallControlNew = BC_DANGER;
		  otherBallControlNew = BC_DANGER;
		}
	    }
	  else   // no opponent seen
	    {
	      ourBallControlNew = ( ourBallControl == BC_LOST ? BC_DANGER : BC_SAFE);
	      otherBallControlNew = ( ourBallControl == BC_LOST ? BC_DANGER : BC_LOST);
	    }
	}
      else   // weird case: no player of our team found...
	{
	  // is there at least an opponent?
	  if( firstDangerousOpponent != UNKNOWN )
	    {
	      firstOpponentReachTime = otherTeamReachBallTime( firstDangerousOpponent );
	  
	      ourBallControlNew = ( ourBallControl == BC_SAFE ? BC_DANGER : BC_LOST);
	      otherBallControlNew = ( ourBallControl == BC_SAFE ? BC_DANGER : BC_SAFE);
	    }
	  else   // there is NOBODY around... try to use friend's ball control values
	    {
	      ourBallControlNew = ourBallControlAudio;
	      otherBallControlNew = otherBallControlAudio;
	    }
	}
      break;

    default :

      ensure( 0 );
      break;
    }

  ourBallControl = ourBallControlNew;
  otherBallControl = otherBallControlNew;
}


bool
TacticalSnapshot::determineOurTacticalAction( Snapshot *snap, const TacticalAction& audioAction, int currentTime )
{
  bool useEmChannel = false;   // use emergency channel if agent reported an action
  bool useHistory = false;     // are tactical history and snap consistent?

  int actionAge = currentTime - ourAction.startingTime;

  TacticalAction ourActionNew( TA_UNKNOWN, UNKNOWN, UNKNOWN, currentTime );

  switch( snap->playMode )
    {
    case PLAY_MODE_UNKNOWN :
    case PLAY_MODE_BEFORE_KICK_OFF :
    case PLAY_MODE_TIME_OVER :
    case PLAY_MODE_HALF_TIME :
    case PLAY_MODE_TIME_EXTENDED :
      
      ourActionNew.quality = TQ_GUESSED;
      break;

    case PLAY_MODE_KICK_OFF_US :
    case PLAY_MODE_KICK_IN_US :
    case PLAY_MODE_FREE_KICK_US :
    case PLAY_MODE_CORNER_KICK_US :
    case PLAY_MODE_GOAL_KICK_US :
    case PLAY_MODE_GOAL_THEM :
    case PLAY_MODE_OFFSIDE_THEM :

      // in a standard situation for us the state of our tactical action is obvious
      ourActionNew.state = TA_CONTROL;

      // take reported information into account
      if( reportedTAction.quality   == TQ_REPORTED &&
	  reportedTAction.state     == TA_CONTROL &&
	  reportedTAction.performer != UNKNOWN )
	{
	  useEmChannel           = true;
	  ourActionNew.quality   = TQ_REPORTED;
	  ourActionNew.performer = reportedTAction.performer;
	  break;
	}

      // take audial information into account
      if( audioAction.state == TA_CONTROL && audioAction.performer != UNKNOWN )
	{
	  // if sending player is sure about what he sent... believe him
	  if( audioAction.quality == TQ_REPORTED )
	    {
	      ourActionNew.quality   = TQ_AUDIAL_REPORTED;
	      ourActionNew.performer = audioAction.performer;
	      break;
	    }
	  // if he is not so sure compare it with our results
	  if( audioAction.quality != TQ_UNKNOWN )
	    {
	      if( audioAction.performer == firstResponsiblePlayer ||
		  audioAction.performer == secondResponsiblePlayer ||
		  audioAction.performer == thirdResponsiblePlayer )
		{
		  ourActionNew.quality   = TQ_AUDIAL;
		  ourActionNew.performer = audioAction.performer;
		  break;
		}
	    }
	}

      // try to determine performer by history
      if( ourAction.state     == TA_CONTROL &&
	  ourAction.quality   != TQ_UNKNOWN &&
	  ourAction.performer != UNKNOWN )
	{
	  // if old performer fits into new worldmodel or is a reported one keep him
	  if( ourAction.performer == firstResponsiblePlayer ||
	      ourAction.performer == secondResponsiblePlayer ||
	      ourAction.performer == thirdResponsiblePlayer ||
	      ourAction.quality   == TQ_REPORTED ||
	      ourAction.quality   == TQ_AUDIAL_REPORTED )
	    {
	      ourActionNew.quality   = ourAction.quality;
	      ourActionNew.performer = ourAction.performer;
	      break;
	    }
	}

      // if all else fails, take firstResponsiblePlayer as performer
      ourActionNew.quality   = TQ_GUESSED;
      ourActionNew.performer = firstResponsiblePlayer;
      break;

    case PLAY_MODE_KICK_OFF_THEM :
    case PLAY_MODE_KICK_IN_THEM :
    case PLAY_MODE_FREE_KICK_THEM :
    case PLAY_MODE_CORNER_KICK_THEM :
    case PLAY_MODE_GOAL_KICK_THEM :
    case PLAY_MODE_GOAL_US :
    case PLAY_MODE_OFFSIDE_US :

      // in a standard situation for the opponent team our tactical action is obvious
      ourActionNew.state   = TA_REACT;
      ourActionNew.quality = TQ_GUESSED;
      break;

    case PLAY_MODE_DROP_BALL :
    case PLAY_MODE_PLAY_ON :

      // CAUTION: very simple implementation!

      // first take reported information into account
      if( reportedTAction.quality == TQ_REPORTED && reportedTAction.state != TA_UNKNOWN )
	{
	  useEmChannel           = true;
	  ourActionNew.quality   = TQ_REPORTED;
	  ourActionNew.state     = reportedTAction.state;
	  ourActionNew.performer = reportedTAction.performer;
	  ourActionNew.receiver  = reportedTAction.receiver;
	  break;
	}

      // if there is no reported information take audio information into account
      if( audioAction.state != TA_UNKNOWN )
	if( audioAction.quality == TQ_REPORTED || audioAction.quality == TQ_AUDIAL_REPORTED )
	  // do not overwrite better information
	  if( ( ourAction.quality != TQ_REPORTED && ourAction.quality != TQ_AUDIAL_REPORTED ) ||
	      audioAction.startingTime > ourAction.startingTime )
	    {
	      ourActionNew.quality   = ( audioAction.quality == TQ_REPORTED ? TQ_AUDIAL_REPORTED : TQ_AUDIAL );
	      ourActionNew.state     = audioAction.state;
	      ourActionNew.performer = audioAction.performer;
	      ourActionNew.receiver  = audioAction.receiver;
	      break;
	    }

      // if there is no good audio information try to use history
      if( ourAction.quality != TQ_UNKNOWN )
	{
	  switch( ourAction.state )
	    {
	    case TA_CONTROL :
	      // keep old action if ball is not lost
	      if( ourBallControl != BC_LOST &&
		  ( ourAction.performer == firstResponsiblePlayer ||
		    ourAction.performer == secondResponsiblePlayer ||
		    ourAction.performer == thirdResponsiblePlayer ||
		    ourAction.quality == TQ_REPORTED ||
		    ourAction.quality == TQ_AUDIAL_REPORTED ) )
		{
		  useHistory = true;
		}
	      break;
	      
	    case TA_PASS :
	      // keep old action if ball is not lost...
	      if( ( ourBallControl != BC_LOST ||
		    firstDangerousOpponent == UNKNOWN ||
		    otherTeamReachBallTime( firstDangerousOpponent ) > 2 ) &&
		  // ...and if no friend has it under control
		  ( // ourBallControl != BC_SAFE ||   // does not make sense
		    firstResponsiblePlayer == ourAction.performer ||
		    // firstResponsiblePlayer == snap->my.number ||   // does not make sense
		    firstResponsiblePlayer == UNKNOWN ||
		    ourTeamReachBallTime( firstResponsiblePlayer ) > 2 ) )
		{
		  useHistory = true;
		}
	      break;

	    case TA_GOALKICK :
	      // keep old action if ball is not lost
	      if( ourBallControl != BC_LOST ||
		  firstDangerousOpponent == UNKNOWN ||
		  otherTeamReachBallTime( firstDangerousOpponent ) > 2 )
		{
		  useHistory = true;
		}
	      break;

	    case TA_REACT :
	      // keep old action if ball control is not safe again
	      if( ourBallControl != BC_SAFE )
		{
		  useHistory = true;
		}
	      break;

	    case TA_UNKNOWN :
	    default :
	      break;
	    }
	}
      
      // use history if it is consistent with world and not too old
      if( useHistory &&
	  ( ourAction.state == TA_CONTROL  ? actionAge < 20 :
	    ( ourAction.state == TA_PASS     ? actionAge < 10 :
	      ( ourAction.state == TA_GOALKICK ? actionAge < 15 :   // 10 ?
		( ourAction.state == TA_REACT    ? actionAge < 20 : false ) ) ) ) )
	{
	  ourActionNew = ourAction;
	  break;
	}

      // history and world are not consistent -> try to determine tactical action
      // by looking at the world

      switch( ourBallControl )
	{
	case BC_SAFE :
	  if( firstResponsiblePlayer == UNKNOWN )
	    ourActionNew.state = TA_REACT;
	  else
	    {
	      ourActionNew.state = TA_CONTROL;
	      ourActionNew.performer = firstResponsiblePlayer;
	    }
	  break;

	case BC_DANGER :
	  // use history in that case
	  if( ourAction.state == TA_REACT || firstResponsiblePlayer == UNKNOWN )
	    ourActionNew.state = TA_REACT;
	  else
	    {
	      ourActionNew.state = TA_CONTROL;
	      ourActionNew.performer = firstResponsiblePlayer;
	    }
	  break;

	case BC_LOST :
	  ourActionNew.state = TA_REACT;
	  break;

	case BC_UNKNOWN :
	default :
	  ourActionNew.state = TA_REACT;
	  break;
	}

      ourActionNew.quality = TQ_GUESSED;
      break;

    default :

      ensure( 0 );
      break;
    }

  // if old and new action for our team are the same, don't change startingTime
  if( ourAction.state == ourActionNew.state &&
      ourAction.performer == ourActionNew.performer &&
      ourAction.receiver == ourActionNew.receiver )
    {
      ourActionNew.startingTime = ourAction.startingTime;
    }
  else
    {
      ourActionNew.startingTime = currentTime;
    }
  ourAction = ourActionNew;

  // if action is very old, downgrade its quality
  if( ourAction.startingTime < currentTime - 10 )
    {
      if( ourAction.quality == TQ_REPORTED )
	ourAction.quality = TQ_GUESSED;
      if( ourAction.quality == TQ_AUDIAL_REPORTED )
	ourAction.quality = TQ_AUDIAL;
    }

  return useEmChannel;
}


void
TacticalSnapshot::determineOtherTacticalAction( Snapshot *snap, const TacticalAction& audioAction, int currentTime )
{
  TacticalAction otherActionNew( TA_UNKNOWN, UNKNOWN, UNKNOWN, currentTime );

  switch( snap->playMode )
    {
    case PLAY_MODE_UNKNOWN :
    case PLAY_MODE_BEFORE_KICK_OFF :
    case PLAY_MODE_TIME_OVER :
    case PLAY_MODE_HALF_TIME :
    case PLAY_MODE_TIME_EXTENDED :
      
      otherActionNew.quality = TQ_GUESSED;
      break;

    case PLAY_MODE_KICK_OFF_US :
    case PLAY_MODE_KICK_IN_US :
    case PLAY_MODE_FREE_KICK_US :
    case PLAY_MODE_CORNER_KICK_US :
    case PLAY_MODE_GOAL_KICK_US :
    case PLAY_MODE_GOAL_THEM :
    case PLAY_MODE_OFFSIDE_THEM :

      // in a standard situation for us opponent's tactical action is obvious
      otherActionNew.state   = TA_REACT;
      otherActionNew.quality = TQ_GUESSED;
      break;

    case PLAY_MODE_KICK_OFF_THEM :
    case PLAY_MODE_KICK_IN_THEM :
    case PLAY_MODE_FREE_KICK_THEM :
    case PLAY_MODE_CORNER_KICK_THEM :
    case PLAY_MODE_GOAL_KICK_THEM :
    case PLAY_MODE_GOAL_US :
    case PLAY_MODE_OFFSIDE_US :

      // in a standard situation for the opponent team the state of its tactical action is obvious
      otherActionNew.state = TA_CONTROL;

      // take audial information into account
      if( audioAction.state == TA_CONTROL &&
	  audioAction.performer != UNKNOWN && 
	  audioAction.quality != TQ_UNKNOWN )
	{
	  if( audioAction.performer == firstDangerousOpponent ||
	      audioAction.performer == secondDangerousOpponent ||
	      audioAction.performer == thirdDangerousOpponent )
	    {
	      otherActionNew.quality = TQ_AUDIAL;
	      otherActionNew.performer = audioAction.performer;
	      break;
	    }
	}

      // try to determine performer by history
      if( otherAction.state == TA_CONTROL &&
	  otherAction.quality != TQ_UNKNOWN &&
	  otherAction.performer != UNKNOWN )
	{
	  // if old performer fits into new worldmodel keep him
	  if( otherAction.performer == firstDangerousOpponent ||
	      otherAction.performer == secondDangerousOpponent ||
	      otherAction.performer == thirdDangerousOpponent )
	    {
	      otherActionNew.quality = otherAction.quality;
	      otherActionNew.performer = otherAction.performer;
	      break;
	    }
	}

      // if all else fails, take firstResponsiblePlayer as performer
      otherActionNew.quality = TQ_GUESSED;
      otherActionNew.performer = firstDangerousOpponent;
      break;

    case PLAY_MODE_DROP_BALL :
    case PLAY_MODE_PLAY_ON :

      // just keep old information
      otherActionNew = otherAction;
      break;

    default :

      ensure( 0 );
      break;
    }

  // if old and new action for other team are the same, don't change startingTime
  if( otherAction.state == otherActionNew.state &&
      otherAction.performer == otherActionNew.performer &&
      otherAction.receiver == otherActionNew.receiver )
    {
      otherActionNew.startingTime = otherAction.startingTime;
    }
  else
    {
      otherActionNew.startingTime = currentTime;
    }
  otherAction = otherActionNew;
}


void
TacticalSnapshot::determineOtherTacticalSystem( Snapshot *snap )
{
  int group[11];         // tactical group players belong to (0,1,2,3)
  int groupSize[11];     // sizes of the (up to 11) tactical groups

  int defGroup = 0;
  int midGroup = 0;
  int offGroup = 0;

  for( int i = 0; i < 11; i++ )
    {
      group[i] = 0;
      groupSize[i] = 0;
    }

  int numGroups = 0;     // number of tactical groups

  double dist[11][11];   // x-distances between players of the opponent team

  for( int i = 0; i < 11; i++ )
    for( int j = i + 1; j < 11; j++ )
      dist[i][j] = fabs( snap->otherTeam( i + 1 ).pos.getX() - snap->otherTeam( j + 1 ).pos.getX() );

  int iMin = 0, jMin = 1;
  bool searchOn = true;
  bool distFound = false;

  while( searchOn )
    {
      searchOn = false;
      distFound = false;
      iMin = 0;
      jMin = 1;

      for( int i = 0; i < 11; i++ )
	if( i != snap->otherGoalieNo - 1 )
	  for( int j = i + 1; j < 11; j++ )
	    if( j != snap->otherGoalieNo - 1 )
	      if( ( !group[i] && group[j] ) ||
		  ( !group[j] && group[i] ) ||
		  ( !group[i] && !group[j] && numGroups < 3 ) )
		if( !distFound || dist[i][j] < dist[iMin][jMin] )
		  {
		    distFound = true;
		    iMin = i;
		    jMin = j;
		  }
      
      if( distFound )
	{
	  if( !group[iMin] && !group[jMin] )
	    {
	      numGroups++;
	      group[iMin] = group[jMin] = numGroups;
	      groupSize[numGroups] += 2;
	      searchOn = true;
	    }
	  else if( !group[iMin] )
	    {
	      group[iMin] = group[jMin];
	      groupSize[group[iMin]]++;
	      searchOn = true;
	    }
	  else if( !group[jMin] )
	    {
	      group[jMin] = group[iMin];
	      groupSize[group[iMin]]++;
	      searchOn = true;
	    }
	}
    }

  int groupPlayer[11];

  for( int i = 1; i <= numGroups; i++ )
    for( int j = 1; j <= 11; j++ )
      if( group[j] == i )
	groupPlayer[i] = j;

  if( numGroups < 3 )
    return;

  offGroup =
    snap->otherTeam( groupPlayer[1] ).pos.getX() > snap->otherTeam( groupPlayer[2] ).pos.getX() &&
    snap->otherTeam( groupPlayer[1] ).pos.getX() > snap->otherTeam( groupPlayer[3] ).pos.getX()
    ? 1 :
    snap->otherTeam( groupPlayer[2] ).pos.getX() > snap->otherTeam( groupPlayer[1] ).pos.getX() &&
    snap->otherTeam( groupPlayer[2] ).pos.getX() > snap->otherTeam( groupPlayer[3] ).pos.getX()
    ? 2 : 3;

  defGroup =
    snap->otherTeam( groupPlayer[1] ).pos.getX() < snap->otherTeam( groupPlayer[2] ).pos.getX() &&
    snap->otherTeam( groupPlayer[1] ).pos.getX() < snap->otherTeam( groupPlayer[3] ).pos.getX()
    ? 1 :
    snap->otherTeam( groupPlayer[2] ).pos.getX() < snap->otherTeam( groupPlayer[1] ).pos.getX() &&
    snap->otherTeam( groupPlayer[2] ).pos.getX() < snap->otherTeam( groupPlayer[3] ).pos.getX()
    ? 2 : 3;

  midGroup =
    ( defGroup == 1 && offGroup == 3 ) ? 2 :
    ( defGroup == 3 && offGroup == 1 ) ? 2 :
    ( defGroup == 1 && offGroup == 2 ) ? 3 :
    ( defGroup == 2 && offGroup == 1 ) ? 3 : 1;

  cerr << "Found " << numGroups << " groups: ";
  cerr << "[" << groupSize[defGroup] << "] [" << groupSize[midGroup] << "] [" << groupSize[offGroup] << "]";
  cerr << endl;
}


//////////////////////////////////////////////////////////////////////
//
// Class TacticalModule
//
//////////////////////////////////////////////////////////////////////

TacticalModule::TacticalModule( int _numSnapshots )
{
  numSnapshots = _numSnapshots;
  ensure( numSnapshots <= 30 );   // just to avoid out of memory (30 IS enough!)

  snapshots = new TacticalSnapshot[numSnapshots];
  ensure( snapshots );
  
  currentTime = 0;

  useInfoChannel = false;
  useEmergencyChannel = false;
}


TacticalModule::TacticalModule( const TacticalModule& tm )
{
  numSnapshots = tm.numSnapshots;

  snapshots = new TacticalSnapshot[numSnapshots];
  ensure( snapshots );
  
  for( int i = 0; i < numSnapshots; i++ )
    snapshots[i] = tm.snapshots[i];

  currentTime = tm.currentTime;
  
  useInfoChannel = tm.useInfoChannel;
  useEmergencyChannel = tm.useEmergencyChannel;
}


TacticalModule& 
TacticalModule::operator = ( const TacticalModule& tm )
{
  numSnapshots = tm.numSnapshots;

  delete [] snapshots;

  snapshots = new TacticalSnapshot[numSnapshots];
  ensure( snapshots );

  for( int i = 0; i < numSnapshots; i++ )
    snapshots[i] = tm.snapshots[i];

  currentTime = tm.currentTime;

  useInfoChannel = tm.useInfoChannel;
  useEmergencyChannel = tm.useEmergencyChannel;

  return *this;
}


TacticalModule::~TacticalModule()
{
  delete [] snapshots;
}


void
TacticalModule::simulate( int time )
{
  TacticalSnapshot &oldSnap = snapshots[getTimeIndex( time )];
  TacticalSnapshot &newSnap = snapshots[getTimeIndex( time + 1 )];

  newSnap.time = oldSnap.time + 1;
  
  // copy old information

  newSnap.ourBallControl = oldSnap.ourBallControl;
  newSnap.otherBallControl = oldSnap.otherBallControl;

  newSnap.ourSituation = oldSnap.ourSituation;
  newSnap.otherSituation = oldSnap.otherSituation;

  newSnap.ourAction = oldSnap.ourAction;
  newSnap.otherAction = oldSnap.otherAction;

  newSnap.ourSystem = oldSnap.ourSystem;
  newSnap.otherSystem = oldSnap.otherSystem;

  newSnap.myTacticalNo = oldSnap.myTacticalNo;

  newSnap.firstResponsiblePlayer  = oldSnap.firstResponsiblePlayer;
  newSnap.secondResponsiblePlayer = oldSnap.secondResponsiblePlayer;
  newSnap.thirdResponsiblePlayer  = oldSnap.thirdResponsiblePlayer;

  newSnap.firstDangerousOpponent  = oldSnap.firstDangerousOpponent;
  newSnap.secondDangerousOpponent = oldSnap.secondDangerousOpponent;
  newSnap.thirdDangerousOpponent  = oldSnap.thirdDangerousOpponent;

  newSnap.ourGoalieFound   = oldSnap.ourGoalieFound;
  newSnap.otherGoalieFound = oldSnap.otherGoalieFound;

  for( int i = 0; i < 11; i++ )
    {
      newSnap.ourTeamRBTime[i]   = oldSnap.ourTeamRBTime[i];
      newSnap.otherTeamRBTime[i] = oldSnap.otherTeamRBTime[i];

      newSnap.ourTeamRBPos[i] = oldSnap.ourTeamRBPos[i];
      newSnap.otherTeamRBPos[i] = oldSnap.otherTeamRBPos[i];

      newSnap.ourTeamTNo[i]   = oldSnap.ourTeamTNo[i];
      newSnap.otherTeamTNo[i] = oldSnap.otherTeamTNo[i];

      newSnap.ourTeamStam[i].value = oldSnap.ourTeamStam[i].value;
      newSnap.ourTeamStam[i].age   = oldSnap.ourTeamStam[i].age + 1;

      newSnap.otherTeamStam[i].value = min( STAMINA_MAX, oldSnap.otherTeamStam[i].value + STAMINA_INC_MAX );
      newSnap.otherTeamStam[i].age   = oldSnap.ourTeamStam[i].age + 1;

      newSnap.otherTeamLastStamValue[i]   = oldSnap.otherTeamLastStamValue[i];
      newSnap.otherTeamLastStamPos[i]     = oldSnap.otherTeamLastStamPos[i];
      newSnap.otherTeamSimStamPossible[i] = oldSnap.otherTeamSimStamPossible[i];
    }

  newSnap.reportedTAction = oldSnap.reportedTAction;
}


bool
TacticalModule::feedbackWorld( Snapshot* snap )
{
  if( !snap )
    return false;

  // CAUTION: What if audio information comes first?
  // well: at the moment there is no audio information...
  if( snap->time < currentTime )
    return false;

  // simulate snapshot
  for( int t = currentTime; t < snap->time; t++ )
    simulate( t );

  currentTime = snap->time;
  TacticalSnapshot &tsnap = snapshots[getTimeIndex( currentTime )];

  // try to determine goalie numbers and my tactical number
  tsnap.updateTacticalNumbers( snap );

  // try to update stamina of our team members
  // (and of the opponent in those situations when stamina is reset)
  for( int i = 0; i < 11; i++ )
    {
      // kick-off, halftime, extended time -> stamina refresh
      if( snap->playMode == PLAY_MODE_HALF_TIME ||
	  snap->playMode == PLAY_MODE_TIME_EXTENDED ||
	  snap->playMode == PLAY_MODE_BEFORE_KICK_OFF )
	{
	  tsnap.ourTeamStam[i].value = STAMINA_MAX;
	  tsnap.ourTeamStam[i].age = 0;
	  tsnap.otherTeamStam[i].value = STAMINA_MAX;
	  tsnap.otherTeamStam[i].age = 0;
	  tsnap.otherTeamSimStamPossible[i] = false;
	}
      else
	{
	  // try to simulate opponent stamina
	  if( snap->otherTeam( i + 1 ).age <= 1 )
	    {
	      if( tsnap.otherTeamSimStamPossible[i] )
		{
		  // player movement since last stamina simulation
		  double d = dist( tsnap.otherTeamLastStamPos[i], snap->otherTeam( i + 1 ).pos );
		  
		  int n = tsnap.otherTeamStam[i].age - snap->otherTeam( i + 1 ).age;

		  d *= 0.9;   // error of about 10 % because of noise and sight blur
		  
		  double tmp = ( 1 - power( PLAYER_DECAY, n ) ) / ( 1 - PLAYER_DECAY );
		  
		  // maximum effect achieved by "sliding"
		  d = max( 0.0, d - PLAYER_SPEED_MAX * PLAYER_DECAY * tmp );
		  
		  // subtract minimum stamina needed for that position change
		  tsnap.otherTeamStam[i].value = tsnap.otherTeamLastStamValue[i] +
		    n * STAMINA_MAX - d / tmp / DASH_POWER_RATE;
		  tsnap.otherTeamStam[i].value = min( STAMINA_MAX,
						      max( EFFORT_DEC_THR * STAMINA_MAX, tsnap.otherTeamStam[i].value ) );

		  STAMSIMTRACE( cout << snap->time << " : Updated stamina of opp. " << i + 1 << " to "
				<< tsnap.otherTeamStam[i].value << ", had not done that for "
				<< n << " steps." << endl; );
		}
	      tsnap.otherTeamStam[i].age        = snap->otherTeam( i + 1 ).age;
	      tsnap.otherTeamSimStamPossible[i] = true;
	      tsnap.otherTeamLastStamValue[i]   = min( STAMINA_MAX, tsnap.otherTeamStam[i].value +
						       snap->otherTeam( i + 1 ).age * STAMINA_INC_MAX );
	      tsnap.otherTeamLastStamPos[i]     = snap->otherTeam( i + 1 ).pos;
	    }

	  // get my own stamina from snapshot
	  if( i == snap->my.number - 1 )
	    {
	      tsnap.ourTeamStam[i].value = snap->senseBody.stamina;
	      tsnap.ourTeamStam[i].age = 0;
	    }
	  // use audial information for the rest of my team
	  else if( receivedInfo.valid )
	    {
	      if( receivedInfo.time < currentTime )
		receivedInfo.ourTeamStam[i].age += currentTime - receivedInfo.time;
	      if( receivedInfo.ourTeamStam[i].age < tsnap.ourTeamStam[i].age )
		{
		  tsnap.ourTeamStam[i].age = receivedInfo.ourTeamStam[i].age;
		  tsnap.ourTeamStam[i].value = receivedInfo.ourTeamStam[i].value;
		}
	    }
	}
    }

  // find out how long it will take players to reach the ball
  // (and determine the 3 "most responsible ones") 
  tsnap.determineResponsibleFriends( snap );

  // find out how long it will take opponents to reach the ball
  // (and determine the 3 "most dangerous ones") 
  tsnap.determineDangerousOpponents( snap );

  // determine ball control for both teams
  tsnap.determineBallControl( snap,
			      receivedInfo.valid ? receivedInfo.ourBallControl : BC_UNKNOWN,
			      receivedInfo.valid ? receivedInfo.otherBallControl : BC_UNKNOWN );
			      
  // tactical action evaluation for our team
  // set emergency channel flag if there is a reported action to broadcast
  if( tsnap.determineOurTacticalAction( snap,
					receivedInfo.valid ? receivedInfo.ourAction : TacticalAction( TA_UNKNOWN ),
					currentTime ) )
    useEmergencyChannel = true;

  // tactical action evaluation for other team
  tsnap.determineOtherTacticalAction( snap,
				      receivedInfo.valid ? receivedInfo.otherAction : TacticalAction( TA_UNKNOWN ),
				      currentTime );
  
  // determine situation for both teams according to ball position and ball control
  // and set situation/info channel flag
  useInfoChannel = tsnap.determineSituation( snap );

  // determine opponent's tactical system
  // tsnap.determineOtherTacticalSystem( snap );

  // reported action is processed, so set it invalid
  tsnap.reportedTAction.quality = TQ_UNKNOWN;
  tsnap.reportedTAction.state = TA_UNKNOWN;
  tsnap.reportedTAction.performer = UNKNOWN;
  tsnap.reportedTAction.receiver = UNKNOWN;

  // audial information is processed, so set receivedInfo invalid
  receivedInfo.valid = false;

  return true;
}


bool
TacticalModule::feedbackAudio( AudioMsg &audioMsg )
{
  TacticalModuleInfo recInfo( false );
  audioMsg >> recInfo;

  if( recInfo.valid )
    {
      receivedInfo = recInfo;
      return true;
    }

  return false;
}


TacticalSnapshot *
TacticalModule::getSnapshot()
{
  // invalid time
  if( currentTime < 0 ) return NULL;

  // invalid snapshot
  if( snapshots[getTimeIndex( currentTime )].time != currentTime ) return NULL;

  return &snapshots[getTimeIndex( currentTime )];
}


bool
TacticalModule::getSayMsg( AudioMsg &audioMsg )
{
  bool ok = true;

  TacticalSnapshot &dummy = snapshots[getTimeIndex( currentTime )];
  TacticalModuleInfo sendInfo( ok, dummy );

  audioMsg.markWrite();
  audioMsg << sendInfo;
      
  // check if error occurred
  if( !audioMsg.getState() )
    {
      audioMsg.rollbackWrite().clearState();
      audioMsg << false;
      ok = false;
    }

  return ok;
}


//////////////////////////////////////////////////////////////////////
//
// Class TacticalModuleInfo
//
//////////////////////////////////////////////////////////////////////

TacticalModuleInfo::TacticalModuleInfo( bool _valid, const TacticalSnapshot &snap, int _sender, int _time )
{
  // if any info is to be sent, use valid = true;
  // to save space in the AudioMsg stream, use valid = false.
  
  valid = _valid;

  if( valid )
    {
      sender = _sender;
      time = _time;

      // set all other variables to values from TacticalSnapshot
      ourBallControl = snap.ourBallControl;
      otherBallControl = snap.otherBallControl;

      ourAction = snap.ourAction;
      otherAction = snap.otherAction;

      firstResponsiblePlayer = snap.firstResponsiblePlayer;
      secondResponsiblePlayer = snap.secondResponsiblePlayer;
      thirdResponsiblePlayer = snap.thirdResponsiblePlayer;
      
      firstDangerousOpponent = snap.firstDangerousOpponent;
      secondDangerousOpponent = snap.secondDangerousOpponent;
      thirdDangerousOpponent = snap.thirdDangerousOpponent;

      for( int i = 0; i < 11; i++ )
	{
	  ourTeamTNo[i] = snap.ourTeamTacticalNo( i + 1 );
	  otherTeamTNo[i] = snap.otherTeamTacticalNo( i + 1 );
	  ourTeamStam[i] = snap.ourTeamStamina( i + 1 );
	  otherTeamStam[i] = snap.otherTeamStamina( i + 1 );
	}
    }
  else
    {
      // set all variables invalid
      sender = UNKNOWN;
      time = -1;
      
      ourBallControl = BC_UNKNOWN;
      otherBallControl = BC_UNKNOWN;

      firstResponsiblePlayer = UNKNOWN;
      secondResponsiblePlayer = UNKNOWN;
      thirdResponsiblePlayer = UNKNOWN;

      firstDangerousOpponent = UNKNOWN;
      secondDangerousOpponent = UNKNOWN;
      thirdDangerousOpponent = UNKNOWN;
      
      for( int i = 0; i < 11; i++ )
	{
	  ourTeamTNo[i] = UNKNOWN;
	  otherTeamTNo[i] = UNKNOWN;
	  ourTeamStam[i].value = STAMINA_MAX;
	  ourTeamStam[i].age = ETERNAL;
	  otherTeamStam[i].value = STAMINA_MAX;
	  otherTeamStam[i].age = ETERNAL;
	}
    }
}


class AudioMsg &operator << ( AudioMsg &audioMsg, const TacticalModuleInfo &info )
{
  audioMsg << info.valid;

  // send tactical module information
  // space needed:
  // validity                        :    1 byte  (bool)
  // ball control for both teams     :    1 byte  (compressed into 1 NodaChar)
  // responsible players             :    3 bytes (3 NodaChars)
  // dangerous opponents             :    3 bytes (same)
  // tactical action for our team    :    7 bytes (1 for state, 1 for quality, 3 for time, 2 for players)
  // tactical action for other team  :    7 bytes (same)
  // tactical numbers for our team   :   11 bytes (11 NodaChars)
  // tactical numbers for other team :   11 bytes (same)
  // stamina for our team            :   44 bytes (3*11 for value, 1*11 for age)
  // --------------------------------------------
  // total                               88 bytes

  if( info.valid )
    {
       // send ballControl variables for both teams
       char c = 0;
       switch( info.ourBallControl )
 	{
 	case BC_SAFE    : c = 0; break;
 	case BC_DANGER  : c = 1; break;
 	case BC_LOST    : c = 2; break;
 	case BC_UNKNOWN : c = 3; break;
 	default         : break;
 	}
       switch( info.otherBallControl )
 	{
 	case BC_SAFE    : c += 5; break;
 	case BC_DANGER  : c += 10; break;
 	case BC_LOST    : c += 15; break;
 	case BC_UNKNOWN : c += 20; break;
 	default         : break;
 	}
       audioMsg << NodaChar( c );

       // send responsible and dangerous players

       // CAUTION : NodaChar does not know negative values (like UNKNOWN)
       audioMsg << NodaChar( char( info.firstResponsiblePlayer + 1 ) );
       audioMsg << NodaChar( char( info.secondResponsiblePlayer + 1 ) );
       audioMsg << NodaChar( char( info.thirdResponsiblePlayer + 1 ) );

       audioMsg << NodaChar( char( info.firstDangerousOpponent + 1 ) );
       audioMsg << NodaChar( char( info.secondDangerousOpponent + 1 ) );
       audioMsg << NodaChar( char( info.thirdDangerousOpponent + 1 ) );

       // send tactical actions for both teams

       switch( info.ourAction.state )
 	{
 	case TA_CONTROL  : c = 0; break;
 	case TA_PASS     : c = 1; break;
 	case TA_GOALKICK : c = 2; break;
 	case TA_REACT    : c = 3; break;
 	case TA_UNKNOWN  : c = 4; break;
 	default : break;
 	}
       audioMsg << NodaChar( c );

       switch( info.ourAction.quality )
 	{
 	case TQ_REPORTED        : c = 0; break;
 	case TQ_AUDIAL_REPORTED : c = 1; break;
 	case TQ_AUDIAL          : c = 2; break;
 	case TQ_GUESSED         : c = 3; break;
 	case TQ_UNKNOWN         : c = 4; break;
 	default : break;
 	}
       audioMsg << NodaChar( c );

       audioMsg << Int3B( info.ourAction.startingTime );
       audioMsg << NodaChar( char( info.ourAction.performer + 1 ) );
       audioMsg << NodaChar( char( info.ourAction.receiver + 1 ) );

       switch( info.otherAction.state )
 	{
 	case TA_CONTROL  : c = 0; break;
 	case TA_PASS     : c = 1; break;
 	case TA_GOALKICK : c = 2; break;
 	case TA_REACT    : c = 3; break;
 	case TA_UNKNOWN  : c = 4; break;
 	default : break;
 	}
       audioMsg << NodaChar( c );

       switch( info.otherAction.quality )
 	{
 	case TQ_REPORTED        : c = 0; break;
 	case TQ_AUDIAL_REPORTED : c = 1; break;
 	case TQ_AUDIAL          : c = 2; break;
 	case TQ_GUESSED         : c = 3; break;
 	case TQ_UNKNOWN         : c = 4; break;
 	default : break;
 	}
       audioMsg << NodaChar( c );

       audioMsg << Int3B( info.otherAction.startingTime );
       audioMsg << NodaChar( char( info.otherAction.performer + 1 ) );
       audioMsg << NodaChar( char( info.otherAction.receiver + 1 ) );

       // send tactical numbers for both teams

       for( int i = 0; i < 11; i++ )
 	{
	  audioMsg << NodaChar( char( info.ourTeamTNo[i] + 1 ) );
  	  audioMsg << NodaChar( char( info.otherTeamTNo[i] + 1 ) );
 	}

       // send stamina for our team

       for( int i = 0; i < 11; i++ )
	 {
	   audioMsg << Int3B( int( info.ourTeamStam[i].value ) );
	   audioMsg << NodaChar( char( min( 63, info.ourTeamStam[i].age ) ) );
	 }
    }
  
  return audioMsg;
}


class AudioMsg &operator >> ( AudioMsg &audioMsg, TacticalModuleInfo &info )
{
  audioMsg >> info.valid;
  if( info.valid )
    {
       NodaChar c;
       Int3B d;

       info.sender = audioMsg.getSource();
       info.time = audioMsg.getServerTime();

       // get ballControl variables for both teams
       audioMsg >> c;
       switch( int( c ) % 5 )
 	{
 	case 0  : info.ourBallControl = BC_SAFE; break;
 	case 1  : info.ourBallControl = BC_DANGER; break;
 	case 2  : info.ourBallControl = BC_LOST; break;
 	case 3  : info.ourBallControl = BC_UNKNOWN; break;
 	default : break;
 	}
       switch( int( c ) / 5 )
 	{
 	case 1  : info.otherBallControl = BC_SAFE; break;
 	case 2  : info.otherBallControl = BC_DANGER; break;
 	case 3  : info.otherBallControl = BC_LOST; break;
 	case 4  : info.otherBallControl = BC_UNKNOWN; break;
 	default : break;
 	}

       // get responsible and dangerous players

       // CAUTION : NodaChar does not know negative values (like UNKNOWN)
       audioMsg >> c;
       info.firstResponsiblePlayer = int( char( c ) ) - 1;
       audioMsg >> c;
       info.secondResponsiblePlayer = int( char( c ) ) - 1;
       audioMsg >> c;
       info.thirdResponsiblePlayer = int( char( c ) ) - 1;

       audioMsg >> c;
       info.firstDangerousOpponent = int( char( c ) ) - 1;
       audioMsg >> c;
       info.secondDangerousOpponent = int( char( c ) ) - 1;
       audioMsg >> c;
       info.thirdDangerousOpponent = int( char( c ) ) - 1;

       // get tactical actions for both teams

       audioMsg >> c;
       switch( char( c ) )
 	{
 	case 0  : info.ourAction.state = TA_CONTROL; break;
 	case 1  : info.ourAction.state = TA_PASS; break;
 	case 2  : info.ourAction.state = TA_GOALKICK; break;
 	case 3  : info.ourAction.state = TA_REACT; break;
 	case 4  : info.ourAction.state = TA_UNKNOWN; break;
 	default : break;
 	}

       audioMsg >> c;
       switch( char( c ) )
 	{
 	case 0 : info.ourAction.quality = TQ_REPORTED; break;
 	case 1 : info.ourAction.quality = TQ_AUDIAL_REPORTED; break;
 	case 2 : info.ourAction.quality = TQ_AUDIAL; break;
 	case 3 : info.ourAction.quality = TQ_GUESSED; break;
 	case 4 : info.ourAction.quality = TQ_UNKNOWN; break;
 	default : break;
 	}
       
       audioMsg >> d;
       info.ourAction.startingTime = int( d );
       audioMsg >> c;
       info.ourAction.performer = int( char( c ) ) - 1;
       audioMsg >> c;
       info.ourAction.receiver = int( char( c ) ) - 1;
       
       audioMsg >> c;
       switch( char( c ) )
 	{
 	case 0  : info.otherAction.state = TA_CONTROL; break;
 	case 1  : info.otherAction.state = TA_PASS; break;
 	case 2  : info.otherAction.state = TA_GOALKICK; break;
 	case 3  : info.otherAction.state = TA_REACT; break;
 	case 4  : info.otherAction.state = TA_UNKNOWN; break;
 	default : break;
 	}

       audioMsg >> c;
       switch( char( c ) )
 	{
 	case 0 : info.otherAction.quality = TQ_REPORTED; break;
 	case 1 : info.otherAction.quality = TQ_AUDIAL_REPORTED; break;
 	case 2 : info.otherAction.quality = TQ_AUDIAL; break;
 	case 3 : info.otherAction.quality = TQ_GUESSED; break;
 	case 4 : info.otherAction.quality = TQ_UNKNOWN; break;
 	default : break;
 	}

       audioMsg >> d;
       info.otherAction.startingTime = int( d );
       audioMsg >> c;
       info.otherAction.performer = int( char( c ) ) - 1;
       audioMsg >> c;
       info.otherAction.receiver = int( char( c ) ) - 1;

       // get tactical numbers for both teams

       for( int i = 0; i < 11; i++ )
	 {
	  audioMsg >> c;
 	  info.ourTeamTNo[i] = int( char( c ) ) - 1;
  	  audioMsg >> c;
  	  info.otherTeamTNo[i] = int( char( c ) ) - 1;
 	}

       // get stamina for our team

       for( int i = 0; i < 11; i++ )
	 {
	   audioMsg >> d;
	   info.ourTeamStam[i].value = double( int( d ) );
	   audioMsg >> c;
	   if( int( char( c ) ) == 63 )
	     info.ourTeamStam[i].age = ETERNAL;
	   else
	     info.ourTeamStam[i].age = int( char( c ) );
	 }
    }

  return audioMsg;
}
