//////////////////////////////////////////////////////////////////////
// Mainz Rolling Brains RoboCup agents developed for the Simulator 
// League of the RoboCup 2000 competition
//
// Copyright (C) 1998, 1999, 2000 Team Members 
// of the Mainz Rolling Brains RoboCup Team 
//
// This program under the terms of the GNU General Public License as
// published by the Free Software Foundation; version 2 of the
// License. See the README file in the root directory of our archive.
//////////////////////////////////////////////////////////////////////

#ifndef _TACTICALMODULE_H_
#define _TACTICALMODULE_H_


#include "worldmodel.h"
#include "Config.h"


//////////////////////////////////////////////////////////////////////
//
// enum BallControlState
//
// tells if (and how) a certain team (not player!) controls the ball
//
//////////////////////////////////////////////////////////////////////

enum BallControlState
{
  BC_SAFE,        // ball is under control
  BC_DANGER,      // ball is still under control, but not safe
  BC_LOST,        // opponent controls ball
  BC_UNKNOWN      // decision impossible
};


//////////////////////////////////////////////////////////////////////
//
// enum SituationState
//
// tells the situation of the match for a certain team (not player!)
//
//////////////////////////////////////////////////////////////////////

enum SituationState
{
  SI_DEFENSIVE,   // we defend
  SI_MIDFIELD,    // ball in midfield
  SI_OFFENSIVE,   // we started an offense
  SI_UNKNOWN      // decision impossible
};


//////////////////////////////////////////////////////////////////////
//
// enum TacticalActionState
//
// tells what a certain team (not player!) is doing at the moment
//
//////////////////////////////////////////////////////////////////////

enum TacticalActionState
{
  TA_CONTROL,     // a team member controls the ball
  TA_PASS,        // team is playing a pass
  TA_GOALKICK,    // a team member tried a goal kick
  TA_REACT,       // team tries to (re)conquer the ball
  TA_UNKNOWN      // decision impossible
};


//////////////////////////////////////////////////////////////////////
//
// enum TacticalActionQuality
//
// tells where a tactical action comes from
//
//////////////////////////////////////////////////////////////////////

enum TacticalActionQuality
{
  TQ_REPORTED,          // player reported that action (i.e. performed it himself)
  TQ_AUDIAL_REPORTED,   // a team member told us about his reported action
  TQ_AUDIAL,            // told by a team member but lesser quality
  TQ_GUESSED,           // guessed from worldmodel data
  TQ_UNKNOWN            // decision impossible
};


//////////////////////////////////////////////////////////////////////
//
// Class TacticalAction
//
// holds TacticalActionState for a team and (in case of PASS or CONTROL)
// the numbers of the involved players
//
//////////////////////////////////////////////////////////////////////

class TacticalAction
{
public:
  TacticalAction( TacticalActionState _state = TA_UNKNOWN,
		  int _perPlayer = UNKNOWN,
		  int _recPlayer = UNKNOWN,
		  int _startingTime = -1 )
  {
    state = _state;
    performer = ( ( state == TA_UNKNOWN || state == TA_REACT ) ? UNKNOWN : _perPlayer );
    receiver = ( state == TA_PASS ? _recPlayer : UNKNOWN );
    startingTime = _startingTime;
  }

  ~TacticalAction() {}

  TacticalActionState state;

  // information quality 
  TacticalActionQuality quality;

  // timestep when that action started
  int startingTime;

  // if state is TA_CONTROL  : number of player who holds the ball
  // if state is TA_PASS     : number of player who sent the pass
  // if state is TA_GOALKICK : number of player who tried the goal kick
  // if state is TA_REACT    : UNKNOWN ( == -1 )
  // if state is TA_UNKNOWN  : UNKNOWN ( == -1 ) 
  int performer;

  // if state is TA_PASS     : number of player who receives the pass
  // otherwise               : UNKNOWN ( == -1 )
  int receiver;
};


//////////////////////////////////////////////////////////////////////
//
// Class TacticalSystem
//
// holds system we currently play (e.g. 4-4-2)
//
//////////////////////////////////////////////////////////////////////

class TacticalSystem
{
public:
  TacticalSystem( int _numDef = 4, int _numMid = 4, int _numOff = 2 );

  ~TacticalSystem() {}

  int numDefenders;
  int numMidfield;
  int numOffenders;
};


//////////////////////////////////////////////////////////////////////
//
// Class Stamina
//
// holds a simulated stamina value and its age 
//
//////////////////////////////////////////////////////////////////////

class Stamina
{
public:
  Stamina( double _value = STAMINA_MAX, int _age = 0 )
    : value( _value ), age( _age ) {}
  ~Stamina() {}

  double value;
  int    age;
};


//////////////////////////////////////////////////////////////////////
//
// Class TacticalSnapshot
//
//////////////////////////////////////////////////////////////////////

class TacticalSnapshot
{
public:
  TacticalSnapshot();
  ~TacticalSnapshot() {}
  
  // servertime this tactical snapshot belongs to
  int time;

  // current state of ball control for both teams
  BallControlState   ourBallControl;
  BallControlState   otherBallControl;

  // current situation state for both teams
  SituationState     ourSituation;
  SituationState     otherSituation;

  // tactical action both teams currently perform
  TacticalAction     ourAction;
  TacticalAction     otherAction;

  // tactical systems both teams currently play
  TacticalSystem     ourSystem;
  TacticalSystem     otherSystem;

  // edges for changing ourBallControl:
  // ourBallControl changes to BC_SAFE if our intercept edge is > ourEdge
  int                ourEdge;
  // ourBallControl changes to BC_LOST if opponent's intercept edge is > otherEdge
  int                otherEdge;

  // my tactical number in the tactical system we currently play (can be UNKNOWN!)
  // --------------|----------------------------------------------
  // role          |  tactical number
  // --------------|----------------------------------------------
  // goalie        |  1
  // defenders     |  2 .. numDefenders+1
  // midfield      |  numDefenders+2 .. numDefenders+numMidfield+1
  // offenders     |  11-numOffenders .. 11
  // -------------------------------------------------------------
  int myTacticalNo;

  // get tactical number of player with number "n" in our/other team (can be UNKNOWN!)
  int ourTeamTacticalNo( int n ) const
  { ensure( n >= 1 && n <= 11 ); return ourTeamTNo[n-1]; }

  int otherTeamTacticalNo( int n ) const
  { ensure( n >= 1 && n <= 11 ); return otherTeamTNo[n-1]; }

  // returns if player with tactical number (NOT number) "tn" of our team is a defender/midfield/offender
  // if player's number is UNKNOWN, false is returned
  // goalie is neither defender nor midfield nor offender
  bool      ourTeamDefender( int tn ) const;
  bool      ourTeamMidfield( int tn ) const;
  bool      ourTeamOffender( int tn ) const;

  // same for other team
  bool      otherTeamDefender( int tn ) const;
  bool      otherTeamMidfield( int tn ) const;
  bool      otherTeamOffender( int tn ) const;

  // returns NUMBER of the player to the left/right of the player with
  // TACTICAL NUMBER "tn" of our team
  // if that player has no left/right neighbour or if he has one but we
  // do not know his number, UNKNOWN is returned
  int       ourTeamLeftNeighbour( int tn ) const;
  int       ourTeamRightNeighbour( int tn ) const;

  // same for other team
  int       otherTeamLeftNeighbour( int tn ) const;
  int       otherTeamRightNeighbour( int tn ) const;
  
  // the three players who will reach the ball first
  // this is just a guess because of lack of reliable stamina information
  // CAUTION 1: ball control module should not blindly use these numbers
  // CAUTION 2: Numbers can be UNKNOWN ( == -1 ) !
  int       firstResponsiblePlayer;
  int       secondResponsiblePlayer;
  int       thirdResponsiblePlayer;

  // most dangerous opponents' numbers
  // (i.e. opponents who will reach the ball first)
  // CAUTION 1: ball control module should not blindly use these numbers
  // CAUTION 2: Numbers can be UNKNOWN ( == -1 ) !
  int       firstDangerousOpponent;
  int       secondDangerousOpponent;
  int       thirdDangerousOpponent;

  // get approximated number of steps player no. "num" will need to reach the ball
  // returns -1 if player is not among the three most responsible ones
  // or if "num" is not a valid player number
  // (should be called with parameter first/second/thirdResponsiblePlayer)
  int       ourTeamReachBallTime( int num )
    { return num < 1 ? -1 : num > 11 ? -1 : ourTeamRBTime[num-1]; };

  // same for opponent team
  // (should be called with first/second/thirdDangerousOpponent)
  int       otherTeamReachBallTime( int num )
    { return num < 1 ? -1 : num > 11 ? -1 : otherTeamRBTime[num-1]; };

  // get approximated position where player no. "num" will reach the ball
  // returns something weird if player is not among the three most responsible ones
  // (should be called with parameter first/second/thirdResponsiblePlayer)
  const Vector2d &ourTeamReachBallPos( int num )
    { ensure( num >= 1 && num <= 11 ); return ourTeamRBPos[num-1]; };

  // same for opponent team
  // (should be called with first/second/thirdDangerousOpponent)
  const Vector2d &otherTeamReachBallPos( int num )
    { ensure( num >= 1 && num <= 11 ); return otherTeamRBPos[num-1]; };
  
  // get approximated stamina of player no. "num" of our team
  // "num" should be a valid player number
  const Stamina&   ourTeamStamina( int num ) const
  {
    if( num < 1 || num > 11 )
      cerr << "ERROR: ourTeamStamina() call with parameter " << num << endl;
    ensure( num >= 1 && num <= 11 );
    return ourTeamStam[num-1];
  }

  // same for opponent team
  const Stamina&   otherTeamStamina( int num ) const
  {
    if( num < 1 || num > 11 )
      cerr << "ERROR: otherTeamStamina() call with parameter " << num << endl;
    ensure( num >= 1 && num <= 11 );
    return otherTeamStam[num-1];
  }

  // Report a tactical action of OUR team to the tactical snapshot.
  // This can e.g. be used by a player who sent a pass or tried a goalkick.
  // It should only be used to report actions the player can be sure of because
  // he is involved. Everything else should be detected by the tactical module itself.
  // Returns false if information is invalid.
  bool      reportOurTacticalAction( TacticalActionState state,
				     int perPlayer = UNKNOWN,
				     int recPlayer = UNKNOWN );

private:
  // do I know the goalie numbers? (important for determination of my tactical number)
  bool      ourGoalieFound;
  bool      otherGoalieFound;

  // approximated numbers of steps players will need to reach the ball
  // public access is via ourTeamReachBallTime() / otherTeamReachBallTime()
  int       ourTeamRBTime[11];
  int       otherTeamRBTime[11];

  // approximated points where players will reach the ball
  // public access is via ourTeamReachBallPos() / otherTeamReachBallPos()
  Vector2d  ourTeamRBPos[11];
  Vector2d  otherTeamRBPos[11];

  // tactical numbers of players in both teams
  // public access is via ourTeamTacticalNo() / otherTeamTacticalNo()
  int       ourTeamTNo[11];
  int       otherTeamTNo[11];

  // guessed stamina for both teams
  // public access is via ourTeamStamina() / otherTeamStamina()
  Stamina   ourTeamStam[11];
  Stamina   otherTeamStam[11];

  // position that opponent players held when their stamina was updated for the last time 
  Vector2d  otherTeamLastStamPos[11];

  // stamina value at that time
  double    otherTeamLastStamValue[11];

  // false at kick-off (esp. halftime) when we have not seen the opponents yet
  bool      otherTeamSimStamPossible[11];

  // tactical action agent reported
  // (is stored here for later update by tactical module)
  TacticalAction   reportedTAction;

  // try (if needed) to determine goalies and then my tactical number
  void updateTacticalNumbers( Snapshot *snap );

  // fill in first/second/third... variables and RBTime arrays
  void determineResponsibleFriends( Snapshot *snap );
  void determineDangerousOpponents( Snapshot *snap );

  // determine if situation is defensive/midfield/offensive for the teams
  // returns if situation channel should be used to inform team members
  bool determineSituation( Snapshot *snap );

  // determine ball control for both teams
  void determineBallControl( Snapshot *snap,
			     BallControlState ourBallControlAudio,
			     BallControlState otherBallControlAudio );

  // determine tactical action for our team
  // returns if emergency channel should be used to inform team members
  bool determineOurTacticalAction( Snapshot *snap, const TacticalAction& audioAction, int currentTime );

  // determine tactical action for other team
  void determineOtherTacticalAction( Snapshot *snap, const TacticalAction& audioAction, int currentTime );

  // determine opponent's tactical system
  void determineOtherTacticalSystem( Snapshot *snap );

  friend class TacticalModule;
};


//////////////////////////////////////////////////////////////////////
//
// Class TacticalModuleInfo
//
//////////////////////////////////////////////////////////////////////

class TacticalModuleInfo
{
public:
  TacticalModuleInfo( bool _valid = false,
		      const TacticalSnapshot &snap = TacticalSnapshot(),
		      int _sender = UNKNOWN,
		      int _time = -1 );
  ~TacticalModuleInfo() {}

  bool valid;

  int  sender;   // sending player
  int  time;     // servertime when object was sent

  // variables from TacticalSnapshot

  BallControlState   ourBallControl;
  BallControlState   otherBallControl;

  TacticalAction     ourAction;
  TacticalAction     otherAction;

  int       firstResponsiblePlayer;
  int       secondResponsiblePlayer;
  int       thirdResponsiblePlayer;

  int       firstDangerousOpponent;
  int       secondDangerousOpponent;
  int       thirdDangerousOpponent;

  int       ourTeamTNo[11];
  int       otherTeamTNo[11];  

  Stamina   ourTeamStam[11];
  Stamina   otherTeamStam[11];
};


//////////////////////////////////////////////////////////////////////
//
// Class TacticalModule
//
//////////////////////////////////////////////////////////////////////

class TacticalModule
{
public:
  TacticalModule( int _numSnapshots = 10 );
  
  TacticalModule( const TacticalModule& tm );
  
  TacticalModule& TacticalModule::operator = ( const TacticalModule& tm );
  
  ~TacticalModule();

  // feed new information about world into module
  bool feedbackWorld( Snapshot* snap );
  
  // feed audio information into module
  bool feedbackAudio( AudioMsg &audioMsg );

  // get simulated tactics for time == currentTime
  TacticalSnapshot* getSnapshot();

  // write information into AudioMsg stream
  bool getSayMsg( AudioMsg &audioMsg );

  // get private property currentTime
  int getCurrentTime()
  { return currentTime; }

  // use emergency channel for tactical model broadcast?
  bool useEmergencyChannel;

  // use situation/info channel for broadcast?
  bool useInfoChannel;

private:
  TacticalSnapshot*   snapshots;      // Simulations of tactics at different time steps
  TacticalModuleInfo  receivedInfo;   // latest audial information

  int                 numSnapshots;   // number of entries in "snapshots"

  // get Index for lookup in "snapshots" array (value for time<0 is 0)
  int                getTimeIndex( int time )
  { return ( time > 0 ) ? time % numSnapshots : 0; }

  // simulate data in snapshot for (time + 1) with data in previous snapshot(s)
  void               simulate( int time );

  // latest time for which there is a valid snapshot
  int currentTime;

};


// put tactical information to AudioMsg-Stream
class AudioMsg& operator << ( AudioMsg &aumsg, const TacticalModuleInfo &info );


// fill with information from AudioMsg-Stream
class AudioMsg& operator >> ( AudioMsg &aumsg, TacticalModuleInfo &info );


#endif  // _TACTICALMODULE_H_
